@@ -1,6 +1,6 @@
 ;;;; remote-ftp.jl -- Remote file access via FTP
 ;;;  Copyright (C) 1998 John Harper <john@dcs.warwick.ac.uk>
-;;;  $Id: remote-ftp.jl,v 1.25 1999/11/25 23:20:27 john Exp $
+;;;  $Id: remote-ftp.jl,v 1.27 2000/03/24 10:28:54 john Exp $
 
 ;;; This file is part of Jade.
 
@@ -167,7 +167,7 @@ file types.")
   (let
       ((process (make-process (lambda (data)
 				(remote-ftp-output-filter session data))
-			      'remote-ftp-sentinel
+			      remote-ftp-sentinel
 			      nil ftp-program
 			      (append remote-ftp-args
 				      (list (aref session remote-ftp-host))))))
@@ -214,7 +214,7 @@ file types.")
 (defun remote-ftp-close-all ()
   "Close all running FTP subprocesses."
   (interactive)
-  (mapc 'remote-ftp-close-session remote-ftp-sessions))
+  (mapc remote-ftp-close-session remote-ftp-sessions))
 
 (defun remote-ftp-get-session-by-process (process)
   (catch 'return
@@ -777,6 +777,7 @@ file types.")
 					 type		;access type
 					 file-name	;remote name
 					 local-file))	;local copy
+	  (remote-register-file-handle local-fh)
 	  local-fh))
        (t
 	(let
