@@ -1,5 +1,5 @@
 # rules.mk.sh -- Build dynamically-loadable objects for librep
-# $Id: rules.mk.sh,v 1.4 1999/10/10 15:39:44 john Exp $
+# $Id: rules.mk.sh,v 1.5 2000/02/02 14:02:15 john Exp $
 
 repdir=$1
 repexecdir=$2
@@ -15,10 +15,10 @@ repdocfile=$repdocfile
 rep_LIBTOOL=\$(repexecdir)/libtool
 
 # use this like:
-# libfoo.la : foo.lo bar.lo
+# foo.la : foo.lo bar.lo
 #	\$(rep_DL_LD) link-opts...
 
-rep_DL_LD=\$(rep_LIBTOOL) \$(CC) -module -rpath \$(repexecdir)
+rep_DL_LD=\$(rep_LIBTOOL) \$(CC) -avoid-version -module -rpath \$(repexecdir)
 rep_DL_INSTALL=\$(rep_LIBTOOL) \$(INSTALL)
 rep_DL_UNINSTALL=\$(rep_LIBTOOL) rm
 
