@@ -1,6 +1,6 @@
 /* lisp.c -- Core of the Lisp, reading and evaluating...
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: lisp.c,v 1.51 1999/11/29 18:37:54 john Exp $
+   $Id: lisp.c,v 1.74.2.1 2000/04/12 13:20:37 john Exp $
 
    This file is part of Jade.
 
@@ -43,6 +43,7 @@ char *alloca ();
 #include <string.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <assert.h>
 
 #ifdef NEED_MEMORY_H
 # include <memory.h>
@@ -117,14 +118,10 @@ DEFSYM(arith_error, "arith-error");
 DEFSTRING(err_arith_error, "Arithmetic error");
 DEFSYM(term_interrupt, "term-interrupt");
 
-#ifdef MINSTACK
-DEFSYM(stack_error, "stack-error");
-DEFSTRING(err_stack_error, "Out of stack space");
-#endif
-
 DEFSYM(debug_on_error, "debug-on-error");
 DEFSYM(backtrace_on_error, "backtrace-on-error");
-DEFSYM(debug_macros, "debug-macros"); /*
+DEFSYM(debug_macros, "debug-macros");
+DEFSYM(error_handler_function, "error-handler-function"); /*
 ::doc:debug-on-error::
 When an error is signalled this variable controls whether or not to
 enter the Lisp debugger immediately. If the variable's value is t or a
@@ -139,6 +136,10 @@ entered.
 ::end::
 ::doc:debug-macros::
 When nil, the debugger isn't entered while expanding macro definitions.
+::end::
+::doc:error-handler-function::
+When set to a function value, called with two arguments (error type
+and data) when lisp errors occur.
 ::end:: */
 
 DEFSYM(print_escape, "print-escape");
@@ -176,14 +177,18 @@ repv rep_env;
 /* A list of the special variables that may be accessed in this
    environment, or Qt to denote all specials. This always contains
    at least one cons cell, which is skipped. */
-repv rep_special_env;
+repv rep_special_env, rep_special_bindings;
+
+/* The bytecode interpreter to use. A three-arg subr or a null pointer */
+repv (*rep_bytecode_interpreter)(repv code, repv consts,
+				 repv stack, repv frame);
 
 /* The lisp-call backtrace; also used for saving and restoring
    the current environment */
 struct rep_Call *rep_call_stack;
 
 /* Prevent infinite recursion */
-static int lisp_depth, max_lisp_depth = 500;
+int rep_lisp_depth, rep_max_lisp_depth = 500;
 
 /* Used to avoid costly interrupt checking too often */
 int rep_test_int_counter = 0;
@@ -199,14 +204,22 @@ void (*rep_test_int_fun)(void) = default_test_int;
 
 /* Reading */
 
-/* All of the read-related functions are now stream based. This will
-   probably add some (much?) overhead but I think it's worth it?
-
-   The `c' variable which keeps coming up is the lookahead character,
-   since each read*() routine normally has to look at the next character
-   to see if it's what it wants. If not, this char is given to someone
-   else... */
+/* The `c' variable which keeps coming up is the lookahead character,
+   since each reader function normally has to look at the next character
+   to see if it's what it wants. If not, the lookahead is given to
+   someone else or unread before exiting... */
 
+/* inline common case of reading from local files; this appears to
+   decrease startup time by about 25% */
+static inline int
+fast_getc (repv stream)
+{
+    if (rep_FILEP (stream) && rep_LOCAL_FILE_P (stream))
+	return getc (rep_FILE (stream)->file.fh);
+    else
+	return rep_stream_getc (stream);
+}
+ 
 DEFSTRING(nodot, "Nothing to dot second element of cons-cell to");
 
 static repv
@@ -229,35 +242,57 @@ read_list(repv strm, register int *c_p)
 	case '\t':
 	case '\n':
 	case '\f':
-	    *c_p = rep_stream_getc(strm);
+	    *c_p = fast_getc(strm);
 	    continue;
 
 	case ';':
 	    {
 		register int c;
-		while((c = rep_stream_getc(strm)) != EOF && c != '\n' && c != '\f')
+		while((c = fast_getc(strm)) != EOF && c != '\n' && c != '\f')
 		    ;
-		*c_p = rep_stream_getc(strm);
+		*c_p = fast_getc(strm);
 		continue;
 	    }
 
-	case '.':
-	    *c_p = rep_stream_getc(strm);
-	    if(last)
-	    {
-		if(!(rep_CDR(last) = rep_readl(strm, c_p)))
-		    result = rep_NULL;
-	    }
-	    else
-		result = Fsignal(Qinvalid_read_syntax,
-				    rep_LIST_1(rep_VAL(&nodot)));
-	    break;
-
 	case ')':
 	case ']':
 	    *c_p = rep_stream_getc(strm);
 	    goto end;
 	    
+	case '.':
+	    *c_p = rep_stream_getc(strm);
+	    switch (*c_p)
+	    {
+	    case EOF:
+		result = Fsignal(Qend_of_stream, rep_LIST_1(strm));
+		goto end;
+
+	    case ' ': case '\t': case '\n': case '\f':
+		if(last)
+		{
+		    repv this = rep_readl(strm, c_p);
+		    if (this != rep_NULL)
+			rep_CDR (last) = this;
+		    else
+		    {
+			result = rep_NULL;
+			goto end;
+		    }
+		}
+		else
+		{
+		    result = Fsignal(Qinvalid_read_syntax,
+				     rep_LIST_1(rep_VAL(&nodot)));
+		    goto end;
+		}
+		continue;
+
+	    default:
+		rep_stream_ungetc (strm, *c_p);
+		*c_p = '.';
+	    }
+	    /* fall through */
+
 	default:
 	    {
 		register repv this = Fcons(Qnil, Qnil);
@@ -276,15 +311,12 @@ end:
     return result;
 }
 
-DEFSTRING(buf_overflow, "Internal buffer overflow");
-DEFSTRING(int_overflow, "Integer limit exceeded");
-
 /* Could be a symbol or a number */
 static repv
 read_symbol(repv strm, int *c_p)
 {
-#define SYM_BUF_LEN 240
     static repv buffer = rep_NULL;
+    static size_t buflen = 240;
 
     repv result;
     u_char *buf;
@@ -293,24 +325,33 @@ read_symbol(repv strm, int *c_p)
 
     /* For parsing numbers, while radix != zero, it might still be
        an integer that we're reading. */
-    int radix = -1, sign = 1;
-    long value = 0;
+    int radix = -1, sign = 1, nfirst = 0;
+    rep_bool exact = rep_TRUE, rational = rep_FALSE, exponent = rep_FALSE;
 
-    if(buffer == rep_NULL)
+    if (buffer == rep_NULL)
     {
-	buffer = rep_make_string(SYM_BUF_LEN + 1);
-	rep_mark_static(&buffer);
+	buffer = rep_make_string (buflen + 2);
+	rep_mark_static (&buffer);
     }
 
     buf = rep_STR(buffer);
 
-    while((c != EOF) && (i < SYM_BUF_LEN))
+    while (c != EOF)
     {
+	if (i == buflen)
+	{
+	    repv new;
+	    buflen = buflen * 2;
+	    new = rep_make_string (buflen + 2);
+	    memcpy (rep_STR (new), buf, buflen / 2);
+	    buf = rep_STR (new);
+	}
 	switch(c)
 	{
 	case ' ':  case '\t': case '\n': case '\f':
 	case '(':  case ')':  case '[':  case ']':
-	case '\'': case '"':  case ';':
+	case '\'': case '"':  case ';':  case ',':
+	case '`':
 	    goto done;
 
 	case '\\':
@@ -324,7 +365,7 @@ read_symbol(repv strm, int *c_p)
 	case '|':
 	    radix = 0;
 	    c = rep_stream_getc(strm);
-	    while((c != EOF) && (c != '|') && (i < SYM_BUF_LEN))
+	    while((c != EOF) && (c != '|') && (i < buflen))	/* XXX */
 	    {
 		buf[i++] = c;
 		c = rep_stream_getc(strm);
@@ -338,8 +379,11 @@ read_symbol(repv strm, int *c_p)
 	    {
 		/* It still may be a number that we're parsing */
 		if(i == 0 && (c == '-' || c == '+'))
+		{
 		    /* A leading sign */
 		    sign = (c == '-') ? -1 : 1;
+		    nfirst = i + 1;
+		}
 		else
 		{
 		    switch(radix)
@@ -347,26 +391,36 @@ read_symbol(repv strm, int *c_p)
 		    case -1:
 			/* Deduce the base next (or that we're not
 			   looking at a number) */
-			if(!isdigit(c))
+			if (c == '.')
+			{
+			    radix = 10;
+			    exact = rep_FALSE;
+			}
+			else if(!isdigit(c))
 			    radix = 0;
 			else if(c == '0')
 			    radix = 1;	/* octal or hex */
 			else
-			{
 			    radix = 10;
-			    value = c - '0';
-			}
 			break;
 
 		    case 1:
 			/* We had a leading zero last character. If
 			   this char's an 'x' it's hexadecimal. */
 			if(toupper(c) == 'X')
+			{
 			    radix = 16;
+			    nfirst = i + 1;
+			}
 			else if(isdigit(c))
 			{
 			    radix = 8;
-			    value = (value * radix) + (c - '0');
+			    nfirst = i;
+			}
+			else if (c == '.' || c == 'e' || c == 'E')
+			{
+			    radix = 10;
+			    exact = rep_FALSE;
 			}
 			else
 			    radix = 0;
@@ -375,53 +429,74 @@ read_symbol(repv strm, int *c_p)
 		    default:
 			/* Now we're speculatively reading a number
 			   of base radix. */
-			if(radix <= 10)
+			switch (c)
 			{
-			    if(c >= '0' && c <= ('0' + radix - 1))
-				value = value * radix + (c - '0');
+			case '.':
+			    if (exact && !rational)
+				exact = rep_FALSE;
 			    else
 				radix = 0;
-			}
-			else
-			{
-			    /* hex */
-			    if(isxdigit(c))
-			    {
-				value = (value * 16
-					 + ((c >= '0' && c <= '9')
-					    ? c - '0'
-					    : 10 + toupper(c) - 'A'));
-			    }
+			    break;
+
+			case '/':
+			    if (exact && !rational)
+				rational = rep_TRUE;
 			    else
 				radix = 0;
+			    break;
+
+			case '-': case '+':
+			    if (!exponent)
+				goto do_default;
+			    break;
+
+			case 'e': case 'E':
+			    if (radix != 16)
+			    {
+				if (!rational && !exponent)
+				{
+				    exponent = rep_TRUE;
+				    exact = rep_FALSE;
+				}
+				else
+				    radix = 0;
+				break;
+			    }
+			    /* fall through */
+
+			default: do_default:
+			    if(radix <= 10
+			       && !(c >= '0' && c <= ('0' + radix - 1)))
+			    {
+				radix = 0;
+			    }
+			    else if(!isxdigit(c))
+				radix = 0;
 			}
 		    }
 		}
 	    }
 	    buf[i++] = c;
 	}
-	c = rep_stream_getc(strm);
-    }
-    if(i >= SYM_BUF_LEN)
-    {
-	/* Guess I'd better fix this! */
-	return Fsignal(Qerror, rep_LIST_1(rep_VAL(&buf_overflow)));
+	c = fast_getc(strm);
     }
 done:
-    if(radix > 0
-       /* Ensure that we don't accept `0x' as hex zero */
-       && ((radix != 16) || (i > 2)))
+    buf[i] = 0;
+    if (radix > 0 && nfirst < i)
     {
-	/* It was a number */
-	value *= sign;
-	if(value < rep_LISP_MIN_INT || value > rep_LISP_MAX_INT)
-	    return Fsignal(Qarith_error, rep_LIST_1(rep_VAL(&int_overflow)));
-	result = rep_MAKE_INT(value);
+	/* It was a number of some sort */
+	if (radix == 1)
+	    result = rep_MAKE_INT (0);
+	else
+	    result = rep_parse_number (buf + nfirst, i - nfirst, radix, sign,
+				       !exact ? rep_NUMBER_FLOAT
+				       : rational ? rep_NUMBER_RATIONAL : 0);
+	if (result == rep_NULL)
+	    goto intern;
     }
     else
     {
-	buf[i] = 0;
-	rep_set_string_len(buffer, i);
+intern:	rep_set_string_len(buffer, i);
 	if(!(result = Ffind_symbol(rep_VAL(buffer), Qnil))
 	   || (rep_NILP(result) && strcmp(buf, "nil")))
 	{
@@ -513,7 +588,7 @@ read_str(repv strm, int *c_p)
 	    else
 	    {
 		*cur++ = c;
-		c = rep_stream_getc(strm);
+		c = fast_getc(strm);
 	    }
 	}
 	if(c == EOF)
@@ -535,13 +610,6 @@ read_str(repv strm, int *c_p)
 repv
 rep_readl(repv strm, register int *c_p)
 {
-#ifdef MINSTACK
-    if(STK_SIZE <= MINSTACK)
-    {
-	STK_WARN("read");
-	return Fsignal(Qstack_error, Qnil);
-    }
-#endif
     while(1)
     {
 	switch(*c_p)
@@ -556,13 +624,13 @@ rep_readl(repv strm, register int *c_p)
 	case '\t':
 	case '\n':
 	case '\f':
-	    *c_p = rep_stream_getc(strm);
+	    *c_p = fast_getc(strm);
 	    continue;
 
 	case ';':
 	    {
 		register int c;
-		while((c = rep_stream_getc(strm)) != EOF && c != '\n' && c != '\f')
+		while((c = fast_getc(strm)) != EOF && c != '\n' && c != '\f')
 		    ;
 		*c_p = rep_stream_getc(strm);
 		continue;
@@ -643,8 +711,11 @@ rep_readl(repv strm, register int *c_p)
 	case '#':
 	    switch(*c_p = rep_stream_getc(strm))
 	    {
+		int comment_terminator;
+
 	    case EOF:
 		goto eof;
+
 	    case '\'':
 		form = Fcons(Qfunction, Fcons(Qnil, Qnil));
 		rep_PUSHGC(gc_form, form);
@@ -659,6 +730,7 @@ rep_readl(repv strm, register int *c_p)
 		    return rep_NULL;
 		else
 		    return form;
+
 	    case '[':
 		{
 		    repv vec = read_vector(strm, c_p);
@@ -675,6 +747,92 @@ rep_readl(repv strm, register int *c_p)
 		    }
 		    break;
 		}
+
+	    case '(':
+		return read_vector (strm, c_p);
+
+	    case '|':
+		/* comment delimited by `#| ... |#' */
+		comment_terminator = '|';
+		{
+		    register int c;
+		read_comment:
+		    while ((c = fast_getc (strm)) != EOF)
+		    {
+		    comment_again:
+			if (c == comment_terminator)
+			{
+			    c = rep_stream_getc (strm);
+			    if (c == EOF || c == '#')
+				break;
+			    else
+				goto comment_again;
+			}
+		    }
+		    if (c != EOF)
+			c = rep_stream_getc (strm);
+		    *c_p = c;
+		    continue;
+		}
+
+	    case '\\':
+		{
+		    static const struct {
+			char *name;
+			int value;
+		    } char_names[] = {
+		      { "space", ' ' },
+		      { "newline", '\n' },
+		      { "backspace", '\010' },
+		      { "tab", '\t' },
+		      { "linefeed", '\n' },
+		      { "return", '\r' },
+		      { "page", '\f' },
+		      { "rubout", '\177' },
+		      { 0, 0 }
+		    };
+
+		    int c = rep_stream_getc (strm), c2, i;
+
+		    if (c == EOF)
+			goto eof;
+		    if (!isalpha (c))
+		    {
+			*c_p = rep_stream_getc (strm);
+			return rep_MAKE_INT (c);
+		    }
+		    c2 = rep_stream_getc (strm);
+		    if (c2 == EOF)
+			goto eof;
+		    if (!isalpha (c2))
+		    {
+			*c_p = c2;
+			return rep_MAKE_INT (c);
+		    }
+		    c = tolower (c);
+		    c2 = tolower (c2);
+		    for (i = 0; char_names[i].name != 0; i++)
+		    {
+			if (char_names[i].name[0] == c
+			    && char_names[i].name[1] == c2)
+			{
+			    char *ptr = char_names[i].name + 2;
+			    while (1)
+			    {
+				c = fast_getc (strm);
+				if (*ptr == 0)
+				{
+				    *c_p = c;
+				    return rep_MAKE_INT (char_names[i].value);
+				}
+				if (c == EOF || tolower (c) != *ptr++)
+				    goto error;
+			    }
+			}
+		    }
+		    goto error;
+		}
+
 	    case '!':
 		if (rep_FILEP(strm))
 		{
@@ -682,23 +840,12 @@ rep_readl(repv strm, register int *c_p)
 		    if (pos && rep_INTP(pos) && rep_INT(pos) == 2)
 		    {
 			/* #! at the start of the file. Skip until !# */
-			register int c;
-			while ((c = rep_stream_getc (strm)) != EOF)
-			{
-			    if (c == '!')
-			    {
-				c = rep_stream_getc (strm);
-				if (c == EOF || c == '#')
-				    break;
-			    }
-			}
-			if (c != EOF)
-			    c = rep_stream_getc (strm);
-			*c_p = c;
-			continue;
+			comment_terminator = '!';
+			goto read_comment;
 		    }
 		}
-		/* FALL THROUGH */
+		/* fall through */
+
 	    default: error:
 		return Fsignal(Qinvalid_read_syntax, rep_LIST_1(strm));
 	    }
@@ -707,7 +854,7 @@ rep_readl(repv strm, register int *c_p)
 	    return read_symbol(strm, c_p);
 	}
     }
-    /* NOT REACHED */
+    /* not reached */
 
 eof:
     return Fsignal(Qend_of_stream, rep_LIST_1(strm));
@@ -765,7 +912,8 @@ eval_list(repv list)
    IMPORTANT: this expects the top of the call stack to have the
    saved environments in which arguments need to be evaluated */
 repv
-rep_bind_lambda_list(repv lambdaList, repv argList, rep_bool eval_args)
+rep_bind_lambda_list(repv lambdaList, repv argList,
+		     rep_bool eval_args, rep_bool eval_in_env)
 {
 #define STATE_REQUIRED 1
 #define STATE_OPTIONAL 2
@@ -775,6 +923,7 @@ rep_bind_lambda_list(repv lambdaList, repv argList, rep_bool eval_args)
     repv *evalled_args;
     int evalled_nargs;
     repv boundlist = Qnil;
+    int state;
 
     rep_GC_root gc_arglist, gc_boundlist;
     rep_GC_n_roots gc_evalled_args;
@@ -791,7 +940,8 @@ rep_bind_lambda_list(repv lambdaList, repv argList, rep_bool eval_args)
 	evalled_nargs = rep_list_length(argList);
 	evalled_args = alloca(sizeof(repv) * evalled_nargs);
 	rep_PUSHGCN(gc_evalled_args, evalled_args, 0);
-	rep_env = rep_call_stack->saved_env;
+	if (!eval_in_env)
+	    rep_env = rep_call_stack->saved_env;
 	rep_PUSHGC(gc_old_env, old_env);
 	for(i = 0; i < evalled_nargs; i++)
 	{
@@ -807,106 +957,110 @@ rep_bind_lambda_list(repv lambdaList, repv argList, rep_bool eval_args)
 	rep_POPGC;
     }
 
-    if(rep_CONSP(lambdaList))
+    state = STATE_REQUIRED;
+    while (1)
     {
-	int state = STATE_REQUIRED;
-	while(rep_CONSP(lambdaList) && rep_SYMBOLP(rep_CAR(lambdaList)))
+	repv argobj;
+	repv argspec;
+
+	if (rep_CONSP(lambdaList) && rep_SYMBOLP(rep_CAR(lambdaList)))
 	{
-	    repv argobj;
-	    repv argspec = rep_CAR(lambdaList);
+	    argspec = rep_CAR(lambdaList);
+	    lambdaList = rep_CDR(lambdaList);
+
 	    if(rep_STR(rep_SYM(argspec)->name)[0] == '&')
 	    {
 		if(argspec == Qamp_optional)
 		{
 		    if(state > STATE_OPTIONAL)
 		    {
-			Fsignal(Qinvalid_lambda_list,
-				   rep_LIST_1(lambdaList));
+			Fsignal(Qinvalid_lambda_list, rep_LIST_1(lambdaList));
 			goto error;
 		    }
 		    state = STATE_OPTIONAL;
-		    lambdaList = rep_CDR(lambdaList);
 		    continue;
 		}
 		else if(argspec == Qamp_rest)
 		{
 		    if(state > STATE_REST)
 		    {
-			Fsignal(Qinvalid_lambda_list,
-				   rep_LIST_1(lambdaList));
+			Fsignal(Qinvalid_lambda_list, rep_LIST_1(lambdaList));
 			goto error;
 		    }
 		    state = STATE_REST;
-		    lambdaList = rep_CDR(lambdaList);
 		    continue;
 		}
 		else if(argspec == Qamp_aux)
 		{
 		    state = STATE_AUX;
-		    lambdaList = rep_CDR(lambdaList);
 		    continue;
 		}
 	    }
+	}
+	else if (lambdaList != Qnil && rep_SYMBOLP(lambdaList))
+	{
+	    state = STATE_REST;
+	    argspec = lambdaList;
+	    lambdaList = Qnil;
+	}
+	else
+	    break;
 
-	    switch(state)
+	switch(state)
+	{
+	case STATE_REQUIRED:
+	case STATE_OPTIONAL:
+	    if(eval_args && evalled_nargs > 0)
 	    {
-	    case STATE_REQUIRED:
-	    case STATE_OPTIONAL:
-		if(eval_args && evalled_nargs > 0)
+		argobj = *evalled_args++;
+		evalled_nargs--;
+		gc_evalled_args.count--;
+	    }
+	    else if(!eval_args && rep_CONSP(argList))
+	    {
+		argobj = rep_CAR(argList);
+		argList = rep_CDR(argList);
+	    }
+	    else if(state == STATE_OPTIONAL)
+		argobj = Qnil;
+	    else
+	    {
+		Fsignal(Qmissing_arg, rep_LIST_1(argspec));
+		goto error;
+	    }
+	    boundlist = rep_bind_symbol(boundlist, argspec, argobj);
+	    break;
+
+	case STATE_REST:
+	    if(eval_args)
+	    {
+		repv list = Qnil;
+		repv *ptr = &list;
+		while(evalled_nargs > 0)
 		{
-		    argobj = *evalled_args++;
+		    *ptr = Fcons(*evalled_args++, Qnil);
+		    ptr = &(rep_CDR(*ptr));
 		    evalled_nargs--;
 		    gc_evalled_args.count--;
 		}
-		else if(!eval_args && rep_CONSP(argList))
-		{
-		    argobj = rep_CAR(argList);
-		    argList = rep_CDR(argList);
-		}
-		else if(state == STATE_OPTIONAL)
-		    argobj = Qnil;
-		else
-		{
-		    Fsignal(Qmissing_arg, rep_LIST_1(argspec));
-		    goto error;
-		}
-		boundlist = rep_bind_symbol(boundlist, argspec, argobj);
-		break;
-
-	    case STATE_REST:
-		if(eval_args)
-		{
-		    repv list = Qnil;
-		    repv *ptr = &list;
-		    while(evalled_nargs > 0)
-		    {
-			*ptr = Fcons(*evalled_args++, Qnil);
-			ptr = &(rep_CDR(*ptr));
-			evalled_nargs--;
-			gc_evalled_args.count--;
-		    }
-		    argobj = list;
-		}
-		else
-		    argobj = argList;
-		boundlist = rep_bind_symbol(boundlist, argspec, argobj);
-		state = STATE_AUX;
-		break;
-
-	    case STATE_AUX:
-		boundlist = rep_bind_symbol(boundlist, argspec, Qnil);
+		argobj = list;
 	    }
+	    else
+		argobj = Fcopy_sequence (argList);
+	    boundlist = rep_bind_symbol(boundlist, argspec, argobj);
+	    state = STATE_AUX;
+	    break;
 
-	    lambdaList = rep_CDR(lambdaList);
+	case STATE_AUX:
+	    boundlist = rep_bind_symbol(boundlist, argspec, Qnil);
+	}
 
-	    rep_TEST_INT;
-	    if(rep_INTERRUPTP)
-	    {
-	    error:
-		rep_unbind_symbols(boundlist);
-		boundlist = rep_NULL;
-		break;
-	    }
+	rep_TEST_INT;
+	if(rep_INTERRUPTP) {
+	error:
+	    rep_unbind_symbols(boundlist);
+	    boundlist = rep_NULL;
+	    break;
 	}
     }
 
@@ -918,7 +1072,8 @@ rep_bind_lambda_list(repv lambdaList, repv argList, rep_bool eval_args)
 }
 
 repv
-rep_eval_lambda(repv lambdaExp, repv argList, rep_bool eval_args)
+rep_eval_lambda(repv lambdaExp, repv argList,
+		rep_bool eval_args, rep_bool eval_in_env)
 {
     repv result = rep_NULL;
     if(rep_CONSP(rep_CDR(lambdaExp)))
@@ -928,7 +1083,8 @@ rep_eval_lambda(repv lambdaExp, repv argList, rep_bool eval_args)
 	rep_PUSHGC(gc_lambdaExp, lambdaExp);
 	rep_PUSHGC(gc_argList, argList);
 	lambdaExp = rep_CDR(lambdaExp);
-	boundlist = rep_bind_lambda_list(rep_CAR(lambdaExp), argList, eval_args);
+	boundlist = rep_bind_lambda_list(rep_CAR(lambdaExp), argList,
+					 eval_args, eval_in_env);
 	if(boundlist)
 	{
 	    rep_GC_root gc_boundlist;
@@ -973,7 +1129,7 @@ rep_load_autoload(repv funarg)
 	|| !rep_STRINGP(rep_CAR(rep_CDR(aload_def))))
     {
 	return Fsignal(Qinvalid_autoload,
-		       rep_list_2(fun, rep_VAL(&invl_autoload)));
+		       rep_list_2(aload_def, rep_VAL(&invl_autoload)));
     }
 
     fun = rep_CAR(aload_def);
@@ -988,8 +1144,8 @@ rep_load_autoload(repv funarg)
 	u_long old_msg_len;
 	repv tmp;
 
-	if (rep_SYM (Qbatch_mode)->value == Qnil
-	    && rep_SYM (Qautoload_verbose)->value != Qnil
+	if (Fsymbol_value (Qbatch_mode, Qt) == Qnil
+	    && Fsymbol_value (Qautoload_verbose, Qt) != Qnil
 	    && rep_message_fun != 0)
 	{
 	    (*rep_message_fun)(rep_save_message, &old_msg, &old_msg_len);
@@ -1007,8 +1163,8 @@ rep_load_autoload(repv funarg)
 	tmp = rep_call_lisp2 (load, file, Qt);
 	rep_POPGC; rep_POPGC;
 
-	if (rep_SYM (Qbatch_mode)->value == Qnil
-	    && rep_SYM (Qautoload_verbose)->value != Qnil
+	if (Fsymbol_value (Qbatch_mode, Qt) == Qnil
+	    && Fsymbol_value (Qautoload_verbose, Qt) != Qnil
 	    && rep_message_fun != 0)
 	{
 	    (*rep_message_fun)(rep_messagef,
@@ -1022,19 +1178,24 @@ rep_load_autoload(repv funarg)
 
 	fun = Fsymbol_value (fun, Qnil);
 
-	/* Magically replace one closure by another without
-	   losing eq-ness */
-	if (fun && rep_FUNARGP(fun))
+	if (fun != rep_NULL)
 	{
-	    rep_FUNARG(funarg)->fun = rep_FUNARG(fun)->fun;
-	    rep_FUNARG(funarg)->name = rep_FUNARG(fun)->name;
-	    rep_FUNARG(funarg)->env = rep_FUNARG(fun)->env;
-	    rep_FUNARG(funarg)->special_env = rep_FUNARG(fun)->special_env;
-	    rep_FUNARG(funarg)->fh_env = rep_FUNARG(fun)->fh_env;
+	    /* Magically replace one closure by another without
+	       losing eq-ness */
+	    tmp = fun;
+	    if (rep_CONSP(tmp) && rep_CAR(tmp) == Qmacro)
+		tmp = rep_CDR(tmp);
+	    if (rep_FUNARGP(tmp))
+	    {
+		rep_FUNARG(funarg)->fun = rep_FUNARG(tmp)->fun;
+		rep_FUNARG(funarg)->name = rep_FUNARG(tmp)->name;
+		rep_FUNARG(funarg)->env = rep_FUNARG(tmp)->env;
+		rep_FUNARG(funarg)->special_env = rep_FUNARG(tmp)->special_env;
+		rep_FUNARG(funarg)->fh_env = rep_FUNARG(tmp)->fh_env;
+	    }
+	    else
+		rep_FUNARG(funarg)->fun = Qnil;
 	}
-	else
-	    rep_FUNARG(funarg)->fun = Qnil;
-
 	return fun;
     }
     else
@@ -1065,28 +1226,25 @@ rep_funcall(repv fun, repv arglist, rep_bool eval_args)
     int type;
     repv result = rep_NULL;
     struct rep_Call lc;
-    rep_bool was_closed = rep_FALSE;
+    repv closure = rep_NULL;
+    rep_GC_root gc_fun, gc_args, gc_closure;
 
     rep_TEST_INT;
     if(rep_INTERRUPTP)
 	return rep_NULL;
 
-#ifdef MINSTACK
-    if(STK_SIZE <= MINSTACK)
+    if(++rep_lisp_depth > rep_max_lisp_depth)
     {
-	STK_WARN("rep_funcall");
-	return Fsignal(Qstack_error, Qnil);
-    }
-#endif
-
-    if(++lisp_depth > max_lisp_depth)
-    {
-	lisp_depth--;
+	rep_lisp_depth--;
 	return Fsignal(Qerror, rep_LIST_1(rep_VAL(&max_depth)));
     }
 
-    /* Note that by putting FUN and ARGLIST in the backtrace,
-       they're automagically protected from gc. */
+    rep_PUSHGC (gc_fun, fun);
+    rep_PUSHGC (gc_args, arglist);
+    rep_PUSHGC (gc_closure, closure);
+
+    rep_MAY_YIELD;
+
     lc.fun = fun;
     lc.args = arglist;
     lc.args_evalled_p = eval_args ? Qnil : Qt;
@@ -1098,9 +1256,8 @@ rep_funcall(repv fun, repv arglist, rep_bool eval_args)
 again:
     if (rep_FUNARGP(fun))
     {
-	rep_USE_FUNARG(fun);
+	closure = fun;
 	fun = rep_FUNARG(fun)->fun;
-	was_closed = rep_TRUE;
     }
     switch(type = rep_TYPE(fun))
     {
@@ -1115,10 +1272,14 @@ again:
 	    if(arglist == rep_NULL)
 		goto end;
 	}
+	if (closure)
+	    rep_USE_FUNARG(closure);
 	result = rep_SUBRNFUN(fun)(arglist);
 	break;
 
     case rep_Subr0:
+	if (closure)
+	    rep_USE_FUNARG(closure);
 	result = rep_SUBR0FUN(fun)();
 	break;
 
@@ -1172,6 +1333,8 @@ again:
 	}
 	if(eval_args)
 	    rep_POPGCN;
+	if (closure)
+	    rep_USE_FUNARG(closure);
 	switch(type)
 	{
 	case rep_Subr1:
@@ -1196,15 +1359,16 @@ again:
     case rep_Cons:
 	car = rep_CAR(fun);
 	/* don't allow unclosed lambdas for security reasons */
-	if(was_closed && car == Qlambda)
+	if(closure && car == Qlambda)
 	{
 	    /* rep_eval_lambda() expanded inline. */
 	    if(rep_CONSP(rep_CDR(fun)))
 	    {
 		repv boundlist;
 		fun = rep_CDR(fun);
-		boundlist = rep_bind_lambda_list(rep_CAR(fun),
-						 arglist, eval_args);
+		rep_USE_FUNARG(closure);
+		boundlist = rep_bind_lambda_list(rep_CAR(fun), arglist,
+						 eval_args, rep_FALSE);
 		if(boundlist != rep_NULL)
 		{
 		    rep_GC_root gc_boundlist;
@@ -1221,17 +1385,15 @@ again:
 	{
 	    /* A macro. This could occur if autoloading from
 	       a macro definition. Try to accommodate.. */
-	    if(!eval_args)
-		/* Args already evaluated. Can't expand the macro */
-		goto invalid;
-	    result = Fmacroexpand(Fcons(lc.fun, arglist), Qnil);
-	    if(result != rep_NULL)
-		result = Feval(result);
+	    if(eval_args)
+		goto invalid;		/* can't expand from evaluated args */
+	    fun = lc.fun = rep_CDR(fun);
+	    goto again;
 	}
-	else if(was_closed && car == Qautoload)
+	else if(closure && car == Qautoload)
 	{
-	    /* lc.fun contains the original closure */
-	    fun = rep_load_autoload(lc.fun);
+	    rep_USE_FUNARG(closure);
+	    fun = rep_load_autoload(closure);
 	    if(fun)
 	    {
 		lc.fun = fun;
@@ -1244,24 +1406,22 @@ again:
 
     case rep_Compiled:
 	/* don't allow unclosed bytecode for security reasons */
-	if (was_closed)
+	if (closure)
 	{
 	    repv boundlist;
-
-	    if (search_environment (Qjade_byte_code) == Qnil)
+	    if (rep_bytecode_interpreter == 0)
 		goto invalid;
 
+	    rep_USE_FUNARG(closure);
 	    boundlist = rep_bind_lambda_list(rep_COMPILED_LAMBDA(fun),
-					     arglist, eval_args);
+					     arglist, eval_args, rep_FALSE);
 	    if(boundlist != rep_NULL)
 	    {
-		rep_GC_root gc_boundlist;
-		rep_PUSHGC(gc_boundlist, boundlist);
-		result = Fjade_byte_code(rep_COMPILED_CODE(fun),
-					 rep_COMPILED_CONSTANTS(fun),
-					 rep_MAKE_INT(rep_COMPILED_STACK(fun)));
-		rep_POPGC;
-		rep_unbind_symbols(boundlist);
+		result = (rep_bytecode_interpreter
+			  (rep_COMPILED_CODE(fun),
+			   rep_COMPILED_CONSTANTS(fun),
+			   rep_MAKE_INT(rep_COMPILED_STACK(fun)),
+			   boundlist));
 	    }
 	    else
 		result = rep_NULL;
@@ -1279,7 +1439,8 @@ again:
 
 end:
     rep_POP_CALL(lc);
-    lisp_depth--;
+    rep_POPGC; rep_POPGC; rep_POPGC;
+    rep_lisp_depth--;
     return result;
 }
 
@@ -1301,14 +1462,6 @@ eval(repv obj)
 {
     repv result = rep_NULL;
 
-#ifdef MINSTACK
-    if(STK_SIZE <= MINSTACK)
-    {
-	STK_WARN("eval");
-	return Fsignal(Qstack_error, Qnil);
-    }
-#endif
-
     switch(rep_TYPE(obj))
     {
 	repv funcobj;
@@ -1318,6 +1471,13 @@ eval(repv obj)
 	break;
 
     case rep_Cons:
+	if (rep_CONSP (rep_CAR (obj)) && rep_CAAR (obj) == Qlambda)
+	{
+	    /* inline lambda; don't need to enclose it.. */
+	    result = rep_eval_lambda (rep_CAR (obj), rep_CDR (obj),
+				      rep_TRUE, rep_TRUE);
+	    break;
+	}
 	funcobj = Feval (rep_CAR(obj));
 	if(funcobj == rep_NULL)
 	    goto end;
@@ -1346,11 +1506,6 @@ eval(repv obj)
 	    result = rep_funcall(funcobj, rep_CDR(obj), rep_TRUE);
 	break;
 
-    case rep_Var:
-	if(!(result = rep_VARFUN(obj)(rep_NULL)))
-	    Fsignal(Qvoid_value, rep_LIST_1(obj));
-	break;
-
     default:
 	result = obj;
 	break;
@@ -1395,7 +1550,8 @@ Evaluates FORM and returns its value.
 
     {
 	repv dbres;
-	repv dbargs = Fcons(obj, Fcons(rep_MAKE_INT(DbDepth), Qnil));
+	repv dbargs = rep_list_3(obj, rep_MAKE_INT(DbDepth),
+				 rep_box_pointer (rep_call_stack));
 	if(dbargs)
 	{
 	    rep_GC_root gc_dbargs;
@@ -1403,8 +1559,9 @@ Evaluates FORM and returns its value.
 	    rep_PUSHGC(gc_dbargs, dbargs);
 	    rep_push_regexp_data(&re_data);
 	    rep_single_step_flag = rep_FALSE;
-	    dbres = rep_funcall(Fsymbol_value (Qdebug_entry, Qt),
-				dbargs, rep_FALSE);
+	    dbres = (rep_call_with_barrier
+		     (Ffuncall, Fcons (Fsymbol_value (Qdebug_entry, Qt),
+				       dbargs), rep_TRUE, 0, 0, 0));
 	    rep_pop_regexp_data();
 	    if (dbres != rep_NULL && rep_CONSP(dbres))
 	    {
@@ -1436,8 +1593,9 @@ Evaluates FORM and returns its value.
 		{
 		    rep_push_regexp_data(&re_data);
 		    rep_CAR(dbargs) = result;
-		    dbres = rep_funcall(Fsymbol_value (Qdebug_exit, Qt),
-					dbargs, rep_FALSE);
+		    dbres = (rep_call_with_barrier
+			     (Ffuncall, Fcons (Fsymbol_value (Qdebug_exit, Qt),
+					       dbargs), rep_TRUE, 0, 0, 0));
 		    if(!dbres)
 			result = rep_NULL;
 		    rep_pop_regexp_data();
@@ -1518,16 +1676,6 @@ rep_lisp_prin(repv strm, repv obj)
 	int print_length;
 	repv tem;
 
-    case rep_Int:
-#ifdef HAVE_SNPRINTF
-	snprintf(tbuf, sizeof(tbuf),
-		 "%" rep_PTR_SIZED_INT_CONV "d", rep_INT(obj));
-#else
-	sprintf(tbuf, "%" rep_PTR_SIZED_INT_CONV "d", rep_INT(obj));
-#endif
-	rep_stream_puts(strm, tbuf, -1, rep_FALSE);
-	break;
-
     case rep_Cons:
 	tem = Fsymbol_value(Qprint_level, Qt);
 	if(tem && rep_INTP(tem) && print_level >= rep_INT(tem))
@@ -1615,28 +1763,16 @@ rep_lisp_prin(repv strm, repv obj)
 	rep_stream_puts(strm, tbuf, -1, rep_FALSE);
 	break;
 
-    case rep_Var:
-#ifdef HAVE_SNPRINTF
-	snprintf(tbuf, sizeof(tbuf), "#<var %s>", rep_STR(rep_XSUBR(obj)->name));
-#else
-	sprintf(tbuf, "#<var %s>", rep_STR(rep_XSUBR(obj)->name));
-#endif
-	rep_stream_puts(strm, tbuf, -1, rep_FALSE);
-	break;
-
     case rep_Funarg:
 	if (rep_STRINGP(rep_FUNARG(obj)->name))
 	{
-#ifdef HAVE_SNPRINTF
-	    snprintf(tbuf, sizeof(tbuf), "#<closure %s>", rep_STR(rep_FUNARG(obj)->name));
-#else
-	    sprintf(tbuf, "#<closure %s>", rep_STR(rep_FUNARG(obj)->name));
-#endif
+	    rep_stream_puts (strm, "#<closure ", -1, rep_FALSE);
+	    rep_stream_puts (strm, rep_STR(rep_FUNARG(obj)->name),
+			     -1, rep_FALSE);
+	    rep_stream_putc (strm, '>');
 	}
 	else
-	    strcpy (tbuf, "#<closure>");
-
-	rep_stream_puts(strm, tbuf, -1, rep_FALSE);
+	    rep_stream_puts(strm, "#<closure>", -1, rep_FALSE);
 	break;
 
     case rep_Void:
@@ -1750,14 +1886,10 @@ rep_copy_list(repv list)
 repv
 rep_handle_var_int(repv val, int *data)
 {
-    if(val)
-    {
-	if(rep_INTP(val))
-	    *data = rep_INT(val);
-	return rep_NULL;
-    }
-    else
-	return rep_MAKE_INT(*data);
+    int old = *data;
+    if(rep_INTP(val))
+	*data = rep_INT(val);
+    return rep_MAKE_INT (old);
 }
 
 /* Similar, but for variables containing greater than 24 bits of data,
@@ -1765,14 +1897,10 @@ rep_handle_var_int(repv val, int *data)
 repv
 rep_handle_var_long_int(repv val, long *data)
 {
-    if(val)
-    {
-	if(rep_LONG_INTP(val))
-	    *data = rep_LONG_INT(val);
-	return rep_NULL;
-    }
-    else
-	return rep_MAKE_LONG_INT(*data);
+    long old = *data;
+    if(rep_LONG_INTP(val))
+	*data = rep_LONG_INT(val);
+    return rep_MAKE_LONG_INT(old);
 }
 
 DEFUN("break", Fbreak, Sbreak, (void), rep_Subr0) /*
@@ -1801,51 +1929,6 @@ Use the Lisp debugger to evaluate FORM.
     return res;
 }
 
-DEFUN("macroexpand", Fmacroexpand, Smacroexpand, (repv form, repv env), rep_Subr2) /*
-::doc:macroexpand::
-macroexpand FORM [ENVIRONMENT]
-
-If FORM is a macro call, expand it until it isn't. If ENVIRONMENT is
-specified it is an alist of `(MACRO-NAME . DEFINITION)'.
-::end:: */
-{
-    repv car;
-    rep_GC_root gc_form, gc_env, gc_car;
-    rep_PUSHGC(gc_form, form);
-    rep_PUSHGC(gc_env, env);
-    rep_PUSHGC(gc_car, car);
-top:
-    if(rep_CONSP(form))
-    {
-	car = rep_CAR(form);
-	if(rep_SYMBOLP(car))
-	{
-	    repv tmp;
-	    if(env != Qnil && (tmp = Fassq(car, env)) && rep_CONSP(tmp))
-		car = rep_CDR(tmp);
-	    else
-	    {
-		car = Fsymbol_value (car, Qt);
-		if (!rep_CONSP(car) || rep_CAR(car) != Qmacro)
-		    goto end;
-		car = rep_CDR(car);
-	    }
-	}
-	else if (rep_CONSP(car) && rep_CAR(car) == Qmacro)
-	    car = rep_CDR(car);
-
-	if (Ffunctionp(car) == Qnil)
-	    goto end;
-
-	form = rep_funcall (car, rep_CDR(form), rep_FALSE);
-	if (form != rep_NULL)
-	    goto top;
-    }
-end:
-    rep_POPGC; rep_POPGC; rep_POPGC;
-    return form;
-}
-
 DEFUN("signal", Fsignal, Ssignal, (repv error, repv data), rep_Subr2) /*
 ::doc:signal::
 signal ERROR-SYMBOL DATA
@@ -1886,8 +1969,11 @@ handler.
 	Fset(Qdebug_on_error, Qnil);
 	rep_single_step_flag = rep_FALSE;
 	rep_PUSHGC(gc_on_error, on_error);
-	tmp = rep_funcall(Fsymbol_value (Qdebug_error_entry, Qt),
-			  Fcons(errlist, Qnil), rep_FALSE);
+	tmp = (rep_call_with_barrier
+	       (Ffuncall, Fcons (Fsymbol_value (Qdebug_error_entry, Qt),
+				 rep_list_2(errlist,
+					    rep_box_pointer (rep_call_stack))),
+		rep_TRUE, 0, 0, 0));
 	rep_POPGC;
 	Fset(Qdebug_on_error, on_error);
 	if(tmp && (tmp == Qt))
@@ -1993,12 +2079,24 @@ DEFSTRING(one_err_fmt, "%s");
 DEFSTRING(two_err_fmt, "%s: %s");
 DEFSTRING(three_err_fmt, "%s: %s, %s");
 DEFSTRING(four_err_fmt, "%s: %s, %s, %s");
-DEFSTRING(n_err_fmt, "%s: ...");
+DEFSTRING(n_err_fmt, "%s: %s");
 
 void
 rep_handle_error(repv error, repv data)
 {
+    static int mutex;
     repv errstr;
+
+    if (mutex++ == 0)
+    {
+	repv fun = Fsymbol_value (Qerror_handler_function, Qt);
+	if (Ffunctionp (fun) != Qnil)
+	{
+	    rep_call_lisp2 (fun, error, data);
+	    goto out;
+	}
+    }
+
     Fbeep();
     if(!(errstr = Fget(error, Qerror_message)) || !rep_STRINGP(errstr))
 	errstr = rep_VAL(&unknown_err);
@@ -2021,8 +2119,11 @@ rep_handle_error(repv error, repv data)
 					  rep_CAR(rep_CDR(rep_CDR(data))))));
 	break;
     default:
-	Fformat(rep_list_3(Qt, rep_VAL(&n_err_fmt), errstr));
+	Fformat(rep_list_4(Qt, rep_VAL(&n_err_fmt), errstr, data));
     }
+
+out:
+    mutex--;
 }
 
 repv
@@ -2075,24 +2176,40 @@ ARGLIST had been evaluated or not before being put into the stack.
     return Qt;
 }
 
-DEFUN("max-lisp-depth", Vmax_lisp_depth, Smax_lisp_depth, (repv val), rep_Var) /*
+DEFUN("debug-frame-environment", Fdebug_frame_environment,
+      Sdebug_frame_environment, (repv pointer), rep_Subr1)
+{
+    struct rep_Call *fp = rep_call_stack;
+    struct rep_Call *ptr = rep_unbox_pointer (pointer);
+    rep_DECLARE(1, pointer, ptr != 0);
+    while (fp != 0 && fp != ptr)
+	fp = fp->next;
+    if (fp != 0)
+    {
+	return rep_list_3 (fp->saved_env,
+			   fp->saved_special_env,
+			   fp->saved_fh_env);
+    }
+    else
+	return Qnil;
+}
+
+DEFUN("max-lisp-depth", Fmax_lisp_depth, Smax_lisp_depth, (repv val), rep_Subr1) /*
 ::doc:max-lisp-depth::
+max-lisp-depth [NEW-VALUE]
+
 The maximum number of times that rep_funcall can be called recursively.
 
 This is intended to stop infinite recursion, if the default value of 250 is
 too small (you get errors in normal use) set it to something larger.
 ::end:: */
 {
-    return rep_handle_var_int(val, &max_lisp_depth);
+    return rep_handle_var_int(val, &rep_max_lisp_depth);
 }
 
 void
 rep_lisp_init(void)
 {
-    rep_USE_DEFAULT_ENV;
-    rep_mark_static (&rep_env);
-    rep_mark_static (&rep_special_env);
-
     rep_INTERN(quote); rep_INTERN(lambda); rep_INTERN(macro);
     rep_INTERN(backquote); rep_INTERN(backquote_unquote); rep_INTERN(backquote_splice);
     rep_INTERN(autoload); rep_INTERN(function);
@@ -2105,11 +2222,11 @@ rep_lisp_init(void)
     rep_ADD_SUBR(Sprogn);
     rep_ADD_SUBR(Sbreak);
     rep_ADD_SUBR_INT(Sstep);
-    rep_ADD_SUBR(Smacroexpand);
     rep_ADD_SUBR(Ssignal);
     rep_ADD_SUBR(Scondition_case);
     rep_ADD_SUBR(Sbacktrace);
     rep_ADD_SUBR(Smax_lisp_depth);
+    rep_ADD_SUBR(Sdebug_frame_environment);
 
     /* Stuff for error-handling */
     rep_INTERN(error_message);
@@ -2128,20 +2245,18 @@ rep_lisp_init(void)
     rep_INTERN(invalid_stream); rep_ERROR(invalid_stream);
     rep_INTERN(setting_constant); rep_ERROR(setting_constant);
     rep_INTERN(process_error); rep_ERROR(process_error);
-#ifdef MINSTACK
-    rep_INTERN(stack_error); rep_ERROR(stack_error);
-#endif
     rep_INTERN(no_memory); rep_ERROR(no_memory);
     rep_INTERN(user_interrupt); rep_ERROR(user_interrupt);
     rep_INTERN(arith_error); rep_ERROR(arith_error);
     rep_INTERN(term_interrupt);
 
     rep_INTERN_SPECIAL(debug_on_error);
-    rep_SYM(Qdebug_on_error)->value = Qnil;
+    Fset (Qdebug_on_error, Qnil);
     rep_INTERN_SPECIAL(backtrace_on_error);
-    rep_SYM(Qbacktrace_on_error)->value = Qnil;
+    Fset (Qbacktrace_on_error, Qnil);
     rep_INTERN_SPECIAL(debug_macros);
-    rep_SYM(Qdebug_macros)->value = Qnil;
+    Fset (Qdebug_macros, Qnil);
+    rep_INTERN_SPECIAL(error_handler_function);
 
     rep_int_cell = Fcons(Quser_interrupt, Qnil);
     rep_mark_static(&rep_int_cell);
@@ -2151,9 +2266,9 @@ rep_lisp_init(void)
     rep_INTERN_SPECIAL(print_escape); 
     rep_INTERN_SPECIAL(print_length);
     rep_INTERN_SPECIAL(print_level);
-    rep_SYM(Qprint_escape)->value = Qnil;
-    rep_SYM(Qprint_length)->value = Qnil;
-    rep_SYM(Qprint_level)->value = Qnil;
+    Fset (Qprint_escape, Qnil);
+    Fset (Qprint_length, Qnil);
+    Fset (Qprint_level, Qnil);
     rep_INTERN(newlines);
 
     rep_INTERN_SPECIAL(autoload_verbose);
