@@ -1,6 +1,6 @@
 /* lispcmds.c -- Lots of standard Lisp functions
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: lispcmds.c,v 1.53 1999/11/29 18:37:55 john Exp $
+   $Id: lispcmds.c,v 1.73 2000/03/28 23:51:50 john Exp $
 
    This file is part of Jade.
 
@@ -31,7 +31,7 @@
 #endif
 
 DEFSTRING(default_rep_directory, REP_DIRECTORY);
-DEFSTRING(div_zero, "Divide by zero");
+DEFSTRING(dot, ".");
 
 DEFSYM(or, "or");
 DEFSYM(and, "and");
@@ -39,6 +39,7 @@ DEFSYM(load_path, "load-path");
 DEFSYM(dl_load_path, "dl-load-path");
 DEFSYM(after_load_alist, "after-load-alist");
 DEFSYM(provide, "provide");
+DEFSYM(features, "features");
 DEFSYM(rep_directory, "rep-directory");
 DEFSYM(lisp_lib_directory, "lisp-lib-directory");
 DEFSYM(site_lisp_directory, "site-lisp-directory");
@@ -100,19 +101,6 @@ Returns ARG.
     return rep_signal_missing_arg(1);
 }
 
-DEFUN("function", Ffunction, Sfunction, (repv args), rep_SF) /*
-::doc:function::
-function ARG
-#'ARG
-
-Returns the closure of ARG.
-::end:: */
-{
-    if(rep_CONSP(args))
-	return Fmake_closure (rep_CAR(args), Qnil);
-    return rep_signal_missing_arg(1);
-}
-
 DEFUN("lambda", Flambda, Slambda, (repv args), rep_SF) /*
 ::doc:lambda::
 lambda LAMBDA-LIST BODY...
@@ -126,109 +114,6 @@ Evaluates to an anonymous function.
 	return rep_signal_missing_arg(1);
 }
 
-DEFUN("defmacro", Fdefmacro, Sdefmacro, (repv args), rep_SF) /*
-::doc:defmacro::
-defmacro NAME LAMBDA-LIST [DOC-STRING] BODY...
-defmacro NAME BYTECODE-OBJECT
-
-Defines a macro called NAME with argument spec. LAMBDA-LIST, documentation
-DOC-STRING (optional) and body BODY. The actual function value is
-    `(macro lambda LAMBDA-LIST [DOC-STRING] BODY...)'
-Macros are called with their arguments un-evaluated, they are expected to
-return a form which will be executed to provide the result of the expression.
-
-A pathetic example could be,
-  (defmacro foo (x) (list 'cons nil x))
-   => foo
-  (foo 'bar)
-   => (nil . bar)
-This makes `(foo X)' a pseudonym for `(cons nil X)'.
-
-Note that macros are expanded at *compile-time* (unless, of course, the Lisp
-code has not been compiled).
-::end:: */
-{
-    repv name;
-    if(!rep_CONSP(args))
-	return rep_signal_missing_arg(1);
-    name = rep_CAR(args);
-    args = rep_CDR(args);
-    if(!rep_CONSP(args))
-	return rep_signal_missing_arg(2);
-    if(!rep_COMPILEDP(rep_CAR(args)))
-	args = Fcons(Qlambda, args);
-    else
-	args = rep_CAR(args);
-    if (Fset(name, Fcons (Qmacro, Fmake_closure (args, rep_SYM(name)->name))))
-    {
-	rep_SYM(name)->car |= rep_SF_DEFVAR;
-	return name;
-    }
-    else
-	return rep_NULL;
-}
-
-DEFUN("defun", Fdefun, Sdefun, (repv args), rep_SF) /*
-::doc:defun::
-defun NAME LAMBDA-LIST [DOC-STRING] BODY...
-defun NAME BYTECODE-OBJECT
-
-Defines a function called NAME with argument specification LAMBDA-LIST,
-documentation DOC-STRING (optional) and body BODY. The actual function
-value is,
-    `(lambda LAMBDA-LIST [DOC-STRING] BODY...)'
-::end:: */
-{
-    repv name;
-    if(!rep_CONSP(args))
-	return rep_signal_missing_arg(1);
-    name = rep_CAR(args);
-    args = rep_CDR(args);
-    if(!rep_CONSP(args))
-	return rep_signal_missing_arg(2);
-    if(!rep_COMPILEDP(rep_CAR(args)))
-	args = Fcons(Qlambda, args);
-    else
-	args = rep_CAR(args);
-    if (Fset(name, Fmake_closure (args, rep_SYM(name)->name)))
-    {
-	rep_SYM(name)->car |= rep_SF_DEFVAR;
-	return name;
-    }
-    else
-	return rep_NULL;
-}
-
-DEFSTRING(const_bound, "Constant already bound");
-
-DEFUN("defconst", Fdefconst, Sdefconst, (repv args), rep_SF) /*
-::doc:defconst::
-defconst NAME repv [DOC-STRING]
-
-Define a constant NAME whose (default) value is repv. If NAME is already
-bound an error is signalled.
-
-Constants are treated specially by the Lisp compiler, basically they are
-hard-coded into the byte-code. For more details see the comments in
-the compiler source (`lisp/compiler.jl').
-::end:: */
-{
-    if(rep_CONSP(args))
-    {
-	repv tmp = Fdefault_boundp(rep_CAR(args));
-	if(tmp && !rep_NILP(tmp))
-	{
-	    return(Fsignal(Qerror, rep_list_2(rep_VAL(&const_bound),
-						rep_CAR(args))));
-	}
-	tmp = Fdefvar(args);
-	if(tmp)
-	    return(Fset_const_variable(tmp, Qnil));
-	return(tmp);
-    }
-    return rep_signal_missing_arg(1);
-}
-
 DEFUN("car", Fcar, Scar, (repv cons), rep_Subr1) /*
 ::doc:car::
 car CONS-CELL
@@ -1292,37 +1177,6 @@ First evals FORM1 then FORMS, returns the value that FORM1 gave.
     return rep_signal_missing_arg(1);
 }
 
-DEFUN("prog2", Fprog2, Sprog2, (repv args), rep_SF) /*
-::doc:prog2::
-prog2 FORM1 FORM2 FORMS...
-
-Evals FORM1 then FORM2 then the rest. Returns whatever FORM2 gave.
-::end:: */
-{
-    if(rep_CONSP(args) && rep_CONSP(rep_CDR(args)))
-    {
-	repv res;
-	rep_GC_root gc_args, gc_res;
-	rep_PUSHGC(gc_args, args);
-	if(Feval(rep_CAR(args)))
-	{
-	    res = Feval(rep_CAR(rep_CDR(args)));
-	    if(res)
-	    {
-		rep_PUSHGC(gc_res, res);
-		if(!Fprogn(rep_CDR(rep_CDR(args))))
-		    res = rep_NULL;
-		rep_POPGC;
-	    }
-	}
-	else
-	    res = rep_NULL;
-	rep_POPGC;
-	return(res);
-    }
-    return rep_signal_missing_arg(rep_CONSP(args) ? 2 : 1);
-}
-
 DEFUN("while", Fwhile, Swhile, (repv args), rep_SF) /*
 ::doc:while::
 while CONDITION FORMS...
@@ -1397,82 +1251,59 @@ like the last else in an else-if statement in C.
     return(res);
 }
 
-DEFUN("if", Fif, Sif, (repv args), rep_SF) /*
-::doc:if::
-if CONDITION THEN-FORM [ELSE-FORMS...]
+DEFUN("case", Fcase, Scase, (repv args), rep_SF) /*
+::doc:case::
+case KEY CLAUSES...
 
-Evaluate CONDITION, if it is non-nil then evaluate THEN-FORM and return
-it's result. Otherwise do an implicit progn with the ELSE-FORMS returning
-its value.
+Each CLAUSE is `((ITEMS... ) FORMS...)'. Find the first CLAUSE with an
+ITEM matching (using `eql') the result of evaluating KEY (only
+evaluated once), then evaluate the associated FORMS in a `progn'. The
+final clause may have the form `(t FORMS...)', which always matches KEY
+if no other CLAUSE has already. Returns `nil' if no clause matches.
+
+If any of the ITEMS appear more than once, then the behaviour is
+undefined.
 ::end:: */
 {
+    repv key;
     rep_GC_root gc_args;
-    repv res;
-    if(!rep_CONSP(args))
-	return rep_signal_missing_arg(1);
+    int i = 2;
+
+    if (!rep_CONSP(args))
+	return rep_signal_missing_arg (0);
+    
     rep_PUSHGC(gc_args, args);
-    res = Feval(rep_CAR(args));
+    key = Feval (rep_CAR(args));
     rep_POPGC;
+    if (key == rep_NULL)
+	return rep_NULL;
     args = rep_CDR(args);
-    if(res != rep_NULL && rep_CONSP(args))
+
+    while(rep_CONSP(args))
     {
-	if(!rep_NILP(res))
-	    res = Feval(rep_CAR(args));
+	if (!rep_CONSP(rep_CAR(args)))
+	    return rep_signal_arg_error (rep_CAR(args), i);
+	if (rep_CONSP(rep_CAAR(args)))
+	{
+	    repv ptr = rep_CAAR(args);
+	    while (rep_CONSP(ptr))
+	    {
+		if (rep_CAR(ptr) == key)
+		    return Fprogn (rep_CDAR(args));
+		ptr = rep_CDR(ptr);
+		rep_TEST_INT;
+		if (rep_INTERRUPTP)
+		    return rep_NULL;
+	    }
+	}
+	else if (rep_CAAR(args) == Qt)
+	    return Fprogn (rep_CDAR(args));
 	else
-	    res = Fprogn(rep_CDR(args));
-    }
-    return res;
-}
-
-DEFUN("and", Fand, Sand, (repv args), rep_SF) /*
-::doc:and::
-and FORMS...
-
-Evaluates each member of FORMS in turn, until one returns nil, and returns
-nil. If none return nil then the value of the last form evaluated is
-returned.
-::end:: */
-{
-    repv res = Qt;
-    rep_GC_root gc_args, gc_res;
-    rep_PUSHGC(gc_args, args);
-    rep_PUSHGC(gc_res, res);
-    while(res && rep_CONSP(args) && !rep_NILP(res))
-    {
-	res = Feval(rep_CAR(args));
+	    return rep_signal_arg_error (rep_CAR(args), i);
 	args = rep_CDR(args);
-	rep_TEST_INT;
-	if(rep_INTERRUPTP)
-	    res = rep_NULL;
+	i++;
     }
-    rep_POPGC;
-    rep_POPGC;
-    return res;
-}
-
-DEFUN("or", For, Sor, (repv args), rep_SF) /*
-::doc:or::
-or FORMS...
-
-Evaluates each member of FORMS in turn until one returns t, the result of
-which is returned. If none are t then return nil.
-::end:: */
-{
-    repv res = Qnil;
-    rep_GC_root gc_args, gc_res;
-    rep_PUSHGC(gc_args, args);
-    rep_PUSHGC(gc_res, res);
-    while(res && rep_CONSP(args) && rep_NILP(res))
-    {
-	res = Feval(rep_CAR(args));
-	args = rep_CDR(args);
-	rep_TEST_INT;
-	if(rep_INTERRUPTP)
-	    res = rep_NULL;
-    }
-    rep_POPGC;
-    rep_POPGC;
-    return res;
+    return Qnil;
 }
 
 DEFUN("apply", Fapply, Sapply, (repv args), rep_SubrN) /*
@@ -1547,6 +1378,7 @@ loaded and a warning is displayed.
     rep_bool no_error_p = !rep_NILP(noerr_p);
     rep_bool no_suffix_p = !rep_NILP(nosuf_p);
     rep_bool in_current_env = !rep_NILP(in_env);
+    rep_bool interp_mode = Fsymbol_value (Qinterpreted_mode, Qt) != Qnil;
 
     repv name = Qnil, path;
     repv dir = rep_NULL, try = rep_NULL;
@@ -1577,65 +1409,68 @@ loaded and a warning is displayed.
 research:
     while(rep_NILP(name) && rep_CONSP(path))
     {
-	dir = Ffile_name_as_directory(rep_STRINGP(rep_CAR(path))
-					 ? rep_CAR(path)
-					 : rep_null_string());
-	if(dir == rep_NULL || !rep_STRINGP(dir))
-	    goto path_error;
-
-	if(!no_suffix_p)
+	if (rep_STRINGP (rep_CAR(path)))
 	{
-	    repv tem;
-	    int i;
-	    static char *suffixes[3] = { ".jl", ".jlc", ".jld" };
-#ifdef rep_DUMPED
-	    i = 2;
-#else
-	    i = 1;
-#endif
-#ifdef HAVE_DYNAMIC_LOADING
-	    if(trying_dl)
-		i = 0;
-#endif
-	    if (rep_SYM(Qinterpreted_mode)->value != Qnil)
-		i = 0;
-	    for(; i >= 0; i--)
+	    dir = Fexpand_file_name (file, rep_CAR(path));
+	    if(dir == rep_NULL || !rep_STRINGP(dir))
+		goto path_error;
+
+	    if(!no_suffix_p)
 	    {
-#ifdef HAVE_DYNAMIC_LOADING
-		if (trying_dl)
-		    try = rep_concat4(rep_STR(dir), "lib", rep_STR(file), ".la");
-		else
-#endif
-		    try = rep_concat3(rep_STR(dir), rep_STR(file), suffixes[i]);
-		tem = load_file_exists_p (try);
-		if(!tem)
-		    goto path_error;
-		if(!rep_NILP(tem))
+		repv tem;
+		static char *suffixes[3] = { ".jl", ".jlc" };
+		int i = 1;
+		if (!trying_dl && interp_mode)
+		    i = 0;
+		for(; i >= 0; i--)
 		{
-		    if(!rep_NILP(name))
+#ifdef HAVE_DYNAMIC_LOADING
+		    if (trying_dl)
 		    {
-			if(rep_file_newer_than(try, name))
+			if (i == 1)
+			    try = rep_concat2(rep_STR(dir), ".la");
+			else
 			{
-			    if (rep_message_fun != 0)
-				(*rep_message_fun)(rep_messagef,"Warning: %s newer than %s, using %s", rep_STR(try), rep_STR(name), rep_STR(try));
-			    name = try;
+			    try = (Fexpand_file_name
+				   (rep_concat3 ("lib", rep_STR(file), ".la"),
+				    rep_CAR(path)));
 			}
 		    }
 		    else
-			name = try;
+#endif
+			try = rep_concat2(rep_STR(dir), suffixes[i]);
+
+		    if (try == rep_NULL || !rep_STRINGP (try))
+			goto path_error;
+
+		    tem = load_file_exists_p (try);
+		    if(!tem)
+			goto path_error;
+		    if(tem != Qnil)
+		    {
+			if(name != Qnil)
+			{
+			    if(rep_file_newer_than(try, name))
+			    {
+				if (rep_message_fun != 0)
+				    (*rep_message_fun)(rep_messagef,"Warning: %s newer than %s, using %s", rep_STR(try), rep_STR(name), rep_STR(try));
+				name = try;
+			    }
+			}
+			else
+			    name = try;
+		    }
 		}
 	    }
-	}
-	if(rep_NILP(name))
-	{
-	    /* Try without a suffix */
-	    repv tem;
-	    try = rep_concat2(rep_STR(dir), rep_STR(file));
-	    tem = load_file_exists_p (try);
-	    if(!tem)
-		goto path_error;
-	    if(!rep_NILP(tem))
-		name = try;
+	    if(name == Qnil)
+	    {
+		/* Try without a suffix */
+		repv tem = load_file_exists_p (dir);
+		if(!tem)
+		    goto path_error;
+		if(tem != Qnil)
+		    name = dir;
+	    }
 	}
 	path = rep_CDR(path);
 	rep_TEST_INT;
@@ -1672,7 +1507,14 @@ path_error:
 
 #ifdef HAVE_DYNAMIC_LOADING
     if(trying_dl)
-	return rep_open_dl_library(name) ? Qt : rep_NULL;
+    {
+	void *handle;
+	rep_PUSHGC(gc_file, file);
+	handle = rep_open_dl_library(name);
+	rep_POPGC;
+	if (handle == 0)
+	    return rep_NULL;
+    }
     else
 #endif
     {
@@ -1731,10 +1573,15 @@ path_error:
 	rep_POPGC; rep_POPGC;
 	rep_unbind_symbols (bindings);
 	Fclose_file (stream);
+	rep_POPGC;
+    }
 
-	/* Loading succeeded. Look for an applicable item in
-	   the after-load-alist. */
-    again:
+    /* Loading succeeded. Look for an applicable item in
+       the after-load-alist. */
+    rep_PUSHGC (gc_file, file);
+    {
+	repv tem;
+again:
 	tem = Fsymbol_value(Qafter_load_alist, Qt);
 	if(tem != rep_NULL && rep_CONSP(tem))
 	{
@@ -1743,9 +1590,8 @@ path_error:
 	    {
 		/* Delete this entry */
 		Fset(Qafter_load_alist,
-			Fdelq(tem, Fsymbol_value
-				 (Qafter_load_alist, Qt)));
-
+		     Fdelq(tem, Fsymbol_value (Qafter_load_alist, Qt)));
+	    
 		/* Then evaluate it */
 		Fprogn(rep_CDR(tem));
 
@@ -1753,208 +1599,10 @@ path_error:
 		goto again;
 	    }
 	}
-	rep_POPGC;
-
-	return Qt;
     }
-}
+    rep_POPGC;
 
-/*
- * some arithmetic commands
- */
-
-#define APPLY_OP(op)					\
-    if(rep_CONSP(args) && rep_INTP(rep_CAR(args)))			\
-    {							\
-	long sum = rep_INT(rep_CAR(args));			\
-	int i = 2;					\
-	args = rep_CDR(args);				\
-	while(rep_CONSP(args))				\
-	{						\
-	    if(!rep_INTP(rep_CAR(args)))			\
-		return rep_signal_arg_error(rep_CAR(args), i);	\
-	    sum = sum op rep_INT(rep_CAR(args));		\
-	    args = rep_CDR(args);				\
-	    i++;					\
-	    rep_TEST_INT;					\
-	    if(rep_INTERRUPTP)					\
-		return(rep_NULL);			\
-	}						\
-	return(rep_MAKE_INT(sum));				\
-    }							\
-    return (rep_CONSP(args)					\
-	    ? rep_signal_arg_error(rep_CAR(args), 1)		\
-	    : rep_signal_missing_arg(1));
-
-DEFUN("+", Fplus, Splus, (repv args), rep_SubrN) /*
-::doc:+::
-+ NUMBERS...
-
-Adds all NUMBERS together. If no arguments are given returns 0.
-::end:: */
-{
-    if(rep_NILP(args))
-	return rep_MAKE_INT(0);
-    APPLY_OP( + )
-}
-
-DEFUN("-", Fminus, Sminus, (repv args), rep_SubrN) /*
-::doc:-::
-- NUMBER [NUMBERS...]
-
-Either returns the negation of NUMBER or the value of NUMBER minus
-NUMBERS
-::end:: */
-{
-    if(rep_CONSP(args))
-    {
-	if(!rep_CONSP(rep_CDR(args)))
-	    return(rep_MAKE_INT(-rep_INT(rep_CAR(args))));
-	else
-	    APPLY_OP( - )
-    }
-    return rep_signal_missing_arg(1);
-}
-
-DEFUN("*", Fproduct, Sproduct, (repv args), rep_SubrN) /*
-::doc:*::
-* NUMBERS...
-
-Multiplies all NUMBERS together. If no numbers are given returns 1.
-::end:: */
-{
-    if(rep_NILP(args))
-	return rep_MAKE_INT(1);
-    APPLY_OP( * )
-}
-
-DEFUN("/", Fdivide, Sdivide, (repv args), rep_SubrN) /*
-::doc:/::
-/ NUMBERS...
-
-Divides NUMBERS (in left-to-right order), ie,
-  (/ 100 2
-   => 10
-::end:: */
-{
-    if(rep_CONSP(args) && rep_INTP(rep_CAR(args)))
-    {
-	long sum = rep_INT(rep_CAR(args));
-	int i = 1;
-	args = rep_CDR(args);
-	while(rep_CONSP(args))
-	{
-	    if(!rep_INTP(rep_CAR(args)))
-	       return rep_signal_arg_error(rep_CAR(args), i);
-	    if(rep_INT(rep_CAR(args)) == 0)
-		return Fsignal(Qarith_error, rep_LIST_1(rep_VAL(&div_zero)));
-	    sum = sum / rep_INT(rep_CAR(args));
-	    args = rep_CDR(args);
-	    i++;
-	    rep_TEST_INT;
-	    if(rep_INTERRUPTP)
-		return(rep_NULL);
-	}
-	return(rep_MAKE_INT(sum));
-    }
-    return (rep_CONSP(args)
-	    ? rep_signal_arg_error(rep_CAR(args), 1) : rep_signal_missing_arg(1));
-}
-
-DEFUN("%", Fremainder, Sremainder, (repv n1, repv n2), rep_Subr2) /*
-::doc:%::
-% DIVIDEND DIVISOR
-
-Returns the integer remainder after dividing DIVIDEND by DIVISOR.
-::end:: */
-{
-    rep_DECLARE1(n1, rep_INTP);
-    rep_DECLARE2(n2, rep_INTP);
-    if(rep_INT(n2) == 0)
-	return Fsignal(Qarith_error, rep_LIST_1(rep_VAL(&div_zero)));
-    return rep_MAKE_INT(rep_INT(n1) % rep_INT(n2));
-}
-
-DEFUN("mod", Fmod, Smod, (repv n1, repv n2), rep_Subr2) /*
-::doc:mod::
-mod DIVIDEND DIVISOR
-
-Returns the value of DIVIDEND modulo DIVISOR; unlike the % (remainder)
-function the behaviour of `mod' is well-defined for negative arguments,
-we have that,
-
-	(mod X Y) == X - (* Y (floor (/ X Y))),	for Y not equal to zero
-
-assuming that (floor Z) gives the least integer greater than or equal to Z,
-and that floating point division is used.
-::end:: */
-{
-    long tem;
-    rep_DECLARE1(n1, rep_INTP);
-    rep_DECLARE2(n2, rep_INTP);
-    if(rep_INT(n2) == 0)
-	return Fsignal(Qarith_error, rep_LIST_1(rep_VAL(&div_zero)));
-
-    /* This code from GNU Emacs */
-    tem = rep_INT(n1) % rep_INT(n2);
-    /* If the "remainder" comes out with the wrong sign, fix it.  */
-    if (rep_INT(n2) < 0 ? tem > 0 : tem < 0)
-	tem += rep_INT(n2);
-
-    return rep_MAKE_INT(tem);
-}
-
-DEFUN("lognot", Flognot, Slognot, (repv num), rep_Subr1) /*
-::doc:lognot::
-lognot NUMBER
-
-Returns the bitwise logical `not' of NUMBER.
-::end:: */
-{
-    rep_DECLARE1(num, rep_INTP);
-    return(rep_MAKE_INT(~rep_INT(num)));
-}
-
-DEFUN("not", Fnot, Snot, (repv arg), rep_Subr1) /*
-::doc:not::
-not ARG
-
-If ARG is nil returns t, else returns nil.
-::end:: */
-{
-    if(rep_NILP(arg))
-	return(Qt);
-    return(Qnil);
-}
-
-DEFUN("logior", Flogior, Slogior, (repv args), rep_SubrN) /*
-::doc:logior::
-logior NUMBERS...
-
-Returns the bitwise logical `inclusive-or' of its arguments.
-::end:: */
-{
-    APPLY_OP( | )
-}
-
-DEFUN("logxor", Flogxor, Slogxor, (repv args), rep_SubrN) /*
-::doc:logxor::
-logxor NUMBERS...
-
-Returns the bitwise logical `exclusive-or' of its arguments.
-::end:: */
-{
-    APPLY_OP( ^ )
-}
-
-DEFUN("logand", Flogand, Slogand, (repv args), rep_SubrN) /*
-::doc:logand::
-logand NUMBERS...
-
-Returns the bitwise logical `and' of its arguments.
-::end:: */
-{
-    APPLY_OP( & )
+    return Qt;
 }
 
 DEFUN("equal", Fequal, Sequal, (repv val1, repv val2), rep_Subr2) /*
@@ -1981,18 +1629,16 @@ this may or may not be true for numbers of the same value (see `eql').
     return (val1 == val2) ? Qt : Qnil;
 }
 
-DEFUN("eql", Feql, Seql, (repv arg1, repv arg2), rep_Subr2) /*
-::doc:eql::
-eql ARG1 ARG2
-Similar to `eq' except that numbers (integers, characters) with the same
-value will always be considered `eql' (this may or may not be the case
-with `eq'.
+DEFUN("not", Fnot, Snot, (repv arg), rep_Subr1) /*
+::doc:not::
+not ARG
+
+If ARG is nil returns t, else returns nil.
 ::end:: */
 {
-    if(rep_INTP(arg1) && rep_INTP(arg2))
-	return(rep_INT(arg1) == rep_INT(arg2) ? Qt : Qnil);
-    else
-	return(arg1 == arg2 ? Qt : Qnil);
+    if(rep_NILP(arg))
+	return(Qt);
+    return(Qnil);
 }
 
 DEFUN("string-head-eq", Fstring_head_eq, Sstring_head_eq, (repv str1, repv str2), rep_Subr2) /*
@@ -2064,54 +1710,52 @@ Returns t if STRING1 is `less' than STRING2, ignoring case.
 }
 
 
-DEFUN("=", Fnum_eq, Snum_eq, (repv num1, repv num2), rep_Subr2) /*
-::doc:=::
-= NUMBER1 NUMBER2
-
-Returns t if NUMBER1 and NUMBER2 are equal.
-::end:: */
-{
-    rep_DECLARE1(num1, rep_INTP);
-    rep_DECLARE2(num2, rep_INTP);
-    if(rep_INT(num1) == rep_INT(num2))
-	return(Qt);
-    return(Qnil);
-}
-
-DEFUN("/=", Fnum_noteq, Snum_noteq, (repv num1, repv num2), rep_Subr2) /*
-::doc:/=::
-/= NUMBER1 NUMBER2
-
-Returns t if NUMBER1 and NUMBER2 are unequal.
-::end:: */
-{
-    rep_DECLARE1(num1, rep_INTP);
-    rep_DECLARE2(num2, rep_INTP);
-    if(rep_INT(num1) != rep_INT(num2))
-	return(Qt);
-    return(Qnil);
-}
-
 #define APPLY_COMPARISON(op)				\
-    if(rep_CONSP(args) && rep_CONSP(rep_CDR(args)))		\
+    if(rep_CONSP(args) && rep_CONSP(rep_CDR(args)))	\
     {							\
 	repv pred = rep_CAR(args);			\
 	int i = 2;					\
 	args = rep_CDR(args);				\
 	while(rep_CONSP(args))				\
 	{						\
-	    if(!(rep_value_cmp(pred, rep_CAR(args)) op 0))	\
+	    if(!(rep_value_cmp(pred, rep_CAR(args)) op 0)) \
 		return Qnil;				\
-	    pred = rep_CAR(args);				\
-	    args = rep_CDR(args);				\
+	    pred = rep_CAR(args);			\
+	    args = rep_CDR(args);			\
 	    i++;					\
-	    rep_TEST_INT;					\
-	    if(rep_INTERRUPTP)					\
+	    rep_TEST_INT;				\
+	    if(rep_INTERRUPTP)				\
 		return(rep_NULL);			\
 	}						\
 	return Qt;					\
     }							\
-    return rep_signal_missing_arg(rep_CONSP(args) ? 2 : 1);					\
+    return rep_signal_missing_arg(rep_CONSP(args) ? 2 : 1);
+
+DEFUN("=", Fnum_eq, Snum_eq, (repv args), rep_SubrN) /*
+::doc:=::
+= ARG1 ARG2 [ARG3 ...]
+
+Returns t if each value is the same as every other value. (Using
+`equal' to compare values.)
+::end:: */
+{
+    APPLY_COMPARISON(==)
+}
+
+DEFUN("/=", Fnum_noteq, Snum_noteq, (repv args), rep_SubrN) /*
+::doc::/=::
+/= ARG1 ARG2 ...
+
+Returns t if each value is different from every other value. (Using
+`equal' to compare values.)
+::end:: */
+{
+    repv ret = Fnum_eq (args);
+    if (ret)
+	return ret == Qnil ? Qt : Qnil;
+    else
+	return rep_NULL;
+}
 
 DEFUN(">", Fgtthan, Sgtthan, (repv args), rep_SubrN) /*
 ::doc:>::
@@ -2208,71 +1852,6 @@ arguments.
     return min;
 }
 
-DEFUN("1+", Fplus1, Splus1, (repv num), rep_Subr1) /*
-::doc:1+::
-1+ NUMBER
-
-Return NUMBER plus 1.
-::end:: */
-{
-    rep_DECLARE1(num, rep_INTP);
-    return(rep_MAKE_INT(rep_INT(num) + 1));
-}
-
-DEFUN("1-", Fsub1, Ssub1, (repv num), rep_Subr1) /*
-::doc:1-::
-1- NUMBER
-
-Return NUMBER minus 1.
-::end:: */
-{
-    rep_DECLARE1(num, rep_INTP);
-    return(rep_MAKE_INT(rep_INT(num) - 1));
-}
-
-DEFUN("lsh", Flsh, Slsh, (repv num, repv shift), rep_Subr2) /*
-::doc:lsh::
-lsh NUMBER COUNT
-
-Shift the bits in NUMBER by COUNT bits to the left, a negative COUNT means
-shift right.
-::end:: */
-{
-    rep_DECLARE1(num, rep_INTP);
-    rep_DECLARE2(shift, rep_INTP);
-    if(rep_INT(shift) > 0)
-	return(rep_MAKE_INT(rep_INT(num) << rep_INT(shift)));
-    /* ensure that a zero is in the top bit. */
-    return(rep_MAKE_INT((rep_INT(num) >> -rep_INT(shift)) & 0x7fffffff));
-}
-
-DEFUN("ash", Fash, Sash, (repv num, repv shift), rep_Subr2) /*
-::doc:ash::
-ash NUMBER COUNT
-
-Use an arithmetic shift to shift the bits in NUMBER by COUNT bits to the left,
-a negative COUNT means shift right.
-::end:: */
-{
-    rep_DECLARE1(num, rep_INTP);
-    rep_DECLARE2(shift, rep_INTP);
-    if(rep_INT(shift) > 0)
-	return(rep_MAKE_INT(rep_INT(num) << rep_INT(shift)));
-    return(rep_MAKE_INT(rep_INT(num) >> -rep_INT(shift)));
-}
-
-DEFUN("zerop", Fzerop, Szerop, (repv num), rep_Subr1) /*
-::doc:zerop::
-zerop NUMBER
-
-t if NUMBER is zero.
-::end:: */
-{
-    if(rep_INTP(num) && (rep_INT(num) == 0))
-	return(Qt);
-    return(Qnil);
-}
-
 DEFUN("null", Fnull, Snull, (repv arg), rep_Subr1) /*
 ::doc:null::
 null ARG
@@ -2313,26 +1892,6 @@ Returns t if ARG is a list, (either a cons-cell or nil).
     return rep_LISTP(arg) ? Qt : Qnil;
 }
 
-DEFUN("numberp", Fnumberp, Snumberp, (repv arg), rep_Subr1) /*
-::doc:numberp::
-numberp ARG
-
-Return t if ARG is a number.
-::end:: */
-{
-    return rep_INTP(arg) ? Qt : Qnil;
-}
-
-DEFUN("integerp", Fintegerp, Sintegerp, (repv arg), rep_Subr1) /*
-::doc:integerp::
-integerp ARG
-
-Return t if ARG is a integer.
-::end:: */
-{
-    return rep_INTP(arg) ? Qt : Qnil;
-}
-
 DEFUN("stringp", Fstringp, Sstringp, (repv arg), rep_Subr1) /*
 ::doc:stringp::
 stringp ARG
@@ -2435,7 +1994,6 @@ Returns t if arg is a primitive function.
     case rep_Subr5:
     case rep_SubrN:
     case rep_SF:
-    case rep_Var:
 	return(Qt);
     default:
 	return(Qnil);
@@ -2472,7 +2030,6 @@ Returns the name (a string) associated with SUBR.
     case rep_Subr5:
     case rep_SubrN:
     case rep_SF:
-    case rep_Var:
 	return(rep_SUBR(subr)->name);
     default:
 	return(Qnil);
@@ -2593,9 +2150,12 @@ DEFUN("unwind-protect", Funwind_protect, Sunwind_protect, (repv args), rep_SF) /
 ::doc:unwind-protect::
 unwind-protect BODY CLEANUP-FORMS...
 
-Eval and return the value of BODY guaranteeing that the CLEANUP-FORMS will
-be evalled no matter what happens (ie, error, non-local exit, etc) while
-BODY is being evaluated.
+Return the result of evaluating BODY. When execution leaves the dynamic
+extent of BODY evaluate `(progn CLEANUP-FORMS)' (even if exiting due to
+an exception within BODY).
+
+Note that when BODY is exited by calling a continuation, it is
+undefined whether or not CLEANUP-FORMS will be evaluated.
 ::end:: */
 {
     if(rep_CONSP(args))
@@ -2659,15 +2219,69 @@ of whatever was changed. Return the value of the last FORM evaluated.
     return res;
 }
 
+DEFUN("featurep", Ffeaturep, Sfeaturep, (repv feature), rep_Subr1) /*
+::doc:featurep::
+featurep FEATURE
+
+Return non-nil if feature FEATURE has already been loaded.
+::end:: */
+{
+    repv value;
+    rep_DECLARE1 (feature, rep_SYMBOLP);
+    value = Fsymbol_value (Qfeatures, Qnil);
+    if (value != rep_NULL)
+	return Fmemq (feature, value);
+    else
+	return rep_NULL;
+}
+
+DEFUN("provide", Fprovide, Sprovide, (repv feature), rep_Subr1) /*
+::doc:provide::
+provide FEATURE
+
+Show that the feature FEATURE (a symbol) has been loaded.
+::end:: */
+{
+    repv value;
+    rep_DECLARE1 (feature, rep_SYMBOLP);
+    value = Fsymbol_value (Qfeatures, Qnil);
+    if (value != rep_NULL)
+    {
+	repv tem = Fmemq (feature, value);
+	if (tem && tem == Qnil)
+	    Fset (Qfeatures, Fcons (feature, value));
+	return feature;
+    }
+    else
+	return rep_NULL;
+}
+
+DEFUN_INT("require", Frequire, Srequire, (repv feature, repv file), rep_Subr2,
+	  "SFeature to load:") /*
+::doc:require::
+require FEATURE [FILE]
+
+If FEATURE (a symbol) has not already been loaded, load it. The file
+loaded is either FILE (if given), or the print name of FEATURE.
+::end:: */
+{
+    repv tem;
+    rep_DECLARE1 (feature, rep_SYMBOLP);
+    tem = Ffeaturep (feature);
+    if (tem && tem == Qnil)
+    {
+	if (!rep_STRINGP(file))
+	    file = rep_SYM(feature)->name;
+	return Fload (file, Qnil, Qnil, Qnil, Qnil);
+    }
+    return feature;
+}
+
 void
 rep_lispcmds_init(void)
 {
     rep_ADD_SUBR(Squote);
-    rep_ADD_SUBR(Sfunction);
     rep_ADD_SUBR(Slambda);
-    rep_ADD_SUBR(Sdefmacro);
-    rep_ADD_SUBR(Sdefun);
-    rep_ADD_SUBR(Sdefconst);
     rep_ADD_SUBR(Scar);
     rep_ADD_SUBR(Scdr);
     rep_ADD_SUBR(Slist);
@@ -2707,28 +2321,14 @@ rep_lispcmds_init(void)
     rep_ADD_SUBR(Scopy_sequence);
     rep_ADD_SUBR(Selt);
     rep_ADD_SUBR(Sprog1);
-    rep_ADD_SUBR(Sprog2);
     rep_ADD_SUBR(Swhile);
     rep_ADD_SUBR(Scond);
-    rep_ADD_SUBR(Sif);
-    rep_ADD_SUBR(Sand);
-    rep_ADD_SUBR(Sor);
+    rep_ADD_SUBR(Scase);
     rep_ADD_SUBR(Sapply);
     rep_ADD_SUBR_INT(Sload);
-    rep_ADD_SUBR(Splus);
-    rep_ADD_SUBR(Sminus);
-    rep_ADD_SUBR(Sproduct);
-    rep_ADD_SUBR(Sdivide);
-    rep_ADD_SUBR(Sremainder);
-    rep_ADD_SUBR(Smod);
-    rep_ADD_SUBR(Slognot);
     rep_ADD_SUBR(Snot);
-    rep_ADD_SUBR(Slogior);
-    rep_ADD_SUBR(Slogxor);
-    rep_ADD_SUBR(Slogand);
     rep_ADD_SUBR(Sequal);
     rep_ADD_SUBR(Seq);
-    rep_ADD_SUBR(Seql);
     rep_ADD_SUBR(Sstring_head_eq);
     rep_ADD_SUBR(Sstring_equal);
     rep_ADD_SUBR(Sstring_lessp);
@@ -2740,17 +2340,10 @@ rep_lispcmds_init(void)
     rep_ADD_SUBR(Slethan);
     rep_ADD_SUBR(Smax);
     rep_ADD_SUBR(Smin);
-    rep_ADD_SUBR(Splus1);
-    rep_ADD_SUBR(Ssub1);
-    rep_ADD_SUBR(Slsh);
-    rep_ADD_SUBR(Sash);
-    rep_ADD_SUBR(Szerop);
     rep_ADD_SUBR(Snull);
     rep_ADD_SUBR(Satom);
     rep_ADD_SUBR(Sconsp);
     rep_ADD_SUBR(Slistp);
-    rep_ADD_SUBR(Snumberp);
-    rep_ADD_SUBR(Sintegerp);
     rep_ADD_SUBR(Sstringp);
     rep_ADD_SUBR(Svectorp);
     rep_ADD_SUBR(Sbytecodep);
@@ -2765,75 +2358,81 @@ rep_lispcmds_init(void)
     rep_ADD_SUBR(Sthrow);
     rep_ADD_SUBR(Sunwind_protect);
     rep_ADD_SUBR(Swith_object);
+    rep_ADD_SUBR(Sfeaturep);
+    rep_ADD_SUBR(Sprovide);
+    rep_ADD_SUBR_INT(Srequire);
 
     rep_INTERN(provide);
+    rep_INTERN_SPECIAL(features);
+    if (Fboundp (Qfeatures) == Qnil)
+	Fset (Qfeatures, Qnil);
 
     rep_INTERN_SPECIAL(rep_directory);
     if(getenv("REPDIR") != 0)
-	rep_SYM(Qrep_directory)->value = rep_string_dup(getenv("REPDIR"));
+	Fset (Qrep_directory, rep_string_dup(getenv("REPDIR")));
     else
-	rep_SYM(Qrep_directory)->value = rep_VAL(&default_rep_directory);
+	Fset (Qrep_directory, rep_VAL(&default_rep_directory));
 
     rep_INTERN_SPECIAL(lisp_lib_directory);
     if(getenv("REPLISPDIR") != 0)
-	rep_SYM(Qlisp_lib_directory)->value = rep_string_dup(getenv("REPLISPDIR"));
+	Fset (Qlisp_lib_directory, rep_string_dup(getenv("REPLISPDIR")));
     else
-	rep_SYM(Qlisp_lib_directory)->value = rep_string_dup(REP_LISP_DIRECTORY);
+	Fset (Qlisp_lib_directory, rep_string_dup(REP_LISP_DIRECTORY));
 
     rep_INTERN_SPECIAL(site_lisp_directory);
     if(getenv("REPSITELISPDIR") != 0)
-	rep_SYM(Qsite_lisp_directory)->value
-	    = rep_string_dup(getenv("REPSITELISPDIR"));
+	Fset(Qsite_lisp_directory, rep_string_dup(getenv("REPSITELISPDIR")));
     else
-	rep_SYM(Qsite_lisp_directory)->value
-	    = rep_concat2(rep_STR(rep_SYM(Qrep_directory)->value), "/site-lisp");
+	Fset (Qsite_lisp_directory,
+	      rep_concat2(rep_STR(Fsymbol_value (Qrep_directory, Qt)),
+			  "/site-lisp"));
 
     rep_INTERN_SPECIAL(exec_directory);
     if(getenv("REPEXECDIR") != 0)
-	rep_SYM(Qexec_directory)->value = rep_string_dup(getenv("REPEXECDIR"));
+	Fset (Qexec_directory, rep_string_dup(getenv("REPEXECDIR")));
     else
-	rep_SYM(Qexec_directory)->value = rep_string_dup(REP_EXEC_DIRECTORY);
+	Fset (Qexec_directory, rep_string_dup(REP_EXEC_DIRECTORY));
 
     rep_INTERN_SPECIAL(documentation_file);
     if(getenv("REPDOCFILE") != 0)
-	rep_SYM(Qdocumentation_file)->value = rep_string_dup(getenv("REPDOCFILE"));
+	Fset (Qdocumentation_file, rep_string_dup(getenv("REPDOCFILE")));
     else
-	rep_SYM(Qdocumentation_file)->value = rep_string_dup(REP_DOC_FILE);
+	Fset (Qdocumentation_file, rep_string_dup(REP_DOC_FILE));
 
     rep_INTERN_SPECIAL(documentation_files);
-    rep_SYM(Qdocumentation_files)->value
-	= Fcons (rep_SYM(Qdocumentation_file)->value, Qnil);
+    Fset (Qdocumentation_files,
+	  Fcons (Fsymbol_value (Qdocumentation_file, Qt), Qnil));
 
     rep_INTERN_SPECIAL(load_path);
-    rep_SYM(Qload_path)->value = rep_list_3(rep_SYM(Qlisp_lib_directory)->value,
-					rep_SYM(Qsite_lisp_directory)->value,
-					rep_null_string ());
+    Fset (Qload_path, rep_list_3(Fsymbol_value (Qlisp_lib_directory, Qt),
+				 Fsymbol_value (Qsite_lisp_directory, Qt),
+				 rep_VAL(&dot)));
 
     rep_INTERN_SPECIAL(dl_load_path);
-    rep_SYM(Qdl_load_path)->value = Qnil;
+    Fset (Qdl_load_path, Qnil);
     {
 	char *ptr = getenv("LD_LIBRARY_PATH");
 	while (ptr != 0 && *ptr != 0)
 	{
 	    char *end = strchr(ptr, ':');
-	    rep_SYM(Qdl_load_path)->value
-		= Fcons(end ? rep_string_dupn(ptr, end - ptr)
-			: rep_string_dup(ptr), rep_SYM(Qdl_load_path)->value);
+	    Fset (Qdl_load_path,
+		  Fcons(end ? rep_string_dupn(ptr, end - ptr)
+			: rep_string_dup(ptr), Fsymbol_value (Qdl_load_path, Qt)));
 	    ptr = end ? end + 1 : 0;
 	}
     }
-    rep_SYM(Qdl_load_path)->value
-	= Fcons(rep_SYM(Qexec_directory)->value,
-		Fcons (rep_string_dup (REP_COMMON_EXEC_DIRECTORY),
-		       Fnreverse(rep_SYM(Qdl_load_path)->value)));
+    Fset (Qdl_load_path,
+	  Fcons (Fsymbol_value (Qexec_directory, Qt),
+		 Fcons (rep_string_dup (REP_COMMON_EXEC_DIRECTORY),
+			Fnreverse(Fsymbol_value (Qdl_load_path, Qt)))));
 
     rep_INTERN_SPECIAL(after_load_alist);
-    rep_SYM(Qafter_load_alist)->value = Qnil;
+    Fset (Qafter_load_alist, Qnil);
 
     rep_INTERN(or); rep_INTERN(and);
 
     rep_INTERN_SPECIAL(dl_load_reloc_now);
-    rep_SYM(Qdl_load_reloc_now)->value = Qnil;
+    Fset (Qdl_load_reloc_now, Qnil);
 
     rep_INTERN_SPECIAL(load_filename);
 }
