@@ -1,6 +1,6 @@
 /* lispmach.c -- Interpreter for compiled Lisp forms
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: lispmach.c,v 1.42 1999/12/11 11:39:16 john Exp $
+   $Id: lispmach.c,v 1.63 2000/04/06 21:47:48 john Exp $
 
    This file is part of Jade.
 
@@ -58,22 +58,30 @@ static u_long byte_code_usage[256];
 #endif
 
 
-/* Bindings */
+/* Helper functions
+
+   Note the careful use of inlining.. the icache is crucial, we want
+   the VM to be as small as possible, so that as much other code as
+   possible fits in cache as well. However, if a helper function is
+   only called once (or maybe is in a crucial path), then inline it..
+
+   The speedup from this (_not_ inlining everything) is _huge_ */
+
 
 /* Unbind one level of the BIND-STACK and return the new head of the stack.
    Each item in the BIND-STACK may be one of:
 	(t . FORM)
 		unwind-protect FORM to always evaluate
-	((SYM . OLD-VAL) ...)
-		list of symbol bindings to undo with rep_unbind_symbols()
-        INTEGER
-		unbind this many function bindings
+	(INTEGER . INTEGER)
+		car special and cdr lexical bindings
 	FUNARG
 		instantiate this closure
-	(PC . STACK-DEPTH)
-		not unbound here; install exception handler at PC */
-inline void
-rep_unbind_object(repv item)
+	(error . (PC . STACK-DEPTH))
+		not unbound here; install exception handler at PC
+
+   returns the number of dynamic bindings removed */
+static inline int
+inline_unbind_object(repv item)
 {
     if(rep_CONSP(item))
     {
@@ -81,31 +89,45 @@ rep_unbind_object(repv item)
 	{
 	    /* unwind-protect protection forms. */
 	    Feval(rep_CDR(item));
+	    return 1;
 	}
-	else if(rep_SYMBOLP(rep_CAR(item)) || rep_CONSP(rep_CAR(item)))
+	else if(rep_INTP(rep_CAR(item)))
 	{
 	    /* A set of symbol bindings (let or let*). */
-	    rep_unbind_symbols(item);
+	    return rep_unbind_symbols(item);
 	}
+	else if (rep_CAR(item) == Qerror)
+	    return 0;
 	else
 	{
 	    rep_type *t = rep_get_data_type(rep_TYPE(rep_CAR(item)));
 	    if (t->unbind != 0)
 		t->unbind(item);
+	    return 1;
 	}
     }
     else if (rep_FUNARGP(item))
+    {
 	rep_USE_FUNARG(item);
+	return 1;
+    }
     else
     {
 	rep_type *t = rep_get_data_type(rep_TYPE(item));
 	if (t->unbind != 0)
 	    t->unbind(item);
+	return 1;
     }
 }
 
+int
+rep_unbind_object (repv item)
+{
+    return inline_unbind_object (item);
+}
+
 /* Bind one object, returning the handle to later unbind by. */
-inline repv
+repv
 rep_bind_object(repv obj)
 {
     rep_type *t = rep_get_data_type(rep_TYPE(obj));
@@ -115,17 +137,53 @@ rep_bind_object(repv obj)
 	return Qnil;
 }
 
-/* copied from symbols.c
-
-   Returns (SYM . VALUE) if a lexical binding. Returns t if the actual
-   value is in the symbol's function slot */
-static inline repv
-search_environment (repv sym)
+static inline void
+unbind_all (repv stack)
 {
-    register repv env = rep_env;
-    while (rep_CONSP(env) && rep_CAR(rep_CAR(env)) != sym)
+    while (rep_CONSP (stack))
+    {
+	rep_unbind_object (rep_CAR (stack));
+	stack = rep_CDR (stack);
+    }
+}
+
+static repv
+unbind_all_but_one (repv stack)
+{
+    if (!rep_CONSP(stack))
+	return Qnil;
+    while (rep_CONSP(rep_CDR(stack)))
+    {
+	rep_unbind_object (rep_CAR (stack));
+	stack = rep_CDR (stack);
+    }
+    return stack;
+}
+
+/* Walk COUNT entries down the environment */
+static inline repv
+snap_environment (int count)
+{
+    register repv ptr = rep_env;
+    while (count-- > 0)
+	ptr = rep_CDR(ptr);
+    return rep_CAR(ptr);
+}
+
+/* The number of special variables bound by FRAME */
+static inline int
+bound_specials (repv frame)
+{
+    return rep_INT (rep_CAR (frame));
+}
+
+static repv
+search_special_bindings (repv sym)
+{
+    register repv env = rep_special_bindings;
+    while (env != Qnil && rep_CAAR(env) != sym)
 	env = rep_CDR(env);
-    return rep_CONSP(env) ? rep_CAR(env) : env;
+    return env != Qnil ? rep_CAR(env) : env;
 }
 
 
@@ -179,9 +237,12 @@ list_ref (repv list, int elt)
 	    arg = insn - op;						\
 	rep_CONCAT(op_, op):
 
-DEFUN("jade-byte-code", Fjade_byte_code, Sjade_byte_code, (repv code, repv consts, repv stkreq), rep_Subr3) /*
+DEFSTRING(max_depth, "max-lisp-depth exceeded, possible infinite recursion?");
+
+DEFUN("jade-byte-code", Fjade_byte_code, Sjade_byte_code,
+      (repv code, repv consts, repv stkreq, repv frame), rep_Subr4) /*
 ::doc:jade-byte-code::
-jade-byte-code CODE-STRING CONST-VEC MAX-STACK
+jade-byte-code CODE-STRING CONST-VEC MAX-STACK [FRAME]
 
 Evaluates the string of byte codes CODE-STRING, the constants that it
 references are contained in the vector CONST-VEC. MAX-STACK is a number
@@ -193,32 +254,65 @@ of byte code. See the functions `compile-file', `compile-directory' and
 `compile-lisp-lib' for more details.
 ::end:: */
 {
-    repv *stackbase;
-    register repv *stackp;
     /* This holds a list of sets of bindings, it can also hold the form of
        an unwind-protect that always gets eval'd (when the car is t).  */
-    repv bindstack = Qnil;
+    repv bindstack;
     register u_char *pc;
     rep_GC_root gc_code, gc_consts, gc_bindstack;
     /* The `gcv_N' field is only filled in with the stack-size when there's
        a chance of gc.	*/
     rep_GC_n_roots gc_stackbase;
 
-    rep_DECLARE1(code, rep_STRINGP);
-    rep_DECLARE2(consts, rep_VECTORP);
+    /* this is the number of dynamic `bindings' in effect
+       (including non-variable bindings). */
+    int impurity;
+
     rep_DECLARE3(stkreq, rep_INTP);
 
-    stackbase = alloca(sizeof(repv) * (rep_INT(stkreq) + 1));
+    if(++rep_lisp_depth > rep_max_lisp_depth)
+    {
+	rep_lisp_depth--;
+	return Fsignal(Qerror, rep_LIST_1(rep_VAL(&max_depth)));
+    }
+
+    /* Jump to this label when tail-calling but the current stack
+       is insufficiently large */
+again_stack: {
+    register repv *stackp;
+    repv *stackbase;
+
+#if defined (__GNUC__)
+    /* Using GCC's variable length auto arrays is better for this since
+       the stack space is freed when leaving the containing scope */
+    repv stack[rep_INT (stkreq) + 1];
+#else
+    /* Otherwise just use alloca (). When tail-calling we'll only
+       allocate a new stack if the current is too small. */
+    repv *stack = alloca(sizeof(repv) * (rep_INT(stkreq) + 1));
+#endif
 
     /* Make sure that even when the stack has no entries, the TOP
        element still != 0 (for the error-detection at label quit:) */
+    stackbase = stack;
     *stackbase++ = Qt;
 
+    /* Jump to this label when tail-calling with a large enough stack */
+again:
+    rep_DECLARE1(code, rep_STRINGP);
+    rep_DECLARE2(consts, rep_VECTORP);
+
     stackp = stackbase - 1;
+    bindstack = Fcons (frame, Qnil);
+    impurity = bound_specials (frame);
     rep_PUSHGC(gc_code, code);
     rep_PUSHGC(gc_consts, consts);
     rep_PUSHGC(gc_bindstack, bindstack);
-    rep_PUSHGCN(gc_stackbase, stackbase, 0);
+    rep_PUSHGCN(gc_stackbase, stackbase, STK_USE);
+
+    if(rep_data_after_gc >= rep_gc_threshold)
+	Fgarbage_collect(Qt);
+
+    rep_MAY_YIELD;
 
     pc = rep_STR(code);
 
@@ -241,20 +335,9 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    int arg;
 	    repv tmp, tmp2;
 	    struct rep_Call lc;
-
-	case 0:
-	    /* Terminating NUL character acts as return instruction */
-	    goto quit;
+	    rep_bool was_closed;
 
 	CASE_OP_ARG(OP_CALL)
-#ifdef MINSTACK
-	    if(STK_SIZE <= MINSTACK)
-	    {
-		STK_WARN("lisp-code");
-		TOP = Fsignal(Qstack_error, Qnil);
-		goto quit;
-	    }
-#endif
 	    /* args are still available above the top of the stack,
 	       this just makes things a bit easier. */
 	    POPN(arg);
@@ -265,10 +348,16 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    rep_PUSH_CALL (lc);
 	    gc_stackbase.count = STK_USE;
 
+	    was_closed = rep_FALSE;
+	    if (rep_FUNARGP(tmp))
+	    {
+		rep_USE_FUNARG(tmp);
+		tmp = rep_FUNARG(tmp)->fun;
+		was_closed = rep_TRUE;
+	    }
+
 	    switch(rep_TYPE(tmp))
 	    {
-		rep_bool was_closed;
-
 	    case rep_Subr0:
 		TOP = rep_SUBR0FUN(tmp)();
 		break;
@@ -381,17 +470,10 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		while(arg--)
 		    tmp2 = Fcons(RET_POP, tmp2);
 		lc.args = tmp2;
-		was_closed = rep_FALSE;
-		if (rep_FUNARGP(tmp))
-		{
-		    rep_USE_FUNARG(tmp);
-		    tmp = rep_FUNARG(tmp)->fun;
-		    was_closed = rep_TRUE;
-		}
 		if (rep_CONSP(tmp))
 		{
 		    if(was_closed && rep_CAR(tmp) == Qlambda)
-			TOP = rep_eval_lambda(tmp, tmp2, rep_FALSE);
+			TOP = rep_eval_lambda(tmp, tmp2, rep_FALSE, rep_FALSE);
 		    else if(rep_CAR(tmp) == Qautoload)
 		    {
 			/* I can't be bothered to go to all the hassle
@@ -408,20 +490,58 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		{
 		    repv bindings;
 
-		    if (search_environment (Qjade_byte_code) == Qnil)
+		    if (rep_bytecode_interpreter == 0)
 			goto invalid;
 
-		    bindings = rep_bind_lambda_list(rep_COMPILED_LAMBDA(tmp),
-						    tmp2, rep_FALSE);
-		    if(bindings != rep_NULL)
+		    if (impurity != 0 || *pc != OP_RETURN)
 		    {
-			rep_GC_root gc_bindings;
-			rep_PUSHGC(gc_bindings, bindings);
-			TOP = Fjade_byte_code(rep_COMPILED_CODE(tmp),
-					      rep_COMPILED_CONSTANTS(tmp),
-					      rep_MAKE_INT(rep_COMPILED_STACK(tmp)));
-			rep_POPGC;
-			rep_unbind_symbols(bindings);
+			bindings = (rep_bind_lambda_list
+				    (rep_COMPILED_LAMBDA(tmp),
+				     tmp2, rep_FALSE, rep_FALSE));
+			if(bindings != rep_NULL)
+			{
+			    TOP = (rep_bytecode_interpreter
+				   (rep_COMPILED_CODE(tmp),
+				    rep_COMPILED_CONSTANTS(tmp),
+				    rep_MAKE_INT(rep_COMPILED_STACK(tmp)),
+				    bindings));
+			}
+		    }
+		    else
+		    {
+			/* A tail call that's safe for eliminating */
+
+			/* snap the call stack */
+			rep_call_stack = lc.next;
+			rep_call_stack->fun = lc.fun;
+			rep_call_stack->args = lc.args;
+			rep_call_stack->args_evalled_p = lc.args_evalled_p;
+
+			/* since impurity==0 there can only be lexical
+			   bindings; these were unbound when switching
+			   environments.. */
+
+			bindings = (rep_bind_lambda_list
+				    (rep_COMPILED_LAMBDA(tmp),
+				     tmp2, rep_FALSE, rep_FALSE));
+			if(bindings != rep_NULL)
+			{
+			    /* set up parameters */
+			    code = rep_COMPILED_CODE (tmp);
+			    consts = rep_COMPILED_CONSTANTS (tmp);
+			    frame = bindings;
+
+			    /* do the goto, after deciding if the
+			       current stack allocation is sufficient. */
+			    rep_POPGCN; rep_POPGC; rep_POPGC; rep_POPGC;
+			    if (rep_COMPILED_STACK (tmp) > rep_INT(stkreq))
+			    {
+				stkreq = rep_MAKE_INT(rep_COMPILED_STACK(tmp));
+				goto again_stack;
+			    }
+			    else
+				goto again;
+			}
 		    }
 		}
 		else
@@ -439,44 +559,46 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    goto fetch;
 
 	CASE_OP_ARG(OP_REFQ)
-	    /* optimise for common case of lexical binding */
+	    /* this instruction is normally only used for special
+	       variables, so optimize the usual path */
 	    tmp = rep_VECT(consts)->array[arg];
-	    if (rep_SYMBOLP(tmp))
+	    if ((rep_SYM(tmp)->car & rep_SF_SPECIAL)
+		&& !(rep_SYM(tmp)->car & rep_SF_LOCAL))
 	    {
-		register repv x = tmp;
-		if ((rep_SYM(x)->car & (rep_SF_LOCAL | rep_SF_SPECIAL)) == 0)
+		/* bytecode interpreter is allowed to assume
+		   unrestricted environment.. */
+		repv tem = search_special_bindings (tmp);
+		if (tem != Qnil)
+		    tem = rep_CDR (tem);
+		else
+		    tem = rep_SYM(tmp)->value;
+		if (!rep_VOIDP(tem))
 		{
-		    /* lexically bound variable */
-		    x = search_environment (x);
-		    if (rep_CONSP(x))
-		    {
-			PUSH(rep_CDR(x));
-			goto fetch;
-		    }
+		    PUSH (tem);
+		    goto fetch;
 		}
 	    }
-	    /* fall back to common method */
+	    /* fall back to common case */
 	    PUSH(Fsymbol_value(rep_VECT(consts)->array[arg], Qnil));
 	    break;
 
 	CASE_OP_ARG(OP_SETQ)
-	    /* optimise for common case of lexical binding */
+	    /* this instruction is normally only used for special
+	       variables, so optimize the usual path */
 	    tmp = rep_VECT(consts)->array[arg];
-	    if (rep_SYMBOLP(tmp))
+	    if ((rep_SYM(tmp)->car & rep_SF_SPECIAL)
+		&& !(rep_SYM(tmp)->car & rep_SF_LOCAL))
 	    {
-		register repv x = tmp;
-		if ((rep_SYM(x)->car & (rep_SF_LOCAL | rep_SF_SPECIAL)) == 0)
-		{
-		    /* lexically bound variable */
-		    x = search_environment (x);
-		    if (rep_CONSP(x))
-		    {
-			rep_CDR(x) = RET_POP;
-			goto fetch;
-		    }
-		}
+		/* bytecode interpreter is allowed to assume
+		   unrestricted environment.. */
+		repv tem = search_special_bindings (tmp);
+		if (tem != Qnil)
+		    rep_CDR (tem) = RET_POP;
+		else
+		    rep_SYM(tmp)->value = RET_POP;
+		goto fetch;
 	    }
-	    /* fall back to common method */
+	    /* fall back to common case */
 	    Fset(rep_VECT(consts)->array[arg], RET_POP);
 	    break;
 
@@ -490,12 +612,44 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	CASE_OP_ARG(OP_BIND)
 	    tmp = rep_VECT(consts)->array[arg];
 	    tmp2 = RET_POP;
-	    if(rep_SYMBOLP(tmp))
-		rep_CAR(bindstack) = rep_bind_symbol(rep_CAR(bindstack),
-						     tmp, tmp2);
-	    else
-		rep_signal_arg_error(tmp, 1);
-	    break;
+	    rep_env = Fcons (Fcons (tmp, tmp2), rep_env);
+	    rep_CDAR(bindstack) = rep_MAKE_INT(rep_INT(rep_CDAR(bindstack))+1);
+	    goto fetch;
+
+	CASE_OP_ARG(OP_BINDSPEC)
+	    tmp = rep_VECT(consts)->array[arg];
+	    tmp2 = RET_POP;
+	    /* assuming non-restricted environment */
+	    rep_special_bindings = Fcons (Fcons (tmp, tmp2),
+					  rep_special_bindings);
+	    rep_CAAR(bindstack) = rep_MAKE_INT(rep_INT(rep_CAAR(bindstack))+1);
+	    impurity++;
+	    goto fetch;
+
+	CASE_OP_ARG(OP_REFN)
+	    tmp = snap_environment (arg);
+	    PUSH(rep_CDR(tmp));
+	    goto fetch;
+
+	CASE_OP_ARG(OP_SETN)
+	    tmp = snap_environment (arg);
+	    rep_CDR(tmp) = RET_POP;
+	    goto fetch;
+
+	CASE_OP_ARG(OP_REFG)
+	    tmp = rep_SYM(rep_VECT(consts)->array[arg])->value;
+	    if (!rep_VOIDP(tmp))
+	    {
+		PUSH(tmp);
+		goto fetch;
+	    }
+	    /* fallback */
+	    goto rep_CONCAT(op_,OP_REFQ);
+
+	CASE_OP_ARG(OP_SETG)
+	    tmp = rep_VECT(consts)->array[arg];
+	    rep_SYM(tmp)->value = RET_POP;
+	    goto fetch;
 
 	case OP_REF:
 	    TOP = Fsymbol_value(TOP, Qnil);
@@ -504,12 +658,12 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	case OP_SET:
 	    CALL_2(Fset);
 
-	case OP_DSET:
+	case OP_DSET:			/* XXX deprecated */
 	    tmp = RET_POP;
 	    tmp2 = TOP;
 	    TOP = Fset (tmp2, tmp);
 	    if (TOP != rep_NULL)
-		rep_SYM(tmp2)->car |= rep_SF_DEFVAR;
+		Fmark_symbol_defined (tmp2);
 	    break;
 
 	case OP_ENCLOSE:
@@ -517,12 +671,13 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    break;
 
 	case OP_INIT_BIND:
-	    bindstack = Fcons(Qnil, bindstack);
+	    bindstack = Fcons(Fcons (rep_MAKE_INT(0), rep_MAKE_INT(0)),
+			      bindstack);
 	    goto fetch;
 
 	case OP_UNBIND:
 	    gc_stackbase.count = STK_USE;
-	    rep_unbind_object(rep_CAR(bindstack));
+	    impurity -= rep_unbind_object(rep_CAR(bindstack));
 	    bindstack = rep_CDR(bindstack);
 	    break;
 
@@ -594,75 +749,66 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    CALL_1(Feval);
 
 	case OP_ADD:
+	    /* open-code fixnum arithmetic */
 	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) + rep_INT(tmp));
-	    else
+	    tmp2 = TOP;
+	    if (rep_INTP (tmp) && rep_INTP (tmp2))
 	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
+		long x = rep_INT (tmp2) + rep_INT (tmp);
+		if (x >= rep_LISP_MIN_INT && x <= rep_LISP_MAX_INT)
+		{
+		    TOP = rep_MAKE_INT (x);
+		    goto fetch;
+		}
 	    }
+	    TOP = rep_number_add (tmp2, tmp);
 	    break;
 
 	case OP_NEG:
-	    if(rep_INTP(TOP))
-		TOP = rep_MAKE_INT(-rep_INT(TOP));
-	    else
-		rep_signal_arg_error(TOP, 1);
+	    /* open-code fixnum arithmetic */
+	    tmp = TOP;
+	    if (rep_INTP (tmp))
+	    {
+		long x = - rep_INT (tmp);
+		if (x >= rep_LISP_MIN_INT && x <= rep_LISP_MAX_INT)
+		{
+		    TOP = rep_MAKE_INT (x);
+		    goto fetch;
+		}
+	    }
+	    TOP = rep_number_neg (tmp);
 	    break;
 
 	case OP_SUB:
+	    /* open-code fixnum arithmetic */
 	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) - rep_INT(tmp));
-	    else
+	    tmp2 = TOP;
+	    if (rep_INTP (tmp) && rep_INTP (tmp2))
 	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
+		long x = rep_INT (tmp2) - rep_INT (tmp);
+		if (x >= rep_LISP_MIN_INT && x <= rep_LISP_MAX_INT)
+		{
+		    TOP = rep_MAKE_INT (x);
+		    goto fetch;
+		}
 	    }
+	    TOP = rep_number_sub (tmp2, tmp);
 	    break;
 
 	case OP_MUL:
-	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) * rep_INT(tmp));
-	    else
-	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
-	    }
-	    break;
+	    CALL_2(rep_number_mul);
 
 	case OP_DIV:
-	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) / rep_INT(tmp));
-	    else
-	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
-	    }
-	    break;
+	    CALL_2(rep_number_div);
 
 	case OP_REM:
 	    CALL_2(Fremainder);
 
 	case OP_LNOT:
-	    if(rep_INTP(TOP))
-		TOP = rep_MAKE_INT(~rep_INT(TOP));
-	    else
-		rep_signal_arg_error(TOP, 1);
-	    break;
+	    CALL_1(Flognot);
 
 	case OP_NOT:
+	case OP_NULL:
 	    if(TOP == Qnil)
 		TOP = Qt;
 	    else
@@ -670,47 +816,23 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    goto fetch;
 
 	case OP_LOR:
-	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) | rep_INT(tmp));
-	    else
-	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
-	    }
-	    break;
+	    CALL_2(rep_number_logior);
 
 	case OP_LXOR:
-	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) ^ rep_INT(tmp));
-	    else
-	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
-	    }
-	    break;
+	    CALL_2(rep_number_logxor);
 
 	case OP_LAND:
-	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) & rep_INT(tmp));
-	    else
-	    {
-		if(rep_INTP(tmp))
-		    rep_signal_arg_error(TOP, 2);
-		else
-		    rep_signal_arg_error(tmp, 1);
-	    }
-	    break;
+	    CALL_2(rep_number_logand);
 
 	case OP_EQUAL:
 	    tmp = RET_POP;
-	    if(!(rep_value_cmp(TOP, tmp)))
+	    tmp2 = TOP;
+	    if (rep_INTP (tmp) && rep_INTP (tmp2))
+	    {
+		TOP = (tmp2 == tmp) ? Qt : Qnil;
+		goto fetch;
+	    }
+	    if(!(rep_value_cmp(tmp2, tmp)))
 		TOP = Qt;
 	    else
 		TOP = Qnil;
@@ -724,12 +846,6 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		TOP = Qnil;
 	    goto fetch;
 
-	case OP_NUM_EQ:
-	    CALL_2(Fnum_eq);
-
-	case OP_NUM_NOTEQ:
-	    CALL_2(Fnum_noteq);
-
 	case OP_GT:
 	    tmp = RET_POP;
 	    if(rep_value_cmp(TOP, tmp) > 0)
@@ -763,36 +879,46 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    break;
 
 	case OP_INC:
-	    if(rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) + 1);
-	    else
-		rep_signal_arg_error(TOP, 1);
+	    tmp = TOP;
+	    if (rep_INTP (tmp))
+	    {
+		long x = rep_INT (tmp) + 1;
+		if (x <= rep_LISP_MAX_INT)
+		{
+		    TOP = rep_MAKE_INT (x);
+		    goto fetch;
+		}
+	    }
+	    TOP = Fplus1 (tmp);
 	    break;
 
 	case OP_DEC:
-	    if(rep_INTP(TOP))
-		TOP = rep_MAKE_INT(rep_INT(TOP) - 1);
-	    else
-		rep_signal_arg_error(TOP, 1);
+	    tmp = TOP;
+	    if (rep_INTP (tmp))
+	    {
+		long x = rep_INT (tmp) - 1;
+		if (x >= rep_LISP_MIN_INT)
+		{
+		    TOP = rep_MAKE_INT (x);
+		    goto fetch;
+		}
+	    }
+	    TOP = Fsub1 (tmp);
 	    break;
 
-	case OP_LSH:
-	    CALL_2(Flsh);
+	case OP_ASH:
+	    CALL_2(Fash);
 
 	case OP_ZEROP:
-	    if(rep_INTP(TOP) && (rep_INT(TOP) == 0))
-		TOP = Qt;
-	    else
-		TOP = Qnil;
+	    tmp = TOP;
+	    if (rep_INTP (tmp))
+	    {
+		TOP = (tmp == rep_MAKE_INT (0)) ? Qt : Qnil;
+		goto fetch;
+	    }
+	    TOP = Fzerop (tmp);
 	    break;
 
-	case OP_NULL:
-	    if(rep_NILP(TOP))
-		TOP = Qt;
-	    else
-		TOP = Qnil;
-	    goto fetch;
-
 	case OP_ATOM:
 	    if(!rep_CONSP(TOP))
 		TOP = Qt;
@@ -815,7 +941,7 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    goto fetch;
 
 	case OP_NUMBERP:
-	    if(rep_INTP(TOP))
+	    if(rep_NUMERICP(TOP))
 		TOP = Qt;
 	    else
 		TOP = Qnil;
@@ -862,7 +988,21 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	       This installs an address in the code string as an
 	       error handler. */
 	    tmp = RET_POP;
-	    bindstack = Fcons(Fcons(tmp, rep_MAKE_INT(STK_USE)), bindstack);
+	    bindstack = Fcons (Fcons (Qerror,
+				      Fcons (tmp, rep_MAKE_INT(STK_USE))),
+			       bindstack);
+	    impurity++;
+	    break;
+
+	case OP_RETURN:
+	    gc_stackbase.count = STK_USE;
+	    unbind_all (bindstack);
+	    goto quit;
+
+	case OP_UNBINDALL:
+	    gc_stackbase.count = STK_USE;
+	    bindstack = unbind_all_but_one (bindstack);
+	    impurity = bound_specials (rep_CAR(bindstack));
 	    break;
 
 	case OP_BOUNDP:
@@ -898,8 +1038,12 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		tmp = rep_CDR(TOP);	/* the error data */
 		tmp2 = stackp[-1];	/* the symbol to bind to */
 		if(rep_SYMBOLP(tmp2) && !rep_NILP(tmp2))
+		{
 		    bindstack = Fcons(rep_bind_symbol(Qnil, tmp2, tmp),
-					 bindstack);
+				      bindstack);
+		    if (rep_SYM(tmp2)->car & rep_SF_SPECIAL)
+			impurity++;
+		}
 		else
 		    /* Placeholder to allow simple unbinding */
 		    bindstack = Fcons(Qnil, bindstack);
@@ -911,6 +1055,9 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    gc_stackbase.count = STK_USE;
 	    CALL_2(Fsignal);
 
+	case OP_QUOTIENT:
+	    CALL_2(Fquotient);
+
 	case OP_REVERSE:
 	    CALL_1(Freverse);
 
@@ -976,12 +1123,7 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    CALL_1(Fsubrp);
 
 	case OP_EQL:
-	    tmp = RET_POP;
-	    if(rep_INTP(tmp) && rep_INTP(TOP))
-		TOP = (rep_INT(TOP) == rep_INT(tmp) ? Qt : Qnil);
-	    else
-		TOP = (TOP == tmp ? Qt : Qnil);
-	    goto fetch;
+	    CALL_2(Feql);
 
 	case OP_MAX:
 	    tmp = RET_POP;
@@ -1099,11 +1241,55 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    TOP = list_ref (TOP, 7);
 	    goto fetch;
 
+	case OP_FLOOR:
+	    CALL_1(Ffloor);
+
+	case OP_CEILING:
+	    CALL_1(Fceiling);
+
+	case OP_TRUNCATE:
+	    CALL_1(Ftruncate);
+
+	case OP_ROUND:
+	    CALL_1(Fround);
+
 	case OP_BINDOBJ:
 	    tmp = RET_POP;
 	    bindstack = Fcons(rep_bind_object(tmp), bindstack);
+	    impurity++;
 	    break;
 
+	case OP_FORBID:
+	    rep_FORBID;
+	    PUSH (rep_PREEMPTABLE_P ? Qnil : Qt);
+	    goto fetch;
+
+	case OP_PERMIT:
+	    rep_PERMIT;
+	    PUSH (rep_PREEMPTABLE_P ? Qnil : Qt);
+	    goto fetch;
+
+	case OP_EXP:
+	    CALL_1(Fexp);
+
+	case OP_LOG:
+	    CALL_1(Flog);
+
+	case OP_COS:
+	    CALL_1(Fcos);
+
+	case OP_SIN:
+	    CALL_1(Fsin);
+
+	case OP_TAN:
+	    CALL_1(Ftan);
+
+	case OP_SQRT:
+	    CALL_1(Fsqrt);
+
+	case OP_EXPT:
+	    CALL_2(Fexpt);
+
 	case OP_SWAP2:
 	    tmp = TOP;
 	    TOP = stackp[-1];
@@ -1126,6 +1312,7 @@ of byte code. See the functions `compile-file', `compile-directory' and
 
 	case OP_BINDENV:
 	    bindstack = Fcons (Fmake_closure (Qnil, Qnil), bindstack);
+	    impurity++;
 	    break;
 
 	/* Jump instructions follow */
@@ -1199,6 +1386,8 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		gc_stackbase.count = STK_USE;
 		Fgarbage_collect(Qt);
 	    }
+	    gc_stackbase.count = STK_USE;
+	    rep_MAY_YIELD;
 	    break;
 
 	default:
@@ -1222,22 +1411,23 @@ of byte code. See the functions `compile-file', `compile-directory' and
 	    while(rep_CONSP(bindstack))
 	    {
 		repv item = rep_CAR(bindstack);
-		if(!rep_CONSP(item)
-		   || !rep_INTP(rep_CAR(item)) || !rep_INTP(rep_CDR(item)))
+		if(!rep_CONSP(item) || rep_CAR(item) != Qerror)
 		{
 		    rep_GC_root gc_throwval;
 		    repv throwval = rep_throw_value;
 		    rep_throw_value = rep_NULL;
 		    rep_PUSHGC(gc_throwval, throwval);
 		    gc_stackbase.count = STK_USE;
-		    rep_unbind_object(item);
+		    impurity -= rep_unbind_object(item);
 		    bindstack = rep_CDR(bindstack);
 		    rep_POPGC;
 		    rep_throw_value = throwval;
 		}
 		else if(rep_throw_value != rep_NULL)
 		{
-		    /* car is an exception-handler, (PC . SP)
+		    item = rep_CDR(item);
+
+		    /* item is an exception-handler, (PC . SP)
 
 		       When the code at PC is called, it will have
 		       the current stack usage set to SP, and then
@@ -1252,6 +1442,7 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		    rep_throw_value = rep_NULL;
 		    pc = rep_STR(code) + rep_INT(rep_CAR(item));
 		    bindstack = rep_CDR(bindstack);
+		    impurity--;
 		    goto fetch;
 		}
 		else
@@ -1259,6 +1450,7 @@ of byte code. See the functions `compile-file', `compile-directory' and
 		    /* car is an exception handler, but rep_throw_value isn't
 		       set, so there's nothing to handle. Keep unwinding. */
 		    bindstack = rep_CDR(bindstack);
+		    impurity--;
 		}
 	    }
 	    TOP = rep_NULL;
@@ -1270,6 +1462,9 @@ quit:
     /* only use this var to save declaring another */
     bindstack = TOP;
 
+    /* close the stack scope */ }
+
+    rep_lisp_depth--;
     rep_POPGCN; rep_POPGC; rep_POPGC; rep_POPGC;
     return bindstack;
 }
@@ -1304,7 +1499,7 @@ Return an object that can be used as the function value of a symbol.
     if(len < rep_COMPILED_MIN_SLOTS)
 	return rep_signal_missing_arg(len + 1);
     
-    if(!rep_LISTP(rep_CAR(args)))
+    if(!rep_CONSP(rep_CAR(args)) && !rep_SYMBOLP(rep_CAR(args)))
 	return rep_signal_arg_error(rep_CAR(args), 1);
     obj[0] = rep_CAR(args); args = rep_CDR(args);
     if(!rep_STRINGP(rep_CAR(args)))
