@@ -1,6 +1,6 @@
 /* main.c -- Entry point for Jade
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: main.c,v 1.39 1999/12/06 11:33:12 john Exp $
+   $Id: main.c,v 1.50 2000/04/03 17:28:50 john Exp $
 
    This file is part of Jade.
 
@@ -55,6 +55,9 @@ void (*rep_on_termination_fun)(void);
 /* The event-loop function, may be entered recursively. */
 repv (*rep_event_loop_fun)(void) = rep_event_loop;
 
+/* rep_init () will set this to an early stack pointer */
+char *rep_stack_bottom;
+
 DEFSYM(exit, "exit");
 DEFSYM(quit, "quit");
 DEFSYM(top_level, "top-level");
@@ -64,6 +67,7 @@ DEFSYM(interpreted_mode, "interpreted-mode");
 DEFSYM(program_name, "program-name");
 DEFSYM(error_mode, "error-mode");
 DEFSYM(interrupt_mode, "interrupt-mode");
+DEFSYM(before_exit_hook, "before-exit-hook");
 
 DEFSTRING(definit, "init");
 static repv init_script = rep_VAL(&definit);
@@ -76,10 +80,9 @@ usage(char *prog_name, void (*sys_usage)(void))
     fprintf(stderr, "usage: %s [OPTIONS...]\n", prog_name);
     fputs ("\
 where OPTIONS may include:\n\
-    --init FILE		use FILE instead of `init.jl' to boot from\n\
-    --batch		don't open any windows; process args and exit\n\
-    --interp		don't load compiled Lisp files\n\
-    --warn-shadowing	warn about redefining functions through binding\n",
+    --init FILE		use FILE instead of `init' to boot from\n\
+    --batch		batch mode: process command line args and exit\n\
+    --interp		don't load compiled Lisp files\n",
 	   stderr);
     if (sys_usage != 0)
 	(*sys_usage)();
@@ -94,7 +97,7 @@ rep_bool
 rep_get_option (char *option, repv *argp)
 {
     int optlen = strlen(option);
-    repv tem = rep_SYM(Qcommand_line_args)->value;
+    repv tem = Fsymbol_value (Qcommand_line_args, Qt);
     while (!rep_INTERRUPTP && rep_CONSP(tem) && rep_STRINGP(rep_CAR(tem)))
     {
 	if (strncmp (option, rep_STR(rep_CAR(tem)), optlen) == 0)
@@ -102,8 +105,8 @@ rep_get_option (char *option, repv *argp)
 	    repv opt = rep_CAR(tem), cdr = rep_CDR(tem);
 	    if (rep_STR(opt)[optlen] == '=' || rep_STR(opt)[optlen] == 0)
 	    {
-		rep_SYM(Qcommand_line_args)->value
-		    = Fdelq (opt, rep_SYM(Qcommand_line_args)->value);
+		Fset (Qcommand_line_args,
+		      Fdelq (opt, Fsymbol_value (Qcommand_line_args, Qt)));
 		if (argp != 0)
 		{
 		    if (rep_STR(opt)[optlen] == '=')
@@ -114,8 +117,8 @@ rep_get_option (char *option, repv *argp)
 		    else if (rep_CONSP(cdr) && rep_STRINGP(rep_CAR(cdr)))
 		    {
 			*argp = rep_CAR(cdr);
-			rep_SYM(Qcommand_line_args)->value
-			    = Fdelq (*argp, rep_SYM(Qcommand_line_args)->value);
+			Fset (Qcommand_line_args,
+			      Fdelq (*argp, Fsymbol_value(Qcommand_line_args, Qt)));
 			return rep_TRUE;
 		    }
 		    else
@@ -173,18 +176,16 @@ get_main_options(char *prog_name, int *argc_p,
 	argc--;
 	argv++;
     }
-    rep_SYM(Qcommand_line_args)->value = head;
+    Fset (Qcommand_line_args, head);
     *argc_p = argc;
     *argv_p = argv;
 
     if (rep_get_option ("--init", &opt))
 	init_script = opt;
     if (rep_get_option("--batch", 0))
-	rep_SYM(Qbatch_mode)->value = Qt;
+	Fset (Qbatch_mode, Qt);
     if (rep_get_option("--interp", 0))
-	rep_SYM(Qinterpreted_mode)->value = Qt;
-    if (rep_get_option("--warn-shadowing", 0))
-	rep_warn_shadowing = rep_TRUE;
+	Fset (Qinterpreted_mode, Qt);
     if (rep_get_option("--help", 0) || rep_get_option ("-?", 0))
     {
 	usage(prog_name, sys_usage);
@@ -194,9 +195,21 @@ get_main_options(char *prog_name, int *argc_p,
     return rep_TRUE;
 }
 
+/* Note that `argc' _must_ (I mean _must_!) be a pointer to the real
+   argc on the stack frame of the outermost procedure */
 void
 rep_init(char *prog_name, int *argc, char ***argv,
 	 void (*sys_symbols)(void), void (*sys_usage)(void))
+{
+    char *dump_file = getenv ("REPDUMPFILE");
+    rep_init_from_dump (prog_name, argc, argv,
+			sys_symbols, sys_usage, dump_file);
+}
+
+void
+rep_init_from_dump(char *prog_name, int *argc, char ***argv,
+		   void (*sys_symbols)(void), void (*sys_usage)(void),
+		   char *dump_file)
 {
     if(sizeof(rep_PTR_SIZED_INT) < sizeof(void *))
     {
@@ -214,15 +227,16 @@ rep_init(char *prog_name, int *argc, char ***argv,
     rep_pre_sys_os_init();
     if(rep_pre_symbols_init())
     {
-#ifdef rep_DUMPED
-	/* Must initialise dumped out symbols before interning _any_
-	   symbols by hand. */
-	rep_dumped_init();
-#endif
+	char *tem = getenv ("REPUNDUMPED");
+	if (dump_file && (!tem || atoi(tem) == 0))
+	    rep_dumped_init (dump_file);
+
 	rep_symbols_init();
+	rep_numbers_init ();
 
 	rep_values_init();
 	rep_lisp_init();
+	rep_macros_init ();
 	rep_lispcmds_init();
 	rep_lispmach_init();
 	rep_find_init();
@@ -232,43 +246,44 @@ rep_init(char *prog_name, int *argc, char ***argv,
 	rep_files_init();
 	rep_sys_os_init();
 
+	/* XXX Assumes that argc is on the stack. I can't think of
+	   XXX any other way to reliably find the real base of the
+	   XXX stack.. */
+	rep_stack_bottom = (char *) argc;
+	rep_continuations_init ();
+
 	if (sys_symbols != 0)
 	    (*sys_symbols)();
 
-	rep_SYM(Qprogram_name)->value = rep_string_dup (prog_name);
+	Fset (Qprogram_name, rep_string_dup (prog_name));
 
 	if(get_main_options(prog_name, argc, argv, sys_usage))
-	{
-	    repv res = Fload(init_script, Qnil, Qnil, Qnil, Qnil);
-	    if (res != rep_NULL)
-		return;
-	}
+	    return;
     }
-
-    if(rep_throw_value && rep_CAR(rep_throw_value) == Qerror)
-    {
-	/* If quitting due to an error, print the error cell if
-	   at all possible. */
-	repv stream = Fstderr_file();
-	repv old_tv = rep_throw_value;
-	rep_GC_root gc_old_tv;
-	rep_PUSHGC(gc_old_tv, old_tv);
-	rep_throw_value = rep_NULL;
-	if(stream && rep_FILEP(stream))
-	{
-	    fputs("error--> ", stderr);
-	    Fprin1(rep_CDR(old_tv), stream);
-	    fputc('\n', stderr);
-	}
-	else
-	    fputs("rep: error in initialisation\n", stderr);
-	rep_throw_value = old_tv;
-	rep_POPGC;
-    }
-
     exit (10);
 }
 
+/* Should be called sometime after calling rep_init*. It will load
+   the standard init script, plus FILE if non-nil. Returns the
+   result of the last form evaluated. */
+repv
+rep_load_environment (repv file)
+{
+    repv res = Qnil;
+    rep_GC_root gc_file;
+
+    rep_PUSHGC (gc_file, file);
+    if (rep_dumped_non_constants != rep_NULL)
+	res = Feval (rep_dumped_non_constants);
+    if (res != rep_NULL)
+	res = Fload (init_script, Qnil, Qnil, Qnil, Qnil);
+    if (res != rep_NULL && rep_STRINGP(file))
+	res = Fload (file, Qnil, Qnil, Qnil, Qnil);
+    rep_POPGC;
+
+    return res;
+}
+
 void
 rep_kill(void)
 {
@@ -390,6 +405,41 @@ rep_handle_input_exception(repv *result_p)
     return rep_FALSE;
 }
 
+/* should be called before exiting (for any reason). returns the value
+   that should be returned by the process */
+int
+rep_top_level_exit (void)
+{
+    rep_GC_root gc_throw;
+    repv throw = rep_throw_value;
+    rep_throw_value = rep_NULL;
+    if(throw && rep_CAR(throw) == Qerror)
+    {
+	/* If quitting due to an error, print the error cell if
+	   at all possible. */
+	repv stream = Fstderr_file();
+	if(stream && rep_FILEP(stream))
+	{
+	    fputs("error--> ", stderr);
+	    Fprin1(rep_CDR(throw), stream);
+	    fputc('\n', stderr);
+	}
+	else
+	    fputs("error in initialisation\n", stderr);
+	return 10;
+    }
+
+    rep_PUSHGC(gc_throw, throw);
+    Fcall_hook (Qbefore_exit_hook, Qnil, Qnil);
+    rep_throw_value = rep_NULL;
+    rep_POPGC;
+
+    if (throw && rep_CAR (throw) == Qquit && rep_INTP (rep_CDR(throw)))
+	return (rep_INT (rep_CDR(throw)));
+
+    return 0;
+}
+
 DEFUN_INT("recursive-edit", Frecursive_edit, Srecursive_edit, (void), rep_Subr0, "") /*
 ::doc:recursive-edit::
 recursive-edit
@@ -434,13 +484,14 @@ rep_main_init(void)
     rep_INTERN_SPECIAL(command_line_args);
     rep_INTERN_SPECIAL(idle_hook);
     rep_INTERN_SPECIAL(batch_mode);
-    rep_SYM(Qbatch_mode)->value = Qnil;
+    Fset (Qbatch_mode, Qnil);
     rep_INTERN_SPECIAL(interpreted_mode);
-    rep_SYM(Qinterpreted_mode)->value = Qnil;
+    Fset (Qinterpreted_mode, Qnil);
     rep_ADD_SUBR(Sget_command_line_option);
     rep_INTERN_SPECIAL(program_name);
     rep_INTERN_SPECIAL(error_mode);
-    rep_SYM(Qerror_mode)->value = Qnil;
+    Fset (Qerror_mode, Qnil);
     rep_INTERN_SPECIAL(interrupt_mode);
-    rep_SYM(Qinterrupt_mode)->value = Qnil;
+    Fset (Qinterrupt_mode, Qnil);
+    rep_INTERN_SPECIAL(before_exit_hook);
 }
