@@ -1,6 +1,6 @@
 /* misc.c -- Miscellaneous functions
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: misc.c,v 1.33 1999/11/29 18:37:56 john Exp $
+   $Id: misc.c,v 1.38 2000/03/29 23:00:37 john Exp $
 
    This file is part of Jade.
 
@@ -37,6 +37,7 @@ DEFSTRING(rep_version_string, REP_VERSION);
 DEFSYM(operating_system, "operating-system");
 DEFSYM(process_environment, "process-environment");
 DEFSYM(rep_version, "rep-version");
+DEFSYM(rep_interface_id, "rep-interface-id");
 DEFSYM(rep_build_id, "rep-build-id"); /*
 ::doc:operating-system::
 A symbol defining the type of operating system that Jade is running
@@ -423,7 +424,7 @@ with the current time of day.
 }
 
 DEFUN("translate-string", Ftranslate_string, Stranslate_string, (repv string, repv table), rep_Subr2) /*
-::doc:Stranslate-string:
+::doc:translate-string:
 translate-string STRING TRANSLATION-TABLE
 
 Applies the TRANSLATION-TABLE to each character in the string STRING.
@@ -533,7 +534,7 @@ Returns the lower-case equivalent of CHAR.
     return rep_MAKE_INT(tolower(rep_INT(ch)));
 }
 
-DEFUN("system", Fsystem, Ssystem, (repv command), rep_Subr1) /*
+DEFUN_INT("system", Fsystem, Ssystem, (repv command), rep_Subr1, "sShell command:") /*
 ::doc:system::
 system SHELL-COMMAND
 
@@ -562,16 +563,18 @@ rep_misc_init(void)
     rep_INTERN_SPECIAL(operating_system);
 #ifdef rep_HAVE_UNIX
     rep_INTERN(unix);
-    rep_SYM(Qoperating_system)->value = Qunix;
+    Fset (Qoperating_system, Qunix);
 #endif
 
     rep_INTERN_SPECIAL(process_environment);
-    rep_SYM(Qprocess_environment)->value = Qnil;
+    Fset (Qprocess_environment, Qnil);
 
     rep_INTERN_SPECIAL(rep_version);
-    rep_SYM(Qrep_version)->value = rep_VAL(&rep_version_string);
+    Fset (Qrep_version, rep_VAL(&rep_version_string));
+    rep_INTERN_SPECIAL(rep_interface_id);
+    Fset (Qrep_interface_id, rep_VAL(rep_MAKE_INT(rep_INTERFACE)));
     rep_INTERN_SPECIAL(rep_build_id);
-    rep_SYM(Qrep_build_id)->value = rep_VAL(&build_id_string);
+    Fset (Qrep_build_id, rep_VAL(&build_id_string));
 
     rep_ADD_SUBR_INT(Sbeep);
     rep_ADD_SUBR(Scomplete_string);
@@ -597,24 +600,35 @@ rep_misc_init(void)
     rep_ADD_SUBR(Sspace_char_p);
     rep_ADD_SUBR(Schar_upcase);
     rep_ADD_SUBR(Schar_downcase);
-    rep_ADD_SUBR(Ssystem);
+    rep_ADD_SUBR_INT(Ssystem);
 
-    rep_INTERN_SPECIAL(upcase_table);
-    rep_SYM(Qupcase_table)->value = rep_make_string(257);
-    rep_INTERN_SPECIAL(downcase_table);
-    rep_SYM(Qdowncase_table)->value = rep_make_string(257);
-    for(i = 0; i < 256; i++)
     {
-	rep_STR(rep_SYM(Qupcase_table)->value)[i] = toupper(i);
-	rep_STR(rep_SYM(Qdowncase_table)->value)[i] = tolower(i);
-    }
-    rep_STR(rep_SYM(Qupcase_table)->value)[256] = 0;
-    rep_STR(rep_SYM(Qdowncase_table)->value)[256] = 0;
+	repv up = rep_make_string (257);
+	repv down = rep_make_string (257);
 
-    rep_INTERN_SPECIAL(flatten_table);
-    rep_SYM(Qflatten_table)->value = rep_make_string(12);
-    for(i = 0; i < 10; i++)
-	rep_STR(rep_SYM(Qflatten_table)->value)[i] = i;
-    rep_STR(rep_SYM(Qflatten_table)->value)[10] = ' ';
-    rep_STR(rep_SYM(Qflatten_table)->value)[11] = 0;
+	for(i = 0; i < 256; i++)
+	{
+	    rep_STR(up)[i] = toupper(i);
+	    rep_STR(down)[i] = tolower(i);
+	}
+	rep_STR(up)[256] = 0;
+	rep_STR(down)[256] = 0;
+
+	rep_INTERN_SPECIAL(upcase_table);
+	rep_INTERN_SPECIAL(downcase_table);
+	Fset (Qupcase_table, up);
+	Fset (Qdowncase_table, down);
+    }
+
+    {
+	repv flatten = rep_make_string (12);
+
+	for(i = 0; i < 10; i++)
+	    rep_STR(flatten)[i] = i;
+	rep_STR(flatten)[10] = ' ';
+	rep_STR(flatten)[11] = 0;
+
+	rep_INTERN_SPECIAL(flatten_table);
+	Fset (Qflatten_table, flatten);
+    }
 }
