@@ -1,6 +1,6 @@
 /* readline.c -- wrap some readline functions when available
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
-   $Id: readline.c,v 1.3 1999/12/13 09:54:02 john Exp $
+   $Id: readline.c,v 1.7 2000/02/28 08:29:28 john Exp $
 
    This file is part of librep.
 
@@ -22,6 +22,7 @@
 
 #include <config.h>
 #include <rep.h>
+#include <sys/time.h>
 
 #ifdef HAVE_LIBREADLINE
 # include <readline/readline.h>
@@ -69,6 +70,92 @@ completion_generator (char *word, int state)
     else
 	return 0;
 }
+
+/* gratuitously stolen from guile, guile-readline/readline.c */
+static void match_paren(int x, int k);
+static int find_matching_paren(int k);
+static void init_bouncing_parens();
+
+static void
+init_bouncing_parens()
+{
+  if(strncmp(rl_get_keymap_name(rl_get_keymap()), "vi", 2)) {
+    rl_bind_key(')', match_paren);
+    rl_bind_key(']', match_paren);
+    rl_bind_key('}', match_paren);
+  }
+}
+
+static int
+find_matching_paren(int k)
+{
+  register int i;
+  register char c = 0;
+  int end_parens_found = 0;
+
+  /* Choose the corresponding opening bracket.  */
+  if (k == ')') c = '(';
+  else if (k == ']') c = '[';
+  else if (k == '}') c = '{';
+
+  for (i=rl_point-2; i>=0; i--)
+    {
+      /* Is the current character part of a character literal?  */
+      if (i - 2 >= 0
+	  && rl_line_buffer[i - 1] == '\\'
+	  && rl_line_buffer[i - 2] == '#')
+	;
+      else if (rl_line_buffer[i] == k)
+	end_parens_found++;
+      else if (rl_line_buffer[i] == '"')
+	{
+	  /* Skip over a string literal.  */
+	  for (i--; i >= 0; i--)
+	    if (rl_line_buffer[i] == '"'
+		&& ! (i - 1 >= 0
+		      && rl_line_buffer[i - 1] == '\\'))
+	      break;
+	}
+      else if (rl_line_buffer[i] == c)
+	{
+	  if (end_parens_found==0) return i;
+	  else --end_parens_found;
+	}
+    }
+  return -1;
+}
+
+static void
+match_paren(int x, int k)
+{
+  int tmp;
+  fd_set readset;
+  struct timeval timeout;
+  
+  rl_insert(x, k);
+
+  /* Did we just insert a quoted paren?  If so, then don't bounce.  */
+  if (rl_point - 1 >= 1
+      && rl_line_buffer[rl_point - 2] == '\\')
+    return;
+
+  /* tmp = 200000 */
+  timeout.tv_sec = 0 /* tmp / 1000000 */ ; 
+  timeout.tv_usec = 200000 /* tmp % 1000000 */ ;
+  FD_ZERO(&readset);
+  FD_SET(fileno(rl_instream), &readset);
+  
+  if(rl_point > 1) {
+    tmp = rl_point;
+    rl_point = find_matching_paren(k);
+    if(rl_point > -1) {
+      rl_redisplay();
+      select(1, &readset, NULL, NULL, &timeout);
+    }
+    rl_point = tmp;
+  }
+}
+
 #endif
 
 DEFUN("readline", Freadline, Sreadline, (repv prompt_), rep_Subr1)
@@ -104,20 +191,19 @@ DEFUN("readline", Freadline, Sreadline, (repv prompt_), rep_Subr1)
 
 /* DL hooks */
 
-rep_xsubr *rep_dl_subrs[] = { &Sreadline, 0 };
-repv rep_dl_feature;
-
 repv
 rep_dl_init(void)
 {
+    rep_ADD_SUBR(Sreadline);
     rep_INTERN(readline);
     rep_INTERN_SPECIAL(rl_completion_generator);
     rep_INTERN(boundp);
     completions = Qnil;
     rep_mark_static (&completions);
-    rep_dl_feature = Qreadline;
 #ifdef HAVE_LIBREADLINE
     rl_completion_entry_function = (void *) completion_generator;
+    rl_basic_quote_characters = "\"";
+    init_bouncing_parens();
 #endif
-    return Qt;
+    return Qreadline;
 }
