@@ -1023,7 +1023,7 @@ regmatch(char *prog)
 		    nextch = *OPERAND(next);
 		if(regnocase)
 		    nextch = toupper(nextch);
-		no = (OP(scan) == STAR) ? 0 : 1;
+		no = (OP(scan) == NGSTAR) ? 0 : 1;
 		save = reginput;
 		max = regrepeat(OPERAND(scan));
 		while (no <= max) {
