@@ -1,6 +1,6 @@
 /* rep_subrs.h -- mostly LISP subr declarations
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: rep_subrs.h,v 1.17 1999/11/25 23:23:13 john Exp $
+   $Id: rep_subrs.h,v 1.42 2000/03/28 23:52:05 john Exp $
 
    This file is part of Jade.
 
@@ -23,6 +23,32 @@
 
 #include <stdarg.h>
 
+/* from continuations.c */
+extern int rep_thread_lock;
+extern rep_bool rep_pending_thread_yield;
+extern repv rep_call_with_barrier (repv (*callback)(repv), repv arg,
+				   rep_bool closed, void (*in)(void *),
+				   void (*out)(void *), void *data);
+extern repv Fcall_cc (repv thunk);
+extern repv Fcontinuation_callable_p (repv cont);
+extern repv Fcall_with_object (repv arg, repv thunk);
+extern repv Fcall_with_dynamic_root (repv thunk);
+extern repv Fcall_with_barrier (repv thunk, repv closed, repv in, repv out);
+extern repv Fmake_thread (repv thunk, repv name);
+extern repv Fthread_yield (void);
+extern repv Fthread_delete (repv thread);
+extern repv Fthread_suspend (repv thread, repv msecs);
+extern repv Fthread_wake (repv thread);
+extern repv Ftheadp (repv arg);
+extern repv Fthread_suspended_p (repv thread);
+extern repv Fthread_exited_p (repv thread);
+extern repv Fcurrent_thread (repv depth);
+extern repv Fall_threads (repv depth);
+extern repv Fthread_forbid (void);
+extern repv Fthread_permit (void);
+extern repv Fthread_name (repv th);
+extern u_long rep_max_sleep_for (void);
+
 /* from debug-buffer.c */
 extern void *rep_db_alloc(char *name, int size);
 extern void rep_db_free(void *db);
@@ -95,6 +121,7 @@ extern repv Fstdin_file(void);
 extern repv Fstdout_file(void);
 extern repv Fstderr_file(void);
 extern repv Fmake_temp_name(void);
+extern repv rep_file_fdopen (int fd, char *mode);
 
 /* from find.c */
 extern rep_regexp *rep_compile_regexp(repv re);
@@ -145,8 +172,11 @@ extern repv Qno_memory, Quser_interrupt, Qarith_error;
 extern repv Qterm_interrupt;
 extern repv Qstack_error;
 extern repv Qprint_escape, Qprint_length, Qprint_level, Qnewlines;
-extern repv rep_env, rep_fenv, rep_special_env;
+extern repv rep_env, rep_fenv, rep_special_env, rep_special_bindings;
+extern repv (*rep_bytecode_interpreter)(repv code, repv consts,
+					repv stack, repv frame);
 extern struct rep_Call *rep_call_stack;
+extern int rep_lisp_depth, rep_max_lisp_depth;
 extern int rep_test_int_counter;
 extern int rep_test_int_period;
 extern void (*rep_test_int_fun)(void);
@@ -155,7 +185,6 @@ extern repv Feval(repv);
 extern repv Fprogn(repv);
 extern repv Fbreak(void);
 extern repv Fstep(repv);
-extern repv Fmacroexpand(repv, repv);
 extern repv Fsignal(repv error, repv data);
 extern repv Fcondition_case(repv args);
 extern repv Fbacktrace(repv strm);
@@ -164,13 +193,9 @@ extern repv Vmax_lisp_depth(repv val);
 /* from lispcmds.c */
 extern repv Qor, Qand;
 extern repv Qload_path, Qafter_load_alist, Qlisp_lib_directory;
-extern repv Qdl_load_path, Qdl_load_reloc_now, Qprovide;
+extern repv Qdl_load_path, Qdl_load_reloc_now, Qprovide, Qfeatures;
 extern repv Qsite_lisp_directory, Qdocumentation_file, Qdocumentation_files;
 extern repv Fquote(repv);
-extern repv Fdefmacro(repv);
-extern repv Fdefun(repv);
-extern repv Fdefvar(repv);
-extern repv Fdefconst(repv);
 extern repv Fcar(repv);
 extern repv Fcdr(repv);
 extern repv Flist(repv);
@@ -210,49 +235,26 @@ extern repv Flength(repv);
 extern repv Fcopy_sequence(repv);
 extern repv Felt(repv, repv);
 extern repv Fprog1(repv);
-extern repv Fprog2(repv);
 extern repv Fwhile(repv);
 extern repv Fcond(repv);
-extern repv Fif(repv args);
-extern repv Fand(repv);
-extern repv For(repv);
 extern repv Fapply(repv);
 extern repv Fload(repv file, repv noerr_p, repv nopath_p,
 		  repv nosuf_p, repv in_env);
-extern repv Fplus(repv);
-extern repv Fminus(repv);
-extern repv Fproduct(repv);
-extern repv Fdivide(repv);
-extern repv Fremainder(repv n1, repv n2);
-extern repv Fmod(repv n1, repv n2);
-extern repv Flognot(repv);
-extern repv Fnot(repv);
-extern repv Flogior(repv);
-extern repv Flogxor(repv);
-extern repv Flogand(repv);
 extern repv Fequal(repv, repv);
 extern repv Feq(repv, repv);
-extern repv Feql(repv arg1, repv arg2);
 extern repv Fstring_head_eq(repv, repv);
-extern repv Fnum_eq(repv num1, repv num2);
-extern repv Fnum_noteq(repv num1, repv num2);
+extern repv Fnum_eq(repv args);
+extern repv Fnum_noteq(repv args);
 extern repv Fgtthan(repv);
 extern repv Fgethan(repv);
 extern repv Fltthan(repv);
 extern repv Flethan(repv);
 extern repv Fmax(repv);
 extern repv Fmin(repv);
-extern repv Fplus1(repv);
-extern repv Fsub1(repv);
-extern repv Flsh(repv, repv);
-extern repv Fash(repv, repv);
-extern repv Fzerop(repv);
 extern repv Fnull(repv);
 extern repv Fatom(repv);
 extern repv Fconsp(repv);
 extern repv Flistp(repv);
-extern repv Fnumberp(repv);
-extern repv Fintegerp(repv);
 extern repv Fstringp(repv);
 extern repv Fvectorp(repv);
 extern repv Fbytecodep(repv);
@@ -272,20 +274,29 @@ extern repv Fprovide(repv);
 extern repv Frequire(repv, repv);
 
 /* from lispmach.c */
-extern void rep_unbind_object(repv item);
+extern int rep_unbind_object(repv item);
 extern repv rep_bind_object(repv obj);
 extern repv Qbytecode_error, Qjade_byte_code;
-extern repv Fjade_byte_code(repv code, repv consts, repv stkreq);
+extern repv Fjade_byte_code(repv code, repv consts, repv stkreq, repv frame);
 extern repv Fvalidate_byte_code(repv bc_major, repv bc_minor);
 extern repv Fmake_byte_code_subr(repv args);
 
+/* from macros.c */
+extern repv Fmacroexpand(repv, repv);
+
 /* from main.c */
 extern void rep_init(char *prog_name, int *argc, char ***argv,
 		     void (*sys_symbols)(void), void (*sys_usage)(void));
+extern void rep_init_from_dump(char *prog_name, int *argc, char ***argv,
+			       void (*sys_symbols)(void),
+			       void (*sys_usage)(void),
+			       char *dump_file);
+extern repv rep_load_environment (repv file);
 extern void rep_kill(void);
 extern rep_bool rep_get_option (char *option, repv *argp);
 extern rep_bool rep_on_idle(long since_last_event);
 extern rep_bool rep_handle_input_exception(repv *result_p);
+extern int rep_top_level_exit (void);
 extern void *rep_common_db;
 extern int rep_recurse_depth;
 extern rep_bool (*rep_on_idle_fun)(int since_last);
@@ -342,6 +353,69 @@ extern repv Fspace_char_p(repv);
 extern repv Fchar_upcase(repv);
 extern repv Fchar_downcase(repv);
 
+/* from numbers.c */
+extern repv rep_make_long_uint (u_long in);
+extern repv rep_make_long_int (long in);
+extern u_long rep_get_long_uint (repv in);
+extern long rep_get_long_int (repv in);
+extern repv rep_make_longlong_int (rep_long_long in);
+extern rep_long_long rep_get_longlong_int (repv in);
+extern repv rep_make_float (double in, rep_bool force);
+extern double rep_get_float (repv in);
+extern char *rep_print_number_to_string (repv obj, int radix, int prec);
+extern repv rep_number_foldl (repv args, repv (*op)(repv, repv));
+extern repv rep_number_add (repv x, repv y);
+extern repv rep_number_neg (repv x);
+extern repv rep_number_sub (repv x, repv y);
+extern repv rep_number_mul (repv x, repv y);
+extern repv rep_number_div (repv x, repv y);
+extern repv rep_number_lognot (repv x);
+extern repv rep_number_logior (repv x, repv y);
+extern repv rep_number_logxor (repv x, repv y);
+extern repv rep_number_logand (repv x, repv y);
+extern repv Feql(repv arg1, repv arg2);
+extern repv Fplus(repv);
+extern repv Fminus(repv);
+extern repv Fproduct(repv);
+extern repv Fdivide(repv);
+extern repv Fremainder(repv n1, repv n2);
+extern repv Fmod(repv n1, repv n2);
+extern repv Fquotient(repv n1, repv n2);
+extern repv Flognot(repv);
+extern repv Fnot(repv);
+extern repv Flogior(repv);
+extern repv Flogxor(repv);
+extern repv Flogand(repv);
+extern repv Fplus1(repv);
+extern repv Fsub1(repv);
+extern repv Fash(repv, repv);
+extern repv Ffloor (repv);
+extern repv Fceiling (repv);
+extern repv Ftruncate (repv);
+extern repv Fround (repv);
+extern repv Fexp (repv);
+extern repv Flog (repv);
+extern repv Fsin (repv);
+extern repv Fcos (repv);
+extern repv Ftan (repv);
+extern repv Fasin (repv);
+extern repv Facos (repv);
+extern repv Fatan (repv, repv);
+extern repv Fsqrt (repv);
+extern repv Fexpt (repv, repv);
+extern repv Fgcd (repv, repv);
+extern repv Fzerop(repv);
+extern repv Fnumberp(repv);
+extern repv Fintegerp(repv);
+extern repv Frationalp(repv);
+extern repv Frealp(repv);
+extern repv Fexactp(repv);
+extern repv Finexactp(repv);
+extern repv Fexact_to_inexact(repv);
+extern repv Finexact_to_exact(repv);
+extern repv Fnumerator(repv);
+extern repv Fdenominator(repv);
+
 /* from streams.c */
 extern repv Qformat_hooks_alist;
 extern int rep_stream_getc(repv);
@@ -371,7 +445,7 @@ extern repv rep_add_subr(rep_xsubr *);
 extern repv rep_add_const_num(repv, long);
 extern void rep_intern_static(repv *, repv);
 extern repv rep_bind_symbol(repv, repv, repv);
-extern void rep_unbind_symbols(repv);
+extern int rep_unbind_symbols(repv);
 extern repv rep_obarray;
 extern repv Qnil, Qt;
 extern repv Qvariable_documentation, Qpermanent_local;
@@ -388,7 +462,9 @@ extern repv Fclosure_function (repv funarg);
 extern repv Fset_closure_function (repv funarg, repv fun);
 extern repv Fclosurep (repv arg);
 extern repv Fsymbol_value(repv, repv);
+extern repv Fdefvar(repv);
 extern repv Fset(repv, repv);
+extern repv Fmark_symbol_defined(repv);
 extern repv Fsetplist(repv, repv);
 extern repv Fsymbol_name(repv);
 extern repv Fdefault_value(repv, repv);
@@ -399,10 +475,7 @@ extern repv Fsymbol_plist(repv);
 extern repv Fgensym(void);
 extern repv Fsymbolp(repv);
 extern repv Fsetq(repv);
-extern repv Fsetq_default(repv);
 extern repv Fmakunbound(repv);
-extern repv Flet(repv);
-extern repv Fletstar(repv);
 extern repv Fget(repv, repv);
 extern repv Fput(repv, repv, repv);
 extern repv Fapropos(repv, repv, repv);
@@ -414,6 +487,13 @@ extern repv Funtrace(repv sym);
 extern repv Vobarray(repv val);
 
 /* from values.c */
+extern repv Qafter_gc_hook;
+extern rep_cons *rep_dumped_cons_start, *rep_dumped_cons_end;
+extern rep_symbol *rep_dumped_symbols_start, *rep_dumped_symbols_end;
+extern repv rep_dumped_non_constants;
+extern int rep_guardian_type;
+extern repv rep_box_pointer (void *p);
+void *rep_unbox_pointer (repv v);
 extern void rep_register_type(u_int code, char *name,
 			      int (*compare)(repv, repv),
 			      void (*princ)(repv, repv),
@@ -443,6 +523,7 @@ extern int rep_value_cmp(repv, repv);
 extern void rep_princ_val(repv, repv);
 extern void rep_print_val(repv, repv);
 extern repv rep_null_string(void);
+extern repv rep_box_string (char *ptr, long len);
 extern repv rep_make_string(long);
 extern repv rep_string_dupn(const u_char *, long);
 extern repv rep_string_dup(const u_char *);
@@ -456,6 +537,9 @@ extern repv rep_list_3(repv, repv, repv);
 extern repv rep_list_4(repv, repv, repv, repv);
 extern repv rep_list_5(repv, repv, repv, repv, repv);
 extern repv rep_make_vector(int);
+extern repv Fmake_primitive_guardian (void);
+extern repv Fprimitive_guardian_push (repv g, repv obj);
+extern repv Fprimitive_guardian_pop (repv g);
 extern void rep_mark_static(repv *);
 extern void rep_mark_value(repv);
 extern repv Fcons(repv, repv);
