@@ -1,6 +1,6 @@
 /* symbols.c -- Lisp symbol handling
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: symbols.c,v 1.29 1999/12/07 17:21:28 john Exp $
+   $Id: symbols.c,v 1.42 2000/04/03 17:28:41 john Exp $
 
    This file is part of Jade.
 
@@ -20,6 +20,8 @@
 
 #define _GNU_SOURCE
 
+#define NDEBUG
+
 /* AIX requires this to be the first thing in the file.  */
 #include <config.h>
 #ifdef __GNUC__
@@ -45,6 +47,17 @@ char *alloca ();
 #include <stdlib.h>
 #include <assert.h>
 
+/* The number of hash buckets in each rep_obarray, this is a prime number. */
+#define rep_OBSIZE		509
+
+#define rep_SYMBOLBLK_SIZE	405		/* ~8k */
+
+/* Symbol allocation blocks */
+typedef struct rep_symbol_block_struct {
+    struct rep_symbol_block_struct *next;
+    rep_ALIGN_CELL(rep_symbol symbols[rep_SYMBOLBLK_SIZE]);
+} rep_symbol_block;
+
 /* Main storage of symbols.  */
 repv rep_obarray;
 
@@ -72,9 +85,6 @@ static rep_symbol_block *symbol_block_chain;
 static rep_symbol *symbol_freelist;
 int rep_allocated_symbols, rep_used_symbols;
 
-/* When true, warn about hiding functions when binding */
-rep_bool rep_warn_shadowing = rep_FALSE;
-
 
 /* Symbol management */
 
@@ -142,33 +152,33 @@ symbol_sweep(void)
 	}
 	sb = nxt;
     }
-#ifdef rep_DUMPED
+    /* Need to clear mark bits of dumped symbols, since they're mutable */
+    if (rep_dumped_symbols_start != rep_dumped_symbols_end)
     {
-	rep_symbol *sym;
-	for(sym = &rep_dumped_symbols_start; sym < &rep_dumped_symbols_end; sym++)
+	rep_symbol *s;
+	for(s = rep_dumped_symbols_start; s < rep_dumped_symbols_end; s++)
 	{
-	    if(!rep_GC_CELL_MARKEDP(rep_VAL(sym)))
+	    if(rep_GC_CELL_MARKEDP(rep_VAL(s)))
 	    {
-		/* Don't put this on the free list. There may
-		   still be references to it from other dumped
-		   constants (that aren't marked).. */
-	    }
-	    else
-	    {
-		rep_GC_CLR_CELL(rep_VAL(sym));
+		rep_GC_CLR_CELL(rep_VAL(s));
 		rep_used_symbols++;
 	    }
 	}
     }
-#endif /* rep_DUMPED */
 }
 
 static int
 symbol_cmp(repv v1, repv v2)
 {
     if(rep_TYPE(v1) == rep_TYPE(v2))
-	return(!(rep_SYM(v1) == rep_SYM(v2)));
-    return(1);
+    {
+	if (v1 == v2)
+	    return 0;
+	else
+	    return rep_value_cmp (rep_SYM(v1)->name, rep_SYM(v2)->name);
+    }
+    else
+	return 1;
 }
 
 static void
@@ -182,26 +192,31 @@ symbol_print(repv strm, repv obj)
 {
     u_char *s = rep_STR(rep_SYM(obj)->name);
     u_char c;
+    rep_bool all_digits = isdigit (*s);
+
     while((c = *s++))
     {
 	switch(c)
 	{
-	case ' ':
-	case '\t':
-	case '\n':
-	case '\f':
-	case '(':
-	case ')':
-	case '[':
-	case ']':
-	case '\'':
-	case '"':
-	case ';':
-	case '\\':
+	case ' ': case '\t': case '\n': case '\f':
+	case '(': case ')': case '[': case ']':
+	case '\'': case '"': case ';': case '\\':
 	case '|':
 	    rep_stream_putc(strm, (int)'\\');
+	    all_digits = rep_FALSE;
 	    break;
+
+	case '.': case '/': case 'e': case 'E':
+	    if (all_digits)
+	    {
+		rep_stream_putc (strm, (int) '\\');
+		all_digits = rep_FALSE;
+	    }
+	    break;
+
 	default:
+	    if (!isdigit (c))
+		all_digits = rep_FALSE;
 	    if(iscntrl(c))
 		rep_stream_putc(strm, (int)'\\');
 	    break;
@@ -218,8 +233,6 @@ rep_add_subr(rep_xsubr *subr)
     {
 	rep_SYM(sym)->value = rep_VAL(subr);
 	rep_SYM(sym)->car |= rep_SF_DEFVAR;
-	if(subr->car == rep_Var)
-	    rep_SYM(sym)->car |= rep_SF_SPECIAL;
     }
     return(sym);
 }
@@ -230,8 +243,9 @@ rep_add_const_num(repv name, long num)
     repv sym = Fintern(name, rep_obarray);
     if(sym)
     {
+	rep_SYM(sym)->car |= rep_SF_SPECIAL;
 	rep_SYM(sym)->value = rep_MAKE_INT(num);
-	rep_SYM(sym)->car |= rep_SF_CONSTANT | rep_SF_SPECIAL | rep_SF_DEFVAR;
+	rep_SYM(sym)->car |= rep_SF_CONSTANT;
     }
     return(sym);
 }
@@ -384,6 +398,8 @@ current environment.
     rep_funarg *f = rep_ALLOC_CELL (sizeof (rep_funarg));
     rep_data_after_gc += sizeof (rep_funarg);
     f->car = rep_Funarg;
+    if (rep_bytecode_interpreter != Fjade_byte_code)
+	f->car |= rep_FF_NO_BYTE_CODE;
     f->fun = fun;
     f->name = name;
     f->env = rep_env;
@@ -455,7 +471,19 @@ set-variable-environment ENV
     if (rep_call_stack != 0)
     {
 	if (rep_LISTP (env))
+	{
+	    repv tem;
 	    rep_call_stack->saved_env = env;
+	    tem = Fassq (Qjade_byte_code, env);
+	    if (tem && tem != Qnil)
+	    {
+		tem = rep_CDR (tem);
+		if (rep_CELL8_TYPEP (tem, rep_Subr3))
+		    rep_bytecode_interpreter = rep_SUBR4FUN (tem);
+		else
+		    rep_bytecode_interpreter = 0;
+	    }
+	}
     }
     return Qt;
 }
@@ -494,11 +522,9 @@ funarg_sweep (void)
     }
 }
 
-/* this is also in lispmach.c
-
-   Returns (SYM . VALUE) if a lexical binding. Returns t if the actual
+/* Returns (SYM . VALUE) if a lexical binding. Returns t if the actual
    value is in the symbol's function slot */
-static inline repv
+static repv
 search_environment (repv sym)
 {
     register repv env = rep_env;
@@ -507,8 +533,24 @@ search_environment (repv sym)
     return rep_CONSP(env) ? rep_CAR(env) : env;
 }
 
+/* this is also in lispmach.c */
+static inline repv
+inlined_search_special_bindings (repv sym)
+{
+    register repv env = rep_special_bindings;
+    while (env != Qnil && rep_CAAR(env) != sym)
+	env = rep_CDR(env);
+    return env != Qnil ? rep_CAR(env) : env;
+}
+
+static repv
+search_special_bindings (repv sym)
+{
+    return inlined_search_special_bindings (sym);
+}
+
 static inline int
-search_special_environment (repv sym)
+inlined_search_special_environment (repv sym)
 {
     register repv env = rep_CDR(rep_special_env);
     while (rep_CONSP(env) && rep_CAR(env) != sym)
@@ -522,68 +564,76 @@ search_special_environment (repv sym)
 	return 0;
 }
 
+static int
+search_special_environment (repv sym)
+{
+    return inlined_search_special_environment (sym);
+}
+
 
 /* Symbol binding */
 
 /* This give SYMBOL a new value, saving the old one onto the front of
-   the list OLDLIST. OLDLIST is structured like,
-     ((SYMBOL . OLDVALUE) ...)
+   the list OLDLIST. OLDLIST is structured like (NSPECIALS . NLEXICALS)
    Returns the new version of OLDLIST.   */
 repv
 rep_bind_symbol(repv oldList, repv symbol, repv newVal)
 {
-    if (rep_warn_shadowing
-	&& Fboundp (symbol) != Qnil
-	&& Ffunctionp (Fsymbol_value (symbol, Qt)) != Qnil)
-    {
-	fprintf (stderr, "warning: shadowing %s\n",
-		 rep_STR(rep_SYM(symbol)->name));
-    }
+    if (oldList == Qnil)
+	oldList = Fcons (rep_MAKE_INT(0), rep_MAKE_INT(0));
 
     if (rep_SYM(symbol)->car & rep_SF_SPECIAL)
     {
-	if (search_special_environment (symbol))
+	/* special binding */
+	if (inlined_search_special_environment (symbol))
 	{
-	    repv newbl = Fcons(Fcons(symbol, Qnil), oldList);
-	    /* Binding to buffer-local values is a recipe for disaster */
-	    rep_CDR(rep_CAR(newbl)) = Fdefault_value(symbol, Qt);
-	    Fset_default(symbol, newVal);
-	    return newbl;
+	    rep_special_bindings = Fcons (Fcons (symbol, newVal),
+					  rep_special_bindings);
+	    rep_CAR(oldList) = rep_MAKE_INT(rep_INT(rep_CAR(oldList)) + 1);
 	}
 	else
-	{
 	    Fsignal (Qvoid_value, rep_LIST_1(symbol));
-	    /* no one expects this function to fail :-( */
-	    return oldList;
-	}
     }
     else
     {
-	/* lexical binding */
+	/* lexical binding (also in lispmach.c:OP_BIND) */
 	rep_env = Fcons (Fcons (symbol, newVal), rep_env);
-	return Fcons (symbol, oldList);
+	rep_CDR(oldList) = rep_MAKE_INT(rep_INT(rep_CDR(oldList)) + 1);
     }
+    return oldList;
 }
 
-/* Undoes what the above function does.  */
-void
+/* Undoes what the above function does. Returns the number of special
+   bindings undone. */
+int
 rep_unbind_symbols(repv oldList)
 {
-    while(rep_CONSP(oldList))
+    if (oldList != Qnil)
     {
-	repv tmp = rep_CAR(oldList);
-	if (rep_CONSP(tmp))
-	{
-	    /* dynamic binding */
-	    Fset_default(rep_CAR(tmp), rep_CDR(tmp));
-	}
-	else
-	{
-	    /* lexical binding */
-	    rep_env = rep_CDR(rep_env);
-	}
-	oldList = rep_CDR(oldList);
+	register repv tem;
+	int i;
+
+	assert (rep_CONSP(oldList));
+	assert (rep_INTP(rep_CAR(oldList)));
+	assert (rep_INTP(rep_CDR(oldList)));
+
+	tem = rep_env;
+	for (i = rep_INT (rep_CDR (oldList)); i > 0; i--)
+	    tem = rep_CDR (tem);
+	rep_env = tem;
+
+	tem = rep_special_bindings;
+	for (i = rep_INT (rep_CAR (oldList)); i > 0; i--)
+	    tem = rep_CDR (tem);
+	rep_special_bindings = tem;
+
+	assert (rep_special_bindings != rep_void_value);
+	assert (rep_env != rep_void_value);
+
+	return rep_INT (rep_CAR (oldList));
     }
+    else
+	return 0;
 }
 
 
@@ -625,19 +675,22 @@ variable will be set (if necessary) not the local value.
 	}
 
 	/* Only allowed to defvar in restricted environments
-	   if the symbol hasn't yet been defvar'd */
+	   if the symbol hasn't yet been defvar'd or it's weak */
 	spec = search_special_environment (sym);
-	if (spec == 0 && (rep_SYM(sym)->car & rep_SF_DEFVAR))
+	if (spec == 0 && (rep_SYM(sym)->car & rep_SF_DEFVAR)
+	    && !(rep_SYM(sym)->car & rep_SF_WEAK))
+	{
 	    return Fsignal (Qvoid_value, rep_LIST_1(sym));	/* XXX */
+	}
 
 	/* Only set the [default] value if its not boundp or
-	   the definition is weak */
-	if(rep_NILP(tmp) || (rep_SYM(sym)->car & rep_SF_WEAK))
+	   the definition is weak and we're currently unrestricted */
+	if(rep_NILP(tmp)
+	   || ((rep_SYM(sym)->car & rep_SF_WEAK)
+	       && !(rep_SYM(sym)->car & rep_SF_WEAK_MOD)
+	       && rep_CDR(rep_special_env) == Qt))
 	{
-	    if(rep_CELL8_TYPEP(rep_SYM(sym)->value, rep_Var))
-		rep_VARFUN(rep_SYM(sym)->value)(val);
-	    else
-		rep_SYM(sym)->value = val;
+	    rep_SYM(sym)->value = val;
 	}
 
 	rep_SYM(sym)->car |= rep_SF_SPECIAL | rep_SF_DEFVAR;
@@ -705,7 +758,7 @@ values look for one of those first.
 
     if (rep_SYM(sym)->car & rep_SF_SPECIAL)
     {
-	int spec = search_special_environment (sym);
+	int spec = inlined_search_special_environment (sym);
 	/* modified-weak specials can only be accessed from an
 	   unrestricted environment */
 	if (spec < 0 || !(rep_SYM(sym)->car & rep_SF_WEAK_MOD))
@@ -713,7 +766,13 @@ values look for one of those first.
 	    if(rep_SYM(sym)->car & rep_SF_LOCAL)
 		val = (*rep_deref_local_symbol_fun)(sym);
 	    if (val == rep_void_value)
-		val = rep_SYM(sym)->value;
+	    {
+		repv tem = inlined_search_special_bindings (sym);
+		if (tem != Qnil)
+		    val = rep_CDR (tem);
+		else
+		    val = rep_SYM(sym)->value;
+	    }
 	}
     }
     else
@@ -726,13 +785,6 @@ values look for one of those first.
 	    val = rep_SYM(sym)->value;
     }
 
-    if(val && (rep_CELL8_TYPEP(val, rep_Var)))
-    {
-	val = rep_VARFUN(val)(rep_NULL);
-	if(val == rep_NULL)
-	    val = rep_void_value;
-    }
-
     if (rep_SYM(sym)->car & rep_SF_DEBUG)
 	rep_single_step_flag = rep_TRUE;
 
@@ -757,13 +809,12 @@ SYMBOL in buffers or windows which do not have their own local value.
     
     spec = search_special_environment (sym);
     if (spec < 0 || !(rep_SYM(sym)->car & rep_SF_WEAK_MOD))
-	val = rep_SYM(sym)->value;
-
-    if(val && (rep_CELL8_TYPEP(val, rep_Var)))
     {
-	val = rep_VARFUN(val)(rep_NULL);
-	if(val == rep_NULL)
-	    val = rep_void_value;
+	repv tem = search_special_bindings (sym);
+	if (tem != Qnil)
+	    val = rep_CDR (tem);
+	else
+	    val = rep_SYM(sym)->value;
     }
 
     if(no_err == Qnil && rep_VOIDP(val))
@@ -791,9 +842,11 @@ SYMBOL the buffer-local value in the current buffer is set. Returns repv.
 
     if (rep_SYM(sym)->car & rep_SF_SPECIAL)
     {
-	int spec = search_special_environment (sym);
+	int spec = inlined_search_special_environment (sym);
 	if (spec)
 	{
+	    repv tem;
+
 	    /* Not allowed to set `modified' variables unless
 	       our environment includes all variables implicitly */
 	    if (spec > 0 && rep_SYM(sym)->car & rep_SF_WEAK_MOD)
@@ -806,8 +859,9 @@ SYMBOL the buffer-local value in the current buffer is set. Returns repv.
 		    return tem;
 		/* Fall through and set the default value. */
 	    }
-	    if (rep_CELL8_TYPEP(rep_SYM(sym)->value, rep_Var))
-		rep_VARFUN(rep_SYM(sym)->value)(val);
+	    tem = inlined_search_special_bindings (sym);
+	    if (tem != Qnil)
+		rep_CDR (tem) = val;
 	    else
 		rep_SYM(sym)->value = val;
 	}
@@ -828,6 +882,30 @@ SYMBOL the buffer-local value in the current buffer is set. Returns repv.
     return val;
 }
 
+DEFUN("mark-symbol-defined", Fmark_symbol_defined,
+      Smark_symbol_defined, (repv sym), rep_Subr1) /*
+::doc:mark-symbol-defined::
+mark-symbol-defined SYMBOL
+::end:: */
+{
+    int spec;
+    rep_DECLARE1(sym, rep_SYMBOLP);
+    spec = search_special_environment (sym);
+    if (spec)
+    {
+	rep_SYM(sym)->car |= rep_SF_DEFVAR;
+	if (rep_CDR(rep_special_env) == Qt
+	    && (rep_SYM(sym)->car & rep_SF_WEAK))
+	{
+	    rep_SYM(sym)->car &= ~rep_SF_WEAK;
+	    rep_SYM(sym)->car |= rep_SF_WEAK_MOD;
+	}
+	return sym;
+    }
+    else
+	return Fsignal (Qvoid_value, rep_LIST_1(sym));	/* XXX */
+}
+
 DEFUN("set-default", Fset_default, Sset_default,
       (repv sym, repv val), rep_Subr2) /*
 ::doc:set-default::
@@ -846,11 +924,14 @@ Sets the default value of SYMBOL to VALUE, then returns VALUE.
     spec = search_special_environment (sym);
     if (spec)
     {
+	repv tem;
+
 	if (spec > 0 && rep_SYM(sym)->car & rep_SF_WEAK_MOD)
 	    return Fsignal (Qvoid_value, rep_LIST_1(sym));	/* XXX */
 
-	if(rep_CELL8_TYPEP(rep_SYM(sym)->value, rep_Var))
-	    rep_VARFUN(rep_SYM(sym)->value)(val);
+	tem = search_special_bindings (sym);
+	if (tem != Qnil)
+	    rep_CDR (tem) = val;
 	else
 	    rep_SYM(sym)->value = val;
 	rep_SYM(sym)->car |= rep_SF_SPECIAL;
@@ -895,8 +976,13 @@ default-boundp SYMBOL
 Returns t if SYMBOL has a default value.
 ::end:: */
 {
+    repv tem;
     rep_DECLARE1(sym, rep_SYMBOLP);
-    return((rep_VOIDP(rep_SYM(sym)->value)) ? Qnil : Qt);
+    tem = search_special_bindings (sym);
+    if (tem != Qnil)
+	return rep_VOIDP (rep_CDR (tem));
+    else
+	return (rep_VOIDP (rep_SYM (sym)->value)) ? Qnil : Qt;
 }
 
 DEFUN("boundp", Fboundp, Sboundp, (repv sym), rep_Subr1) /*
@@ -987,218 +1073,14 @@ end:
     return(res);
 }
 
-DEFUN("setq-default", Fsetq_default, Ssetq_default, (repv args), rep_SF) /*
-::doc:setq-default::
-setq-default { SYMBOL FORM }...
-
-Sets the default value of each SYMBOL to the value of its corresponding
-FORM evaluated, returns the value of the last evaluation. See also setq.
-::end:: */
-{
-    repv res = Qnil;
-    rep_GC_root gc_args;
-    rep_PUSHGC(gc_args, args);
-    while(rep_CONSP(args) && rep_CONSP(rep_CDR(args)) && rep_SYMBOLP(rep_CAR(args)))
-    {
-	if(!(res = Feval(rep_CAR(rep_CDR(args)))))
-	    goto end;
-	if(!Fset_default(rep_CAR(args), res))
-	{
-	    res = rep_NULL;
-	    goto end;
-	}
-	args = rep_CDR(rep_CDR(args));
-    }
-end:
-    rep_POPGC;
-    return(res);
-}
-
 DEFUN("makunbound", Fmakunbound, Smakunbound, (repv sym), rep_Subr1) /*
 ::doc:makunbound::
 makunbound SYMBOL
 
-Make SYMBOL have no value as a variable. If the variable was marked
-as being special, this status is removed.
+Make SYMBOL have no value as a variable.
 ::end:: */
 {
-    rep_DECLARE1(sym, rep_SYMBOLP);
-    if (rep_SYM(sym)->car & rep_SF_SPECIAL)
-    {
-	rep_SYM(sym)->value = rep_void_value;
-	rep_SYM(sym)->car &= ~(rep_SF_SPECIAL | rep_SF_WEAK | rep_SF_WEAK_MOD);
-    }
-    else
-    {
-	repv tem = search_environment (sym);
-	if (rep_CONSP(tem))
-	    rep_CDR(tem) = rep_void_value;
-	else if (tem == Qt)
-	    rep_SYM(sym)->value = rep_void_value;
-    }
-    return(sym);
-}
-
-DEFSTRING(no_symbol, "No symbol to bind to in let");
-
-static repv
-do_let (repv args, repv (*bind)(repv, repv, repv), void (*unbind)(repv))
-{
-    repv tmp, *store, oldvals, res = rep_NULL;
-    int numsyms = 0;
-    if(!rep_CONSP(args))
-	return rep_NULL;
-    oldvals = Qnil;
-    for(tmp = rep_CAR(args); rep_CONSP(tmp); numsyms++)
-	tmp = rep_CDR(tmp);
-    if(numsyms == 0)
-	return(Fprogn(rep_CDR(args)));
-
-    store = alloca(sizeof(repv) * numsyms);
-    if(store != NULL)
-    {
-	int i;
-	rep_GC_root gc_args;
-	rep_GC_n_roots gc_store;
-	rep_PUSHGC(gc_args, args);
-	rep_PUSHGCN(gc_store, store, 0);
-	i = 0;
-	tmp = rep_CAR(args);
-	while(rep_CONSP(tmp))
-	{
-	    if(rep_CONSP(rep_CAR(tmp)))
-	    {
-		if(!(store[i] = Fprogn(rep_CDR(rep_CAR(tmp)))))
-		{
-		    rep_POPGCN; rep_POPGC;
-		    goto end;
-		}
-	    }
-	    else
-		store[i] = Qnil;
-	    tmp = rep_CDR(tmp);
-	    i++;
-	    gc_store.count = i;
-	}
-	rep_POPGCN;
-	rep_POPGC;
-	i = 0;
-	tmp = rep_CAR(args);
-	while(rep_CONSP(tmp))
-	{
-	    repv sym;
-	    switch(rep_TYPE(rep_CAR(tmp)))
-	    {
-	    case rep_Symbol:
-		sym = rep_CAR(tmp);
-		break;
-	    case rep_Cons:
-		sym = rep_CAR(rep_CAR(tmp));
-		if(rep_SYMBOLP(sym))
-		    break;
-		/* FALL THROUGH */
-	    default:
-		Fsignal(Qerror, rep_LIST_1(rep_VAL(&no_symbol)));
-		goto end;
-	    }
-	    if(!(oldvals = bind (oldvals, sym, store[i])))
-		goto end;
-	    tmp = rep_CDR(tmp);
-	    i++;
-	}
-	rep_PUSHGC(gc_args, oldvals);
-	res = Fprogn(rep_CDR(args));
-	rep_POPGC;
-end:
-	unbind (oldvals);
-	return(res);
-    }
-    return rep_NULL;
-}
-
-static repv
-do_letstar (repv args, repv (*bind)(repv, repv, repv), void (*unbind)(repv))
-{
-    repv binds, res = rep_NULL;
-    repv oldvals = Qnil;
-    rep_GC_root gc_args, gc_oldvals;
-    if(!rep_CONSP(args))
-	return rep_NULL;
-    binds = rep_CAR(args);
-    rep_PUSHGC(gc_args, args);
-    rep_PUSHGC(gc_oldvals, oldvals);
-    while(rep_CONSP(binds))
-    {
-	if(rep_CONSP(rep_CAR(binds)))
-	{
-	    if(rep_SYMBOLP(rep_CAR(rep_CAR(binds))))
-	    {
-		repv val;
-		if(!(val = Fprogn(rep_CDR(rep_CAR(binds)))))
-		    goto error;
-		if(!(oldvals = bind (oldvals, rep_CAR(rep_CAR(binds)), val)))
-		    goto error;
-	    }
-	}
-	else
-	{
-	    if(!(oldvals = bind (oldvals, rep_CAR(binds), Qnil)))
-		goto error;
-	}
-	binds = rep_CDR(binds);
-    }
-    res = Fprogn(rep_CDR(args));
-error:
-    rep_POPGC; rep_POPGC;
-    unbind (oldvals);
-    return(res);
-}
-
-DEFUN("let", Flet, Slet, (repv args), rep_SF) /*
-::doc:let::
-let (SYMBOL-BINDINGS...) BODY...
-
-Binds temporary values to symbols while BODY is being evaluated.
-Each SYMBOL-BINDING is either a symbol, in which case that symbol is bound to
-nil, or a list. The symbol at the head of this list is bound to the progn'ed
-value of the forms making up the tail. ie,
-  (let
-      ((foo 1 2 3)
-       bar)
-    (cons foo bar))
-   => (3 . nil)
-
-All values of the new bindings are evaluated before any symbols are bound.
-::end:: */
-{
-    return do_let (args, rep_bind_symbol, rep_unbind_symbols);
-}
-
-DEFUN("let*", Fletstar, Sletstar, (repv args), rep_SF) /*
-::doc:let*::
-let* (SYMBOL-BINDINGS...) BODY...
-
-Binds temporary values to symbols while BODY is being evaluated.
-Each SYMBOL-BINDING is either a symbol, in which case that symbol is bound to
-nil, or a list. The symbol at the head of this list is bound to the progn'ed
-value of the forms making up the tail. ie,
-  (let*
-      ((foo 1 2 3)
-       bar)
-    (cons foo bar))
-   => (3 . nil)
-
-The value of each binding is evaluated just before that symbol is bound,
-this means that,
-  (setq x 'foo)
-  (let*
-      ((x 10)
-       (y x))
-    (cons x y))
-   => (10 . 10)
-::end:: */
-{
-    return do_letstar (args, rep_bind_symbol, rep_unbind_symbols);
+    return Fset (sym, rep_void_value);
 }
 
 DEFUN("get", Fget, Sget, (repv sym, repv prop), rep_Subr2) /*
@@ -1213,8 +1095,12 @@ Returns the value of SYMBOL's property PROPERTY. See `put'.
     plist = rep_SYM(sym)->prop_list;
     while(rep_CONSP(plist) && rep_CONSP(rep_CDR(plist)))
     {
-	if(rep_CAR(plist) == prop)
+	if(rep_CAR(plist) == prop
+	   || (!rep_SYMBOLP(prop)
+	       && rep_value_cmp (rep_CAR(plist), prop) == 0))
+	{
 	    return(rep_CAR(rep_CDR(plist)));
+	}
 	plist = rep_CDR(rep_CDR(plist));
     }
     return(Qnil);
@@ -1238,7 +1124,9 @@ retrieved with the `get' function.
     plist = rep_SYM(sym)->prop_list;
     while(rep_CONSP(plist) && rep_CONSP(rep_CDR(plist)))
     {
-	if(rep_CAR(plist) == prop)
+	if(rep_CAR(plist) == prop
+	   || (!rep_SYMBOLP(prop)
+	       && rep_value_cmp (rep_CAR(plist), prop) == 0))
 	{
 	    if(!rep_CONS_WRITABLE_P(rep_CDR(plist)))
 	    {
@@ -1389,14 +1277,17 @@ Cancel the effect of (trace SYMBOL).
     return(sym);
 }
 
-DEFUN("obarray", Vobarray, Sobarray, (repv val), rep_Var) /*
+DEFUN("obarray", Fobarray, Sobarray, (repv val), rep_Subr1) /*
 ::doc:obarray::
-The obarray used by the Lisp reader.
+obarray [NEW-VALUE]
 ::end:: */
 {
-    if(val && rep_VECTORP(val))
+    if(val != Qnil)
+    {
+	rep_DECLARE1(val, rep_VECTORP);
 	rep_obarray = val;
-    return(rep_obarray);
+    }
+    return rep_obarray;
 }
 
 int
@@ -1432,6 +1323,12 @@ rep_symbols_init(void)
     rep_SYM(Qt)->value = Qt;
     rep_SYM(Qt)->car |= rep_SF_CONSTANT;
 
+    rep_USE_DEFAULT_ENV;
+    rep_special_bindings = Qnil;
+    rep_mark_static (&rep_env);
+    rep_mark_static (&rep_special_env);
+    rep_mark_static (&rep_special_bindings);
+
     rep_INTERN(documentation);
     rep_INTERN(permanent_local);
     rep_ADD_SUBR(Smake_symbol);
@@ -1450,6 +1347,7 @@ rep_symbols_init(void)
     rep_ADD_SUBR(Sdefvar);
     rep_ADD_SUBR(Ssymbol_value);
     rep_ADD_SUBR_INT(Sset);
+    rep_ADD_SUBR(Smark_symbol_defined);
     rep_ADD_SUBR(Ssetplist);
     rep_ADD_SUBR(Ssymbol_name);
     rep_ADD_SUBR(Sdefault_value);
@@ -1460,10 +1358,7 @@ rep_symbols_init(void)
     rep_ADD_SUBR(Sgensym);
     rep_ADD_SUBR(Ssymbolp);
     rep_ADD_SUBR(Ssetq);
-    rep_ADD_SUBR(Ssetq_default);
     rep_ADD_SUBR(Smakunbound);
-    rep_ADD_SUBR(Slet);
-    rep_ADD_SUBR(Sletstar);
     rep_ADD_SUBR(Sget);
     rep_ADD_SUBR(Sput);
     rep_ADD_SUBR(Sapropos);
