@@ -1,6 +1,6 @@
 /* timers.c -- call a function after a period of time has passed
    Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
-   $Id: timers.c,v 1.6 1999/11/29 18:37:58 john Exp $
+   $Id: timers.c,v 1.9 2000/02/28 08:29:28 john Exp $
 
    This file is part of librep.
 
@@ -20,6 +20,24 @@
 
 #define _GNU_SOURCE
 
+/* AIX requires this to be the first thing in the file.  */
+#include <config.h>
+#ifdef __GNUC__
+# define alloca __builtin_alloca
+#else
+# if HAVE_ALLOCA_H
+#  include <alloca.h>
+# else
+#  ifdef _AIX
+ #pragma alloca
+#  else
+#   ifndef alloca /* predefined by HP cc +Olibcalls */
+char *alloca ();
+#   endif
+#  endif
+# endif
+#endif
+
 #include "repint.h"
 #include <signal.h>
 #include <time.h>
@@ -337,17 +355,12 @@ timer_print (repv stream, repv arg)
 
 /* DL hooks */
 
-rep_xsubr *rep_dl_subrs[] = { &Smake_timer, &Sdelete_timer, &Sset_timer, 0 };
-repv rep_dl_feature;
-
 repv
 rep_dl_init (void)
 {
     timer_type = rep_register_new_type ("timer", 0, timer_print, timer_print,
 					timer_sweep, timer_mark,
 					timer_mark_active, 0, 0, 0, 0, 0, 0);
-    rep_INTERN (timers);
-    rep_dl_feature = Qtimers;
     pipe (pipe_fds);
     rep_register_input_fd (pipe_fds[0], timer_fd_handler);
 #ifdef HAVE_UNIX
@@ -356,7 +369,13 @@ rep_dl_init (void)
     sigemptyset (&alrm_sigset);
     sigaddset (&alrm_sigset, SIGALRM);
     rep_sig_restart (SIGALRM, rep_TRUE);
-    return Qt;
+
+    rep_ADD_SUBR(Smake_timer);
+    rep_ADD_SUBR(Sdelete_timer);
+    rep_ADD_SUBR(Sset_timer);
+
+    rep_INTERN (timers);
+    return Qtimers;
 }
 
 void
