@@ -1,6 +1,6 @@
 /* unix_dl.c -- Dynamic loading of C modules
    Copyright (C) 1998 John Harper <john@dcs.warwick.ac.uk>
-   $Id: unix_dl.c,v 1.18 1999/11/25 23:23:13 john Exp $
+   $Id: unix_dl.c,v 1.26 2000/03/21 17:28:47 john Exp $
 
    This file is part of Jade.
 
@@ -20,10 +20,40 @@
 
 #define _GNU_SOURCE
 
+/* AIX requires this to be the first thing in the file.  */
+#include <config.h>
+#ifdef __GNUC__
+# define alloca __builtin_alloca
+#else
+# if HAVE_ALLOCA_H
+#  include <alloca.h>
+# else
+#  ifdef _AIX
+ #pragma alloca
+#  else
+#   ifndef alloca /* predefined by HP cc +Olibcalls */
+char *alloca ();
+#   endif
+#  endif
+# endif
+#endif
+
 #include "repint.h"
 #include <assert.h>
 #include <string.h>
 
+/* we define some extensions to the libtool .la file. As well as using
+   the dlname entry to find the .so file to open, we also look for:
+
+     rep_open_globally=[yes|no]
+
+	whether or not to open with RTLD_GLOBAL
+
+     rep_requires='FEATURES...'
+
+	FEATURES is space separated list of feature symbols.
+	Each of which must be provided by a dl object. */
+
 #ifdef HAVE_DYNAMIC_LOADING
 
 #if defined (HAVE_DLFCN_H)
@@ -37,6 +67,20 @@
 #   define RTLD_LAZY 1
 #  endif
 # endif
+# if ! defined (RTLD_GLOBAL)
+#  if defined (DL_GLOBAL)
+#   define RTLD_GLOBAL DL_GLOBAL
+#  else
+#   define RTLD_GLOBAL 0
+#  endif
+# endif
+# if ! defined (RTLD_LOCAL)
+#  if defined (DL_LOCAL)
+#   define RTLD_LOCAL DL_LOCAL
+#  else
+#   define RTLD_LOCAL 0
+#  endif
+# endif
 # if ! defined (RTLD_NOW)
 #  if defined (DL_NOW)
 #   define RTLD_NOW DL_NOW
@@ -44,6 +88,10 @@
 #   define RTLD_NOW 0
 #  endif
 # endif
+# if defined (BROKEN_RTLD_GLOBAL)
+#  undef RTLD_GLOBAL
+#  define RTLD_GLOBAL 0
+# endif
 
 #elif defined (HAVE_DL_H)
 # include <dl.h>
@@ -95,14 +143,10 @@ static void *
 x_dlsym (void *handle, char *sym)
 {
     void *ptr = 0;
-    char *tem = rep_alloc (strlen(sym) + 2);
-    if (tem != 0)
-    {
-	tem[0] = '_';
-	strcpy (tem + 1, sym);
-	ptr = dlsym (handle, tem);
-	free (tem);
-    }
+    char *tem = alloca (strlen(sym) + 2);
+    tem[0] = '_';
+    strcpy (tem + 1, sym);
+    ptr = dlsym (handle, tem);
     return ptr;
 }
 #endif
@@ -126,16 +170,47 @@ static struct dl_lib_info *
 find_dl_by_feature(repv feature)
 {
     struct dl_lib_info *x = dl_list;
-    assert(rep_SYMBOLP(feature));
+    assert (rep_STRINGP(feature));
     while(x != 0)
     {
-	if(x->feature_sym == feature)
+	if(rep_SYMBOLP(x->feature_sym)
+	   && strcmp (rep_STR(rep_SYM(x->feature_sym)->name),
+		      rep_STR(feature)) == 0)
+	{
 	    return x;
+	}
 	x = x->next;
     }
     return 0;
 }
 
+static rep_bool
+load_requires (char *ptr)
+{
+    ptr += strspn (ptr, " \t");
+    while (*ptr != 0)
+    {
+	char *end = ptr + strcspn (ptr, " \t");
+	repv name = rep_string_dupn (ptr, end - ptr);
+	if (find_dl_by_feature (name) == 0)
+	{
+	    if (Fload (name, Qnil, Qnil, Qnil, Qnil) == rep_NULL)
+		return rep_FALSE;
+	}
+	ptr = end + strspn (end, " \t");
+    }
+    return rep_TRUE;
+}
+
+static void
+signal_error (char *msg)
+{
+    if (Qerror != 0)
+	Fsignal (Qerror, rep_LIST_1 (rep_string_dup (msg)));
+    else
+	fprintf (stderr, "error: %s\n", msg);
+}
+
 void *
 rep_open_dl_library(repv file_name)
 {
@@ -148,6 +223,7 @@ rep_open_dl_library(repv file_name)
 
 	char buf[256];
 	char *dlname = 0;
+	rep_bool open_globally = rep_FALSE;
 	FILE *fh = fopen(rep_STR(file_name), "r");
 	if (fh == 0)
 	{
@@ -161,49 +237,79 @@ rep_open_dl_library(repv file_name)
 		char *ptr = buf + sizeof("dlname='") - 1;
 		u_char *base;
 		char *end = strchr(ptr, '\'');
-		if (end == 0)
-		    break;
-		*end = 0;
-		base = strrchr(rep_STR(file_name), '/');
-		if (base == 0)
-		    dlname = ptr;
-		else
+		if (end != 0)
 		{
-		    char tem[256];
-		    base++;
-		    memcpy(tem, rep_STR(file_name), base - rep_STR(file_name));
-		    strcpy(tem + (base - rep_STR(file_name)), ptr);
-		    strcpy(buf, tem);
-		    dlname = buf;
+		    *end = 0;
+		    base = strrchr(rep_STR(file_name), '/');
+		    if (base == 0)
+		    {
+			dlname = alloca (strlen (ptr) + 1);
+			strcpy (dlname, ptr);
+		    }
+		    else
+		    {
+			base++;
+			dlname = alloca (strlen(ptr) +
+					 base - rep_STR(file_name) + 1);
+			memcpy(dlname, rep_STR(file_name),
+			       base - rep_STR(file_name));
+			strcpy(dlname + (base - rep_STR(file_name)), ptr);
+		    }
+		}
+	    }
+	    else if (strncmp("rep_open_globally=", buf,
+			     sizeof("rep_open_globally=") - 1) == 0)
+	    {
+		char *ptr = buf + sizeof ("rep_open_globally=") - 1;
+		if (strncmp ("yes", ptr, 3) == 0)
+		    open_globally = rep_TRUE;
+	    }
+	    else if (strncmp("rep_requires='", buf,
+			     sizeof ("rep_requires='") - 1) == 0)
+	    {
+		char *ptr = buf + sizeof ("rep_requires='") - 1;
+		char *end = strchr (ptr, '\'');
+		if (end != 0)
+		{
+		    rep_GC_root gc_file_name;
+		    rep_bool success;
+		    char *string = alloca (end - ptr + 1);
+		    memcpy (string, ptr, end - ptr);
+		    string[end - ptr] = 0;
+		    rep_PUSHGC (gc_file_name, file_name);
+		    success = load_requires (string);
+		    rep_POPGC;
+		    if (!success)
+			goto out;
 		}
-		break;
 	    }
 	}
 	fclose(fh);
 	if (!dlname)
-	{
-	    Fsignal(Qerror,
-		       rep_LIST_2(rep_string_dup("Can't find dlname in .la object"),
-			      file_name));
-	}
+	    signal_error ("Can't find dlname in .la object");
 	else
 	{
 	    rep_xsubr **functions;
 	    repv (*init_func)(repv);
 	    repv *feature_sym;
 	    void *handle;
+	    rep_bool relocate_now = rep_FALSE;
+	    if (Qdl_load_reloc_now
+		&& Fsymbol_value (Qdl_load_reloc_now, Qt) != Qnil)
+	    {
+		relocate_now = rep_TRUE;
+	    }
 
 #if defined (HAVE_DLOPEN)
 	    handle = dlopen(dlname,
-			    rep_SYM(Qdl_load_reloc_now)->value == Qnil
-			    ? RTLD_LAZY : RTLD_NOW);
+			    (relocate_now ? RTLD_NOW : RTLD_LAZY)
+			    | (open_globally ? RTLD_GLOBAL : RTLD_LOCAL));
 #elif defined (HAVE_SHL_LOAD)
+	    /* XXX how do we open these locally/globally? */
 	    handle = shl_load (dlname,
-			       (rep_SYM(Qdl_load_reloc_now)->value == Qnil
-				? BIND_DEFERRED : BIND_IMMEDIATE)
+			       (relocate_now ? BIND_IMMEDIATE : BIND_DEFERRED)
 			       | BIND_NONFATAL | DYNAMIC_PATH, 0L);
 #endif
-
 	    if(handle == 0)
 	    {
 		char *err;
@@ -213,7 +319,7 @@ rep_open_dl_library(repv file_name)
 		err = "unknown dl error";
 #endif
 		if(err != 0)
-		    Fsignal(Qerror, rep_LIST_1(rep_string_dup(err)));
+		    signal_error (err);
 		return 0;
 	    }
 	    x = rep_alloc(sizeof(struct dl_lib_info));
@@ -225,18 +331,32 @@ rep_open_dl_library(repv file_name)
 	    }
 	    x->file_name = file_name;
 	    x->handle = handle;
+	    x->feature_sym = Qnil;
+
+	    x->next = dl_list;
+	    dl_list = x;
 
 	    init_func = x_dlsym(handle, "rep_dl_init");
 	    if(init_func != 0)
 	    {
 		repv ret = init_func(file_name);
-		if(ret == rep_NULL)
+		if(Qnil != rep_NULL			/* initialising */
+		   && (ret == rep_NULL || ret == Qnil))
 		{
 		    /* error. abort abort.. */
+		    struct dl_lib_info **ptr;
+		    for (ptr = &dl_list; *ptr != 0; ptr = &((*ptr)->next))
+		    {
+			if (*ptr == x)
+			    *ptr = x->next;
+			break;
+		    }		    
 		    rep_free(x);
 		    dlclose(handle);
 		    return 0;
 		}
+		else if (ret && rep_SYMBOLP(ret) && ret != Qt)
+		    x->feature_sym = ret;
 	    }
 
 	    feature_sym = x_dlsym (handle, "rep_dl_feature");
@@ -245,8 +365,6 @@ rep_open_dl_library(repv file_name)
 	    {
 		x->feature_sym = *feature_sym;
 	    }
-	    else
-		x->feature_sym = Qnil;
 
 	    functions = x_dlsym(handle, "rep_dl_subrs");
 	    if(functions != 0)
@@ -258,13 +376,11 @@ rep_open_dl_library(repv file_name)
 		}
 	    }
 
-	    x->next = dl_list;
-	    dl_list = x;
-
 	    if (x->feature_sym != Qnil)
-		rep_call_lisp1 (Fsymbol_value (Qprovide, Qt), x->feature_sym);
+		Fprovide (x->feature_sym);
 	}
     }
+out:
     return x;
 }
 
@@ -306,7 +422,9 @@ rep_kill_dl_libraries(void)
 void *
 rep_find_dl_symbol (repv feature, char *symbol)
 {
-    struct dl_lib_info *x = find_dl_by_feature (feature);
+    struct dl_lib_info *x;
+    assert(rep_SYMBOLP(feature));
+    x = find_dl_by_feature (rep_SYM(feature)->name);
     if (x != 0)
 	return x_dlsym (x->handle, symbol);
     else
