@@ -1,6 +1,6 @@
 /* unix_main.c -- Miscellaneous functions for Unix
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: unix_main.c,v 1.49 1999/11/29 18:37:58 john Exp $
+   $Id: unix_main.c,v 1.57 2000/02/08 14:16:10 john Exp $
 
    This file is part of Jade.
 
@@ -324,29 +324,29 @@ rep_unix_set_fd_cloexec(int fd)
 void
 rep_sig_restart(int sig, rep_bool flag)
 {
+#if defined (HAVE_SIGINTERRUPT)
+    siginterrupt (sig, !flag);
+#else
     struct sigaction act;
     sigaction (sig, 0, &act);
     if(flag)
     {
-#ifdef SA_RESTART
+# if defined (SA_RESTART)
 	act.sa_flags |= SA_RESTART;
-#else
-# ifdef SA_INTERRUPT
+# elif defined (SA_INTERRUPT)
 	act.sa_flags &= ~SA_INTERRUPT;
 # endif
-#endif
     }
     else
     {
-#ifdef SA_RESTART
+# if defined (SA_RESTART)
 	act.sa_flags &= ~SA_RESTART;
-#else
-# ifdef SA_INTERRUPT
+# elif defined (SA_INTERRUPT)
 	act.sa_flags |= SA_INTERRUPT;
 # endif
-#endif
     }
     sigaction(sig, &act, 0);
+#endif /* !HAVE_SIGINTERRUPT */
 }
 
 /* Wait for input for no longer than TIMEOUT-MSECS for input fds defined
@@ -389,10 +389,13 @@ wait_for_input(fd_set *inputs, u_long timeout_msecs)
        interrupt between each call to select. */
     do {
 	struct timeval timeout;
+	u_long max_sleep = rep_max_sleep_for ();
 	u_long this_timeout_msecs = MIN (timeout_msecs,
 					 rep_input_timeout_secs * 1000);
-	timeout.tv_sec = this_timeout_msecs / 1000;
-	timeout.tv_usec = (this_timeout_msecs % 1000) * 1000;
+	u_long actual_timeout_msecs = MIN (this_timeout_msecs, max_sleep);
+
+	timeout.tv_sec = actual_timeout_msecs / 1000;
+	timeout.tv_usec = (actual_timeout_msecs % 1000) * 1000;
 	memcpy (&copy, inputs, sizeof (copy));
 
 	/* Dont test for interrupts before the first call to select() */
@@ -411,6 +414,12 @@ wait_for_input(fd_set *inputs, u_long timeout_msecs)
 	rep_sig_restart(SIGALRM, rep_TRUE);
 	rep_sig_restart(SIGCHLD, rep_TRUE);
 
+	if (ready == 0 && actual_timeout_msecs < this_timeout_msecs)
+	{
+	    Fthread_suspend (Qnil, rep_MAKE_INT (this_timeout_msecs
+						 - actual_timeout_msecs));
+	}
+	
 	timeout_msecs -= this_timeout_msecs;
     } while (ready == 0 && timeout_msecs > 0);
 
@@ -721,16 +730,32 @@ fatal_signal_handler(int sig)
 static RETSIGTYPE
 interrupt_signal_handler(int sig)
 {
-    rep_throw_value = rep_int_cell;
-    signal(SIGINT, interrupt_signal_handler);
+    if (rep_throw_value == rep_int_cell)
+    {
+	signal (sig, SIG_DFL);
+	raise (sig);
+    }
+    else
+    {
+	rep_throw_value = rep_int_cell;
+	signal (sig, interrupt_signal_handler);
+    }
 }
 
 /* Invoked by trappable termination signals */
 static RETSIGTYPE
 termination_signal_handler(int sig)
 {
-    rep_throw_value = rep_term_cell;
-    signal(sig, termination_signal_handler);
+    if (rep_throw_value == rep_term_cell)
+    {
+	signal (sig, SIG_DFL);
+	raise (sig);
+    }
+    else
+    {
+	rep_throw_value = rep_term_cell;
+	signal (sig, termination_signal_handler);
+    }
 }
 
 /* Invoked by SIGUSR1 or SIGUSR2 */
@@ -836,7 +861,7 @@ rep_sys_os_init(void)
     ptr = environ;
     while(*ptr != 0)
 	env = Fcons(rep_string_dup(*ptr++), env);
-    rep_SYM(Qprocess_environment)->value = env;
+    Fset (Qprocess_environment, env);
 
 #ifdef DEBUG_SYS_ALLOC
     rep_ADD_SUBR(Sunix_print_allocations);
