@@ -1,6 +1,6 @@
 /* values.c -- Handling of Lisp data (includes garbage collection)
    Copyright (C) 1993, 1994 John Harper <john@dcs.warwick.ac.uk>
-   $Id: values.c,v 1.39 1999/11/29 18:37:59 john Exp $
+   $Id: values.c,v 1.52 2000/04/06 21:47:38 john Exp $
 
    This file is part of Jade.
 
@@ -20,6 +20,24 @@
 
 #define _GNU_SOURCE
 
+/* AIX requires this to be the first thing in the file.  */
+#include <config.h>
+#ifdef __GNUC__
+# define alloca __builtin_alloca
+#else
+# if HAVE_ALLOCA_H
+#  include <alloca.h>
+# else
+#  ifdef _AIX
+ #pragma alloca
+#  else
+#   ifndef alloca /* predefined by HP cc +Olibcalls */
+char *alloca ();
+#   endif
+#  endif
+# endif
+#endif
+
 #include "repint.h"
 #include <string.h>
 #include <stdlib.h>
@@ -33,10 +51,44 @@
 
 /* #define DEBUG_GC */
 
+#define rep_CONSBLK_SIZE	510		/* ~4k */
+#define rep_STRINGBLK_SIZE	510		/* ~4k */
+
+/* Structure of cons allocation blocks */
+typedef struct rep_cons_block_struct {
+    union {
+	struct rep_cons_block_struct *p;
+	/* ensure that the following cons cell is aligned to at
+	   least sizeof (rep_cons) (for the dcache) */
+	rep_cons dummy;
+    } next;
+    rep_cons cons[rep_CONSBLK_SIZE];
+} rep_cons_block;
+
+/* Structure of string header allocation blocks */
+typedef struct rep_string_block_struct {
+    union {
+	struct rep_string_block_struct *p;
+	/* ensure that the following cons cell is aligned to at
+	   least sizeof (rep_string) (for the dcache) */
+	rep_string dummy;
+    } next;
+    rep_string data[rep_STRINGBLK_SIZE];
+} rep_string_block;
+
 #ifdef DEBUG_GC
 static int debug_gc = 0;
 #endif
 
+/* Dumped data */
+rep_cons *rep_dumped_cons_start, *rep_dumped_cons_end;
+rep_symbol *rep_dumped_symbols_start, *rep_dumped_symbols_end;
+repv rep_dumped_non_constants;
+
+int rep_guardian_type;
+
+DEFSYM(after_gc_hook, "after-gc-hook");
+
 
 /* Type handling */
 
@@ -69,7 +121,7 @@ rep_register_type(u_int code, char *name,
     }
     t->code = code;
     t->name = name;
-    t->compare = compare;
+    t->compare = compare ? compare : rep_ptr_cmp;
     t->princ = princ;
     t->print = print;
     t->sweep = sweep;
@@ -175,8 +227,9 @@ rep_type_cmp(repv val1, repv val2)
 
 /* Strings */
 
-static rep_string *strings;
-static int allocated_strings, allocated_string_bytes;
+static rep_string_block *string_block_chain;
+static rep_string *string_freelist;
+static int allocated_strings, used_strings, allocated_string_bytes;
 
 DEFSTRING(null_string_const, "");
 
@@ -188,30 +241,56 @@ rep_null_string(void)
 
 DEFSTRING(string_overflow, "String too long");
 
+/* PTR should have been allocated using rep_alloc or malloc. Ownership
+   of its memory passes to the lisp system. LEN _doesn't_ include the zero
+   terminator */
+repv
+rep_box_string (char *ptr, long len)
+{
+    rep_string *str;
+
+    if(len > rep_MAX_STRING)
+	return Fsignal(Qerror, rep_LIST_1(rep_VAL(&string_overflow)));
+
+    /* find a string header */
+    str = string_freelist;
+    if(str == NULL)
+    {
+	rep_string_block *cb;
+	cb = rep_alloc(sizeof(rep_string_block));
+	if(cb != NULL)
+	{
+	    int i;
+	    allocated_strings += rep_STRINGBLK_SIZE;
+	    cb->next.p = string_block_chain;
+	    string_block_chain = cb;
+	    for(i = 0; i < (rep_STRINGBLK_SIZE - 1); i++)
+		cb->data[i].car = rep_VAL(&cb->data[i + 1]);
+	    cb->data[i].car = 0;
+	    string_freelist = cb->data;
+	}
+	else
+	    return rep_mem_error();
+	str = string_freelist;
+    }
+    string_freelist = rep_STRING(str->car);
+    used_strings++;
+    rep_data_after_gc += sizeof(rep_string);
+
+    str->car = rep_MAKE_STRING_CAR (len);
+    rep_data_after_gc += len;
+    str->data = ptr;
+    return rep_VAL (str);
+}
+
 /* Return a string object with room for exactly LEN characters. No extra
    byte is allocated for a zero terminator; do this manually if required. */
 repv
 rep_make_string(long len)
 {
-    rep_string *str;
-    int memlen;
-
-    if(len > rep_MAX_STRING)
-	return Fsignal(Qerror, rep_LIST_1(rep_VAL(&string_overflow)));
-
-    memlen = rep_DSTRING_SIZE(len);
-    str = rep_ALLOC_CELL(memlen);
-    if(str != NULL)
-    {
-	str->car = rep_MAKE_STRING_CAR(len - 1);
-	str->next = strings;
-	str->data = ((char *)str) + sizeof (rep_string);
-	strings = str;
-	allocated_strings++;
-	allocated_string_bytes += memlen;
-	rep_data_after_gc += memlen;
-	return rep_VAL(str);
-    }
+    char *data = rep_alloc (len);
+    if(data != NULL)
+	return rep_box_string (data, len - 1);
     else
 	return rep_NULL;
 }
@@ -278,24 +357,57 @@ string_cmp(repv v1, repv v2)
 static void
 string_sweep(void)
 {
-    rep_string *x = strings;
-    strings = 0;
-    allocated_strings = 0;
+    rep_string_block *cb = string_block_chain;
+    string_block_chain = NULL;
+    string_freelist = NULL;
+    used_strings = 0;
     allocated_string_bytes = 0;
-    while(x != 0)
+    while(cb != NULL)
     {
-	rep_string *next = x->next;
-	if(rep_GC_CELL_MARKEDP(rep_VAL(x)))
+	rep_string_block *nxt = cb->next.p;
+	rep_string *newfree = NULL, *newfreetail = NULL, *this;
+	int i, newused = 0;
+	for(i = 0, this = cb->data; i < rep_STRINGBLK_SIZE; i++, this++)
 	{
-	    rep_GC_CLR_CELL(rep_VAL(x));
-	    x->next = strings;
-	    strings = x;
-	    allocated_strings++;
-	    allocated_string_bytes += rep_DSTRING_SIZE(rep_STRING_LEN(rep_VAL(x)));
+	    /* if on the freelist then the CELL_IS_8 bit
+	       will be unset (since the pointer is long aligned) */
+	    if(rep_CELL_CONS_P(rep_VAL(this))
+	       || !rep_GC_CELL_MARKEDP(rep_VAL(this)))
+	    {
+		if(!newfreetail)
+		    newfreetail = this;
+		if (!rep_CELL_CONS_P(rep_VAL(this)))
+		    rep_free (this->data);
+		this->car = rep_VAL(newfree);
+		newfree = this;
+	    }
+	    else
+	    {
+		rep_GC_CLR_CELL(rep_VAL(this));
+		allocated_string_bytes += rep_STRING_LEN(rep_VAL(this));
+		newused++;
+	    }
+	}
+	if(newused == 0)
+	{
+	    /* Whole block is unused, get rid of it.  */
+	    rep_free(cb);
+	    allocated_strings -= rep_STRINGBLK_SIZE;
 	}
 	else
-	    rep_free(x);
-	x = next;
+	{
+	    if(newfreetail != NULL)
+	    {
+		/* Link this mini-freelist onto the main one.  */
+		newfreetail->car = rep_VAL(string_freelist);
+		string_freelist = newfree;
+		used_strings += newused;
+	    }
+	    /* Have to rebuild the block chain as well.  */
+	    cb->next.p = string_block_chain;
+	    string_block_chain = cb;
+	}
+	cb = nxt;
     }
 }
 
@@ -315,15 +427,6 @@ rep_set_string_len(repv str, long len)
 
 /* Misc */
 
-static int
-number_cmp(repv v1, repv v2)
-{
-    if(rep_TYPE(v1) == rep_TYPE(v2))
-	return rep_INT(v1) - rep_INT(v2);
-    else
-	return 1;
-}
-
 int
 rep_ptr_cmp(repv v1, repv v2)
 {
@@ -333,6 +436,40 @@ rep_ptr_cmp(repv v1, repv v2)
 	return 1;
 }
 
+repv
+rep_box_pointer (void *p)
+{
+    unsigned rep_PTR_SIZED_INT low;
+    low = (unsigned rep_PTR_SIZED_INT)p;
+    if (low <= rep_LISP_MAX_INT)
+	return rep_MAKE_INT (low);
+    else
+    {
+	int i;
+	unsigned rep_PTR_SIZED_INT high = (unsigned rep_PTR_SIZED_INT)p;
+	for (i = rep_PTR_SIZED_INT_BITS / 2; i < rep_PTR_SIZED_INT_BITS; i++)
+	    low &= ~(1 << i);
+	high = high >> (rep_PTR_SIZED_INT_BITS/2);
+	return Fcons (rep_MAKE_INT(high), rep_MAKE_INT(low));
+    }
+}
+
+void *
+rep_unbox_pointer (repv v)
+{
+    if (rep_INTP(v))
+	return (void *) rep_INT(v);
+    else if (rep_CONSP(v))
+    {
+	unsigned rep_PTR_SIZED_INT low, high;
+	low = rep_INT(rep_CDR(v));
+	high = rep_INT(rep_CAR(v));
+	return (void *) (low | high << (rep_PTR_SIZED_INT_BITS/2));
+    }
+    else
+	return 0;
+}
+
 
 /* Cons */
 
@@ -357,7 +494,7 @@ Returns a new cons-cell with car CAR and cdr CDR.
 	{
 	    int i;
 	    allocated_cons += rep_CONSBLK_SIZE;
-	    cb->next = cons_block_chain;
+	    cb->next.p = cons_block_chain;
 	    cons_block_chain = cb;
 	    for(i = 0; i < (rep_CONSBLK_SIZE - 1); i++)
 		cb->cons[i].cdr = rep_CONS_VAL(&cb->cons[i + 1]);
@@ -387,50 +524,26 @@ rep_cons_free(repv cn)
 static void
 cons_sweep(void)
 {
-    rep_cons_block *cb = cons_block_chain;
-    cons_block_chain = NULL;
-    cons_freelist = NULL;
+    rep_cons_block *cb;
+    cons_freelist = 0;
     used_cons = 0;
-    while(cb != NULL)
+    for (cb = cons_block_chain; cb != 0; cb = cb->next.p)
     {
-	rep_cons_block *nxt = cb->next;
-	rep_cons *newfree = NULL, *newfreetail = NULL, *this;
-	int i, newused = 0;
-	for(i = 0, this = cb->cons; i < rep_CONSBLK_SIZE; i++, this++)
+	int i;
+	rep_cons *this = cb->cons;
+	for (i = 0; i < rep_CONSBLK_SIZE; i++, this++)
 	{
-	    if(!rep_GC_CONS_MARKEDP(rep_CONS_VAL(this)))
+	    if (!rep_GC_CONS_MARKEDP (rep_CONS_VAL (this)))
 	    {
-		if(!newfreetail)
-		    newfreetail = this;
-		this->cdr = rep_CONS_VAL(newfree);
-		newfree = this;
+		this->cdr = rep_CONS_VAL (cons_freelist);
+		cons_freelist = rep_CONS (this);
 	    }
 	    else
 	    {
-		rep_GC_CLR_CONS(rep_CONS_VAL(this));
-		newused++;
+		rep_GC_CLR_CONS (rep_CONS_VAL (this));
+		used_cons++;
 	    }
 	}
-	if(newused == 0)
-	{
-	    /* Whole ConsBlk unused, lets get rid of it.  */
-	    rep_free(cb);
-	    allocated_cons -= rep_CONSBLK_SIZE;
-	}
-	else
-	{
-	    if(newfreetail != NULL)
-	    {
-		/* Link this mini-freelist onto the main one.  */
-		newfreetail->cdr = rep_CONS_VAL(cons_freelist);
-		cons_freelist = newfree;
-		used_cons += newused;
-	    }
-	    /* Have to rebuild the ConsBlk chain as well.  */
-	    cb->next = cons_block_chain;
-	    cons_block_chain = cb;
-	}
-	cb = nxt;
     }
 }
 
@@ -535,6 +648,127 @@ vector_cmp(repv v1, repv v2)
     return rc;
 }
 
+
+/* Guardians */
+
+static rep_guardian *guardians;
+
+DEFUN("make-primitive-guardian", Fmake_primitive_guardian,
+      Smake_primitive_guardian, (void), rep_Subr0)
+{
+    rep_guardian *g = rep_ALLOC_CELL (sizeof (rep_guardian));
+    rep_data_after_gc += sizeof (rep_guardian);
+    g->car = rep_guardian_type;
+    g->accessible = Qnil;
+    g->inaccessible = Qnil;
+    g->next = guardians;
+    guardians = g;
+    return rep_VAL(g);
+}
+
+DEFUN("primitive-guardian-push", Fprimitive_guardian_push,
+       Sprimitive_guardian_push, (repv g, repv obj), rep_Subr2)
+{
+    rep_DECLARE1 (g, rep_GUARDIANP);
+    rep_GUARDIAN(g)->accessible = Fcons (obj, rep_GUARDIAN(g)->accessible);
+    return g;
+}
+
+DEFUN("primitive-guardian-pop", Fprimitive_guardian_pop,
+      Sprimitive_guardian_pop, (repv g), rep_Subr1)
+{
+    rep_DECLARE1 (g, rep_GUARDIANP);
+    if (rep_GUARDIAN(g)->inaccessible != Qnil)
+    {
+	repv ret = rep_CAR (rep_GUARDIAN(g)->inaccessible);
+	rep_GUARDIAN(g)->inaccessible = rep_CDR (rep_GUARDIAN(g)->inaccessible);
+	return ret;
+    }
+    else
+	return Qnil;
+}
+
+static void
+mark_guardian (repv g)
+{
+    /* accessible list is marked by run_guardians */
+    rep_MARKVAL (rep_GUARDIAN(g)->inaccessible);
+}
+
+static void
+run_guardians (void)
+{
+    struct saved {
+	struct saved *next;
+	repv obj;
+    } *changed = 0;
+
+    /* scan all guardians for unmarked objects that used to be accessible */
+    rep_guardian *g;
+    for (g = guardians; g != 0; g = g->next)
+    {
+	repv *ptr = &g->accessible;
+	/* cons cells store mark bit in CDR, so mask it out. */
+	while ((*ptr & ~rep_VALUE_CONS_MARK_BIT) != Qnil)
+	{
+	    repv cell = *ptr & ~rep_VALUE_CONS_MARK_BIT;
+	    if (!rep_GC_MARKEDP (rep_CAR (cell)))
+	    {
+		/* move object to inaccessible list */
+		struct saved *new;
+		/* have to preserve the cons mark bit in *ptr */
+		*ptr = rep_GCDR (cell) | (*ptr & rep_VALUE_CONS_MARK_BIT);
+		rep_CDR (cell) = g->inaccessible;
+		g->inaccessible = cell;
+
+		/* note that we need to mark this object */
+		new = alloca (sizeof (struct saved));
+		new->obj = rep_CAR (cell);
+		new->next = changed;
+		changed = new;
+	    }
+	    else
+		ptr = rep_CDRLOC (cell);
+
+	    /* mark the list infrastructure */
+	    rep_GC_SET_CONS (cell);
+	}
+    }
+
+    /* mark any objects that changed state */
+    while (changed != 0)
+    {
+	rep_MARKVAL (changed->obj);
+	changed = changed->next;
+    }
+}
+
+static void
+sweep_guardians (void)
+{
+    rep_guardian *g = guardians;
+    guardians = 0;
+    while (g)
+    {
+	rep_guardian *next = g->next;
+	if (!rep_GC_CELL_MARKEDP (rep_VAL (g)))
+	    rep_FREE_CELL (g);
+	else
+	{
+	    rep_GC_CLR_CELL (rep_VAL (g));
+	    g->next = guardians;
+	    guardians = g;
+	}
+	g = next;
+    }
+}
+
+static void
+print_guardian (repv stream, repv obj)
+{
+    rep_stream_puts (stream, "#<guardian>", -1, rep_FALSE);
+}
+
 
 /* Garbage collection */
 
@@ -589,20 +823,6 @@ rep_mark_value(register repv val)
 	gc_stack_high_tide = &dummy;
 #endif
 
-#ifdef GC_MINSTACK
-    /* This is a real problem. I can't safely stop marking since this means
-       that some lisp data won't have been marked and therefore the sweep
-       will screw up. But if I just keep on merrily recursing I risk
-       blowing the stack.  */
-    if(STK_SIZE <= GC_MINSTACK)
-    {
-	STK_WARN("garbage-collect(major problem!)");
-	/* Perhaps I should longjmp() back to the start of the gc, then quit
-	   totally?  */
-	return;
-    }
-#endif
-
 again:
     if(rep_INTP(val))
 	return;
@@ -651,10 +871,7 @@ again:
 
     case rep_Vector:
     case rep_Compiled:
-#ifdef rep_DUMPED
-	/* Ensure that read-only objects aren't marked */
 	if(rep_VECTOR_WRITABLE_P(val))
-#endif
 	{
 	    int i, len = rep_VECT_LEN(val);
 	    rep_GC_SET_CELL(val);
@@ -680,7 +897,13 @@ again:
 	rep_GC_SET_CELL(val);
 	break;
 
+    case rep_Number:
+	rep_GC_SET_CELL(val);
+	break;
+
     case rep_Funarg:
+	if (!rep_FUNARG_WRITABLE_P(val))
+	    break;
 	rep_GC_SET_CELL(val);
 	rep_MARKVAL(rep_FUNARG(val)->name);
 	rep_MARKVAL(rep_FUNARG(val)->env);
@@ -691,7 +914,6 @@ again:
 	    goto again;
 	break;
 
-    case rep_Var:
     case rep_Subr0:
     case rep_Subr1:
     case rep_Subr2:
@@ -710,8 +932,10 @@ again:
     }
 }
 
-DEFUN("garbage-threshold", Vgarbage_threshold, Sgarbage_threshold, (repv val), rep_Var) /*
+DEFUN("garbage-threshold", Fgarbage_threshold, Sgarbage_threshold, (repv val), rep_Subr1) /*
 ::doc:garbage-threshold::
+garbage-threshold [NEW-VALUE]
+
 The number of bytes of storage which must be used before a garbage-
 collection is triggered.
 ::end:: */
@@ -719,8 +943,10 @@ collection is triggered.
     return rep_handle_var_int(val, &rep_gc_threshold);
 }
 
-DEFUN("idle-garbage-threshold", Vidle_garbage_threshold, Sidle_garbage_threshold, (repv val), rep_Var) /*
+DEFUN("idle-garbage-threshold", Fidle_garbage_threshold, Sidle_garbage_threshold, (repv val), rep_Subr1) /*
 ::doc:idle-garbage-threshold::
+idle-garbage-threshold [NEW-VALUE]
+
 The number of bytes of storage which must be used before a garbage-
 collection is triggered when the editor is idle.
 ::end:: */
@@ -757,6 +983,8 @@ last garbage-collection is greater than `garbage-threshold'.
     }
 #endif
 
+    rep_macros_before_gc ();
+
     /* mark static objects */
     for(i = 0; i < next_static_root; i++)
 	rep_MARKVAL(*static_roots[i]);
@@ -799,10 +1027,14 @@ last garbage-collection is greater than `garbage-threshold'.
 	rep_MARKVAL(lc->args);
 	rep_MARKVAL(lc->saved_env);
 	rep_MARKVAL(lc->saved_special_env);
+	rep_MARKVAL(lc->saved_fh_env);
 	/* don't bother marking `args_evalled_p' it's always `nil' or `t'  */
 	lc = lc->next;
     }
 
+    /* move and mark any guarded objects that became inaccessible */
+    run_guardians ();
+
     /* Finished marking, start sweeping. */
 
     for(i = 0; i < TYPE_HASH_SIZE; i++)
@@ -832,14 +1064,18 @@ last garbage-collection is greater than `garbage-threshold'.
     }
 #endif
 
+    Fcall_hook (Qafter_gc_hook, Qnil, Qnil);
+
     if(rep_NILP(noStats))
     {
 	return(rep_list_4(Fcons(rep_MAKE_INT(used_cons),
-			       rep_MAKE_INT(allocated_cons - used_cons)),
+				rep_MAKE_INT(allocated_cons - used_cons)),
 		      Fcons(rep_MAKE_INT(rep_used_symbols),
-			       rep_MAKE_INT(rep_allocated_symbols - rep_used_symbols)),
-		      Fcons(rep_MAKE_INT(allocated_strings),
-			       rep_MAKE_INT(allocated_string_bytes)),
+			    rep_MAKE_INT(rep_allocated_symbols
+					 - rep_used_symbols)),
+		      rep_list_3(rep_MAKE_INT(used_strings),
+				 rep_MAKE_INT(allocated_strings),
+				 rep_MAKE_INT(allocated_string_bytes)),
 		      rep_MAKE_INT(used_vector_slots)));
     }
 
@@ -852,8 +1088,6 @@ rep_pre_values_init(void)
 {
     rep_register_type(rep_Cons, "cons", cons_cmp,
 		  rep_lisp_prin, rep_lisp_prin, cons_sweep, 0, 0, 0, 0, 0, 0, 0, 0);
-    rep_register_type(rep_Int, "integer", number_cmp,
-		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
     rep_register_type(rep_Vector, "vector", vector_cmp,
 		  rep_lisp_prin, rep_lisp_prin, vector_sweep, 0, 0, 0, 0, 0, 0, 0, 0);
     rep_register_type(rep_String, "string", string_cmp, rep_string_princ,
@@ -862,8 +1096,6 @@ rep_pre_values_init(void)
 		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
     rep_register_type(rep_Void, "void", rep_type_cmp,
 		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
-    rep_register_type(rep_Var, "var", rep_ptr_cmp,
-		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
     rep_register_type(rep_SF, "special-form", rep_ptr_cmp,
 		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
     rep_register_type(rep_Subr0, "subr0", rep_ptr_cmp,
@@ -880,6 +1112,11 @@ rep_pre_values_init(void)
 		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
     rep_register_type(rep_SubrN, "subrn", rep_ptr_cmp,
 		  rep_lisp_prin, rep_lisp_prin, 0, 0, 0, 0, 0, 0, 0, 0, 0);
+
+    rep_guardian_type = rep_register_new_type ("guardian", rep_ptr_cmp,
+					       print_guardian, print_guardian,
+					       sweep_guardians, mark_guardian,
+					       0, 0, 0, 0, 0, 0, 0);
 }
 
 void
@@ -889,6 +1126,10 @@ rep_values_init(void)
     rep_ADD_SUBR(Sgarbage_threshold);
     rep_ADD_SUBR(Sidle_garbage_threshold);
     rep_ADD_SUBR_INT(Sgarbage_collect);
+    rep_ADD_SUBR(Smake_primitive_guardian);
+    rep_ADD_SUBR(Sprimitive_guardian_push);
+    rep_ADD_SUBR(Sprimitive_guardian_pop);
+    rep_INTERN_SPECIAL(after_gc_hook);
 }
 
 void
@@ -896,10 +1137,10 @@ rep_values_kill(void)
 {
     rep_cons_block *cb = cons_block_chain;
     rep_vector *v = vector_chain;
-    rep_string *s = strings;
+    rep_string_block *s = string_block_chain;
     while(cb != NULL)
     {
-	rep_cons_block *nxt = cb->next;
+	rep_cons_block *nxt = cb->next.p;
 	rep_free(cb);
 	cb = nxt;
     }
@@ -911,48 +1152,50 @@ rep_values_kill(void)
     }
     while(s != NULL)
     {
-	rep_string *nxt = s->next;
-	rep_FREE_CELL(s);
+	int i;
+	rep_string_block *nxt = s->next.p;
+	for (i = 0; i < rep_STRINGBLK_SIZE; i++)
+	{
+	    if (!rep_CELL_CONS_P (rep_VAL(s->data + i)))
+		rep_free (s->data[i].data);
+	}
+	rep_free(s);
 	s = nxt;
     }
     cons_block_chain = NULL;
     vector_chain = NULL;
-    strings = NULL;
+    string_block_chain = NULL;
 }
 
 
 /* Support for dumped Lisp code */
 
-#ifdef rep_DUMPED
-
 void
-rep_dumped_init(void)
+rep_dumped_init(char *file)
 {
-    /* Main function is to go through all dumped symbols, interning
-       them, and changing rep_NULL values to be void. */
-    rep_symbol *sym;
-
-    /* But first, intern nil, it'll be filled in later. */
-    Qnil = Fintern_symbol(rep_VAL(rep_DUMPED_SYM_NIL), rep_void_value);
-
-    /* Initialise allocated_X counts */
-    allocated_cons = &rep_dumped_cons_end - &rep_dumped_cons_start;
-    rep_allocated_symbols = &rep_dumped_symbols_end - &rep_dumped_symbols_start;
-    /* ish.. */
-    used_vector_slots = ((&rep_dumped_vectors_end - &rep_dumped_vectors_start)
-			 + (&rep_dumped_bytecode_end - &rep_dumped_bytecode_start));
-
-    /* Stop one symbol too early, since we've already added it (nil) */
-    for(sym = &rep_dumped_symbols_start; sym < (&rep_dumped_symbols_end)-1; sym++)
+    void *dl = rep_open_dl_library (rep_string_dup (file));
+    if (dl == 0)
+	fprintf (stderr, "warning: couldn't open dumped filed %s\n", file);
+    else
     {
-	/* Second arg is [OBARRAY], but it's only checked against
-	   being a vector. */
-	Fintern_symbol(rep_VAL(sym), rep_void_value);
-	if(sym->value == rep_NULL)
-	    sym->value = rep_void_value;
-	if(sym->prop_list == rep_NULL)
-	    sym->prop_list = Qnil;
+	/* Main function is to go through all dumped symbols, interning
+	   them, and changing rep_NULL values to be void. */
+	rep_symbol *s;
+
+	/* But first, intern nil, it'll be filled in later. */
+	Qnil = Fintern_symbol (rep_VAL(rep_dumped_symbols_end - 1),
+			       rep_void_value);
+
+	/* Stop one symbol too early, since we've already added it */
+	for (s = rep_dumped_symbols_start; s < rep_dumped_symbols_end - 1; s++)
+	{
+	    /* Second arg is [OBARRAY], but it's only checked against
+	       being a vector. */
+	    Fintern_symbol (rep_VAL(s), rep_void_value);
+	    if (s->value == rep_NULL)
+		s->value = rep_void_value;
+	    if (s->prop_list == rep_NULL)
+		s->prop_list = Qnil;
+	}
     }
 }
-
-#endif /* rep_DUMPED */
